/*
 * $Id: cli.cc,v 1.10 2006/10/05 08:12:34 gleissner Exp $
 *
 * SCPM Command Line Interface
 *
 * Copyright 2001 SuSE GmbH
 *					 2002-2004 SuSE Linux AG
 *
 * Author: Joachim Gleissner <jg@suse.de>
 *
 *	 This program is free software; you can redistribute it and/or modify
 *	 it under the terms of the GNU General Public License as published by
 *	 the Free Software Foundation; either version 2 of the License, or
 *	 (at your option) any later version.
 *
 *	 This program is distributed in the hope that it will be useful,
 *	 but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 *	 GNU General Public License for more details.
 *
 *	 You should have received a copy of the GNU General Public License
 *	 along with this program; if not, write to the Free Software
 *	 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA	 02111-1307	 USA
 */

#include <stdio.h>
#include "globals.h"
#include "help_texts.h"
#include "db_commands.h"
#include <scpm.h>
#include <scpm_types.h>
#include <tokenizer.h>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <locale.h>
#include <libintl.h>
#include <signal.h>
#include <algorithm>

using namespace std;


int options = 0;
int command;
string cmd_prompt = "> ";
bool auto_switch = true;
bool list_descriptions = true;
bool list_active_only = false;
bool list_groups = true;
bool auto_group = true;
bool group_defaults = false;
bool diff_backup = false;
bool switch_only = false;
string scpm_root="";
string profile;
int journal_cmd_count = -1;
bool rollback=false;

SCPM *scpm;
SCDB *db;
SCDBUtil *util;

void cleanup_and_exit( )
{
	if ( util ) delete util;
	if ( scpm ) delete scpm;
	exit( err_no );
}

void die( string errortext, int errorno )
{
	cerr << "SCPM: " << errortext << endl;
	err_no = errorno;
	cleanup_and_exit( );
}

void sighand(int sig)
{
	switch(sig) {
		case SIGPIPE:
				// most probably some output was piped into less or something
				// so we clean up and exit quietly
			cleanup_and_exit();
		case SIGINT:
		case SIGTERM:
		case SIGQUIT:
				// currently ignored
			break;
	}
}

int get_global_opts( int argc, char **argv )
{
	int optptr=0;

	while ( optptr<argc && (argv[optptr][0] == '-') ) {
			// maybe we have a single dash as stdin identifier
		if ( argv[optptr][1] == '\0' ) {
			optptr++;
			continue;
		}
			// we accept --help as long only option
		if ( strcmp( argv[optptr], "--help" ) == 0 ) {
			cerr << HELP_TEXT_USAGE << endl;
			exit(0);
		}
		switch (argv[optptr][1])
		{
			case 'h': cerr << HELP_TEXT_USAGE << endl; exit(0); break;
			case 'v': options |= scpm_flag_verbose; break;
			case 'q': options |= scpm_flag_quiet; break;
			case 'f': options |= scpm_flag_force; break;
			case 's': options |= scpm_flag_skip; break;
			case 'm': options |= scpm_flag_hash; break;
			case 'd': options |= scpm_flag_debug; break;
			case 'b': options |= scpm_flag_boot; break;
			case 'w': options |= scpm_flag_exit_on_warning; break;
			case 'x': options |= scpm_flag_force_db; break;
			case 'o': options |= scpm_flag_read_only; break;
			case 'r':
				if (argc<=optptr+1) {
					cerr << gettext("Global option '-r' needs an argument") << endl;
					exit(1);
				}
				scpm_root = argv[++optptr];
				break;
			case 'j':
				if ((argc<=optptr+1) || (argv[optptr+1][0]=='-')) journal_cmd_count = 0;
				else journal_cmd_count = (int)strtol( argv[++optptr], NULL, 10 );
				break;
			default: cerr << gettext("Unknown global option ") << argv[optptr] << endl; exit(1);
		}
		optptr++;
	}
	return(optptr);
}
		
int get_command( int argc, char **argv)
{
	if ( argc == 0 ) {
		cerr << gettext("No command given") << endl;
		exit(1);
	}
	if			( strcmp(argv[0], "active") == 0 )					 command=CMD_ACTIVE;
	else if ( strcmp(argv[0], "list") == 0 )						 command=CMD_LIST;
	else if ( strcmp(argv[0], "enable") == 0 )					 command=CMD_ENABLE;
	else if ( strcmp(argv[0], "disable") == 0 )					 command=CMD_DISABLE;
	else if ( strcmp(argv[0], "add") == 0 )							 command=CMD_ADD;
	else if ( strcmp(argv[0], "copy") == 0 )						 command=CMD_COPY;
	else if ( strcmp(argv[0], "delete") == 0 )					 command=CMD_DELETE;
	else if ( strcmp(argv[0], "remove") == 0 )					 command=CMD_DELETE;
	else if ( strcmp(argv[0], "rename") == 0 )					 command=CMD_RENAME;
	else if ( strcmp(argv[0], "reload") == 0 )					 command=CMD_RELOAD;
	else if ( strcmp(argv[0], "switch") == 0 )					 command=CMD_SWITCH;
	else if ( strcmp(argv[0], "modify") == 0 )					 command=CMD_MODIFY;
	else if ( strcmp(argv[0], "set") == 0 )							 command=CMD_SET;
	else if ( strcmp(argv[0], "get") == 0 )							 command=CMD_GET;
	else if ( strcmp(argv[0], "status") == 0 )					 command=CMD_STATUS;
	else if ( strcmp(argv[0], "save") == 0 )						 command=CMD_SAVE;
	else if ( strcmp(argv[0], "db") == 0 )							 command=CMD_DB;
	else if ( strcmp(argv[0], "diff") == 0 )						 command=CMD_DIFF;
	else if ( strcmp(argv[0], "list_groups") == 0 )			 command=CMD_LIST_GROUPS;
	else if ( strcmp(argv[0], "activate_group") == 0 )	 command=CMD_ACT_GROUP;
	else if ( strcmp(argv[0], "deactivate_group") == 0 ) command=CMD_DEACT_GROUP;
	else if ( strcmp(argv[0], "show_group") == 0 )			 command=CMD_SHOW_GROUP;
	else if ( strcmp(argv[0], "remove_group") == 0 )		 command=CMD_REMOVE_GROUP;
	else if ( strcmp(argv[0], "reset_group") == 0 )			 command=CMD_RESET_GROUP;
	else if ( strcmp(argv[0], "reset_all") == 0 )				 command=CMD_RESET_ALL;
	else if ( strcmp(argv[0], "rebuild") == 0 )					 command=CMD_REBUILD;
	else if ( strcmp(argv[0], "update_script") == 0 )		 command=CMD_UPDATE_SCRIPT;
	else if ( strcmp(argv[0], "update_resources") == 0 ) command=CMD_UPDATE_RESOURCES;
	else if ( strcmp(argv[0], "backup") == 0 )					 command=CMD_BACKUP;
	else if ( strcmp(argv[0], "recover") == 0 )					 command=CMD_RECOVER;
	else {
		cerr << gettext("Unknown command ") << argv[0] << endl;
		exit(1);
	}

	return(1);
}

int get_command_opts( int argc, char **argv )
{
	int optptr=0;

	while ( optptr < argc && (argv[optptr][0] == '-') ) {
		if ( argv[optptr][1] == '\0' ) {
				// is no option, maybe stdin identifier
			return(optptr);
		}
		switch (command) {
			case CMD_ADD:
				if ( argv[optptr][1] == 'a' ) auto_switch=true;
				else if ( argv[optptr][1] == 'n' ) auto_switch=false;
				else die( (string)gettext("Unknown add option")+" "+argv[optptr], ERR_USER );
				break;
			case CMD_LIST:
				if ( argv[optptr][1] == 'd' ) list_descriptions=true;
				else if ( argv[optptr][1] == 'n' ) list_descriptions=false;
				else die( (string)gettext("Unknown list option")+" "+argv[optptr], ERR_USER );
				break;
			case CMD_SWITCH:
			case CMD_SAVE:
			case CMD_RELOAD:
				if ( argv[optptr][1] == 'r' ) list_groups=false;
				else if ( argv[optptr][1] == 'g' ) list_groups=true;
				else die( (string)gettext("Unknown switch/save option")+" "+argv[optptr], ERR_USER );
				break;
			case CMD_LIST_GROUPS:
				if ( argv[optptr][1] == 'd' ) list_descriptions=true;
				else if ( argv[optptr][1] == 'n' ) list_descriptions=false;
				else if ( argv[optptr][1] == 'a' ) list_active_only=true;
				else die( (string)gettext("Unknown list_groups option")+" "+argv[optptr], ERR_USER );
				break;
			case CMD_ACT_GROUP:
				if ( argv[optptr][1] == 'd' ) auto_group=false;
				if ( argv[optptr][1] == 'a' ) auto_group=true;
				else die( (string)gettext("Unknown activate_group option")+" "+argv[optptr], ERR_USER );
				break;
			case CMD_DEACT_GROUP:
				if ( argv[optptr][1] == 'd' ) auto_group=false;
				if ( argv[optptr][1] == 'a' ) auto_group=true;
				else die( (string)gettext("Unknown deactivate_group option")+" "+argv[optptr], ERR_USER );
				break;
			case CMD_SHOW_GROUP:
				if ( argv[optptr][1] == 'd' ) group_defaults=true;
				else die( (string)gettext("Unknown show_group option")+" "+argv[optptr], ERR_USER );
				break;
			case CMD_BACKUP:
				if ( argv[optptr][1] == 'a' ) profile="all";
				else if ( argv[optptr][1] == 'p' ) {
					if ( argc==optptr+1 ) die( gettext("unexpected end of command"), ERR_USER );
					profile=argv[++optptr];
				}
				else die( (string)gettext("Unknown backup option")+" "+argv[optptr], ERR_USER );
				break;
			case CMD_DIFF:
				if ( argv[optptr][1] == 'b' ) diff_backup=true;
				break;
			case CMD_RECOVER:
				if ( argv[optptr][1] == 'b' ) rollback=true;
				break;
			default:
				die( gettext("Unknown command option"), ERR_USER );
		}
		optptr++;
	}
	return(optptr);
}

string get_command( vector<string> &args )
{
	if ( isatty( 0 ) ) cerr << cmd_prompt;
	size_t n = 0;
	ssize_t read;
	char *line = NULL;
	args.clear();
	read = getline( &line, &n, stdin );
	if ( read == -1 ) return "quit";
	line[strlen(line)-1]='\0'; // kill newline
	Tokenizer tok( line );
	string cmd_str;
	try {
		cmd_str = tok.GetToken();
		for (;;) args.push_back( tok.GetToken() );
	}
	catch( Tokenizer::end_of_tokens ) {
	}
	return cmd_str;
}

int db_command_by_name( string cmd )
{
	if			( cmd == "" ) return DB_NOOP;
	else if ( cmd == "help" ) return DB_HELP;
	else if ( cmd == "quit" || cmd == "exit" ) return DB_EXIT;
	else if ( cmd == "load" ) return DB_LOAD;
	else if ( cmd == "save" ) return DB_SAVE;
	else if ( cmd == "info" ) return DB_INFO;
	else if ( cmd == "db_version" ) return DB_DB_VERSION;
	else if ( cmd == "scpm_version" ) return DB_SCPM_VERSION;
	else if ( cmd == "create" ) return DB_CREATE;
	else if ( cmd == "drop" ) return DB_DROP;
	else if ( cmd == "dump" ) return DB_DUMP;
	else if ( cmd == "show" ) return DB_SHOW;
	else if ( cmd == "minimize" ) return DB_MINIMIZE;
	else if ( cmd == "maximize" ) return DB_MAXIMIZE;
	else if ( cmd == "customize" ) return DB_CUSTOMIZE;
	else if ( cmd == "rebuild_deps" ) return DB_REBUILD_DEPS;
	else if ( cmd == "convert" ) return DB_CONVERT;
	else if ( cmd == "clean" ) return DB_CLEAN;
	else if ( cmd == "get" ) return DB_GET;
	else if ( cmd == "set" ) return DB_SET;
	else if ( cmd == "export" ) return DB_EXPORT;
	else if ( cmd == "import" ) return DB_IMPORT;
	return DB_UNKNOWN;
}

void exec_db_command( int cmd, vector<string> &args )
{
	switch( cmd ) {
		case DB_UNKNOWN:
			cerr << gettext( "Unknown command" ) << " "
					 << gettext( "(type 'help' for help)" ) << endl;
			break;
		case DB_NOOP: break;
		case DB_QUIT:
		case DB_EXIT: cleanup_and_exit( ); break;
		case DB_HELP: cmd_help( args ); break;
		case DB_LOAD: cmd_load( args ); break;
		case DB_SAVE: cmd_save( args ); break;
		case DB_INFO: cmd_info( args ); break;
		case DB_DB_VERSION: cmd_db_version( args ); break;
		case DB_SCPM_VERSION: cmd_scpm_version( args ); break;
		case DB_CREATE: cmd_create( args ); break;
		case DB_DROP: cmd_drop( args ); break;
		case DB_DUMP: cmd_dump( args ); break;
		case DB_SHOW: cmd_show( args ); break;
		case DB_MINIMIZE: cmd_minimize( args ); break;
		case DB_MAXIMIZE: cmd_maximize( args ); break;
		case DB_CUSTOMIZE: cmd_customize( args ); break;
		case DB_REBUILD_DEPS: cmd_rebuild_deps( args ); break;
		case DB_CONVERT: cmd_convert( args ); break;
		case DB_CLEAN: cmd_clean( args ); break;
		case DB_GET: cmd_get( args ); break;
		case DB_SET: cmd_set( args ); break;
		case DB_EXPORT: cmd_export( args ); break;
		case DB_IMPORT: cmd_import( args ); break;
	}
}

void db_command( int argc, char *argv[] )
{
	vector<string> args;
	if ( argc != 0 ) {
			// command given
		for ( int i=1; i<argc; i++ ) args.push_back( argv[i] );
		scpm = new SCPM( scpm_flag_quiet|scpm_flag_force_db, cerr, cerr, scpm_root );
		db = SCDB::GetHandle();
		util = new SCDBUtil();
		exec_db_command( db_command_by_name( argv[0] ), args );
	}
	else {
			// interactive mode
		if ( isatty( 0 ) ) cerr << "SCDB Utility (SCPM version "<< SCPM_VERSION << ")" << endl;
		cerr << "Warning: SCDB gets now auto loaded also in interactive mode." << endl
				 << "Warning: You can still use the 'load' command to load another db file." << endl;
		scpm = new SCPM( scpm_flag_quiet|scpm_flag_force_db, cerr, cerr, scpm_root );
		db = SCDB::GetHandle();
		util = new SCDBUtil();
		for (;;) {
			string cmd = get_command( args );
			exec_db_command( db_command_by_name( cmd ), args );
		}
	}
}

void get_conf_opts()
{
		// this is a dirty hack to check whether SWITCH_MODE is set in
		// /etc/scpm.conf
		// it should be replace by something more accurate
	ifstream cfgfile( "/etc/sysconfig/scpm" );
	char buf[255];
	while( cfgfile.getline( buf, 254 ) ) {
		if ( static_cast<string>(buf) == "SWITCH_MODE=\"skip\"" ) {
			options |= scpm_flag_skip;
			break;
		}
		else if ( static_cast<string>(buf) == "SWITCH_MODE=\"force\"" ) {
			options |= scpm_flag_force;
			break;
		}
	}
	cfgfile.close();
		// here comes the next hack, should be done by libscpm sometime
	struct stat tmp;
	if ( stat( "/var/lib/scpm/active_groups", &tmp ) == -1 )
		list_groups = false; // forcing single resource mode, not in group mode
}

void query_by_resources( switch_info_t &switch_info )
{
	cout << gettext( "The following resources are modified" ) << ": " << endl;
	for (unsigned int i=0; i<switch_info.modified_resources.size(); i++) {
		char c;
		do {
			cout << switch_info.modified_resources[i].resource_name << " (" <<
				switch_info.modified_resources[i].resource_type << ") " ;
			if ( switch_info.modified_resources[i].is_new )
				cout << gettext(" NEW  add ([y]es/[n]o/[a]bort)? ");
			else if ( switch_info.modified_resources[i].is_deleted )
				cout << gettext(" DELETED  remove ([y]es/[n]o/[a]bort)? ");
			else 
				cout << gettext("  save ([y]es/[n]o/[a]bort/[d]iff/app[l]y/[p]atch)? ");
			cin >> c;
			if ( c == 'y' || c == 'j' ) switch_info.modified_resources[i].save = true;
			if ( c == 'n' ) switch_info.modified_resources[i].save = false;
			if ( c == 'a' ) { err_no=ERR_USERABORT; cleanup_and_exit(); }
			if ( c == 'd' ) scpm->ShowChanges( cout, 
																				 switch_info.modified_resources[i].resource_type,
																				 switch_info.modified_resources[i].resource_name );
			if ( c == 'l' ) {
				switch_info.modified_resources[i].save = true;
				switch_info.modified_resources[i].save_mode = save_all;
			}
			if ( c == 'p' ) {
				switch_info.modified_resources[i].save = true;
				switch_info.modified_resources[i].save_mode = patch_all;
			}
		} while ( c != 'y' && c != 'n' && c != 'j' && c != 'l' && c != 'p' );
	}
}

void query_by_group( switch_info_t &switch_info )
{
	cout << gettext( "The following resource groups are modified" ) << ": " << endl;

	vector<resource_group_t> groups;

	for (unsigned int i=0; i<switch_info.modified_resources.size(); i++)
		for (unsigned int j=0; j<switch_info.modified_resources[i].groups.size(); j++)
			groups.push_back( switch_info.modified_resources[i].groups[j] );
	for (unsigned int i=0; i<groups.size(); i++ ) {
		bool found = false;
		for (unsigned int j=0; j<i; j++ )
			if ( groups[i].name == groups[j].name ) found = true;
		if ( !found ) {
			char c='@';
			while(c!='y' && c!='j' && c!='n' && c!='a' && c!='r') {
				cout << groups[i].name;
				for ( unsigned int j=groups[i].name.size(); j-15>0; j++ ) cout << " ";
				cout << "(" << gettext(groups[i].description.c_str()) << ") "
						 << gettext("  save ([y]es/[n]o/show [r]esources/[a]bort)? ");
				cin >> c;
			}
			if ( c == 'y' || c == 'j' ) {
				for (unsigned int k=0; k<switch_info.modified_resources.size(); k++ )
					for (unsigned int l=0; l<switch_info.modified_resources[k].groups.size(); l++ )
						if (switch_info.modified_resources[k].groups[l].name == groups[i].name) {
							switch_info.modified_resources[k].save = true;
							break;
						}
			}
			if ( c == 'n' ) {
				for (unsigned int k=0; k<switch_info.modified_resources.size(); k++ )
					for (unsigned int l=0; l<switch_info.modified_resources[k].groups.size(); l++ )
						if (switch_info.modified_resources[k].groups[l].name == groups[i].name) {
							switch_info.modified_resources[k].save = false;
							break;
						}
			}
			if ( c == 'r' ) {
				for (unsigned int k=0; k<switch_info.modified_resources.size(); k++ )
					for (unsigned int l=0; l<switch_info.modified_resources[k].groups.size(); l++ )
						if (switch_info.modified_resources[k].groups[l].name == groups[i].name) {
							char ci;
							do {
								cout << switch_info.modified_resources[k].resource_name << " (" <<
									switch_info.modified_resources[k].resource_type << ") " ;
								if ( switch_info.modified_resources[k].is_new )
									cout << gettext(" NEW  add ([y]es/[n]o/[a]bort)? ");
								else if ( switch_info.modified_resources[i].is_deleted )
									cout << gettext(" DELETED  remove ([y]es/[n]o/[a]bort)? ");
								else 
									cout << gettext("  save ([y]es/[n]o/[a]bort/[d]iff/app[l]y/[p]atch)? ");
								cin >> ci;
								if ( ci == 'y' || c == 'j' ) switch_info.modified_resources[k].save = true;
								if ( ci == 'n' ) switch_info.modified_resources[k].save = false;
								if ( ci == 'a' ) { err_no=ERR_USERABORT; cleanup_and_exit(); }
								if ( ci == 'd' ) scpm->ShowChanges( cout, 
																										switch_info.modified_resources[k].resource_type,
																										switch_info.modified_resources[k].resource_name );
								if ( ci == 'l' ) {
									switch_info.modified_resources[k].save = true;
									switch_info.modified_resources[k].save_mode = save_all;
								}
								if ( ci == 'p' ) {
									switch_info.modified_resources[k].save = true;
									switch_info.modified_resources[k].save_mode = patch_all;
								}
							} while ( ci != 'y' && ci != 'n' && ci != 'j' && ci != 'l' && ci != 'p' );
						}
			}
			if ( c == 'a' ) { err_no=ERR_USERABORT; cleanup_and_exit(); }
		}
	}
}

string GetBackupVersionByNum(string type, string name, int num)
{
	vector<backup_info_t> bi;
	string ret;

	// we need the list of backups
	if (scpm->ListBackups(profile, &bi)) {
		// Search 
		if (bi.size()>0) {
			unsigned int j;
			for (j=0; j < bi.size(); j++)
				if (bi[j].type == type && bi[j].name==name) break;
			if (j <  bi.size()) {
				// sort descending
				sort(bi[j].versions.begin(), bi[j].versions.end(), greater< pair<time_t, string> >());
				if (num<0 || num >= (int)bi[j].versions.size())
					die(gettext("backup does not exists"), ERR_USER);
				// found right backup
				ret =  bi[j].versions[num].second;
			} else die(gettext("no backup found"), ERR_USER);
		} else die(gettext("no backup found"), ERR_USER);
	} else die( scpm_error, ERR_SUBSYS );
	return ret;
}

int main( int argc, char **argv )
{
	int optptr;
	int retval;
	switch_info_t switch_info;
	scpm = NULL;
	util = NULL;
	string version="" ;

	setlocale( LC_ALL, "" );
	bindtextdomain( "scpm", NULL );
	textdomain( "scpm" );

	argc--;
	argv=&argv[1];

	if (argc==0) {
		cerr << "SCPM Version " << SCPM_VERSION << endl;
		cerr << HELP_TEXT_USAGE << endl;
		exit(1);
	}

		//if ( geteuid() != 0 ) die( gettext("you have to be root to use SCPM"), ERR_UNSPEC );

	optptr=get_global_opts( argc, argv );
	argc -= optptr;
	argv=&argv[optptr];
	optptr=get_command( argc, argv );
	argc -= optptr;
	argv=&argv[optptr];
	optptr=get_command_opts( argc, argv );
	argc -= optptr;
	argv=&argv[optptr];

	get_conf_opts();

	if ( getuid() != 0 ) {
		if ( geteuid() == 0 ) {
				// ok, we are suid root, check if the user is authorized
			struct passwd *pw;
			pw = getpwuid( getuid() );
			if (!pw) {
				cerr << gettext("Could not determine your login name") << endl
						 << gettext("Running scpm in read-only mode") << endl;
				options |= scpm_flag_read_only;
			}
			else {
				bool found=false;
				char line[255];
				ifstream in( "/etc/scpm.users" );
				while ( in.getline( line, 254 ) ) {
					if ( line[0] == '#' ) continue;
					char user[255];
					char rights[255];
					sscanf(line, "%s %s", user, rights);
					if ( strcmp( pw->pw_name, user ) == 0 ) {
						found = true;
						if ( strcmp( rights, "switch" ) == 0 ) {
							switch_only = true;
						}
						else if ( strcmp( rights, "all" ) == 0 ) {
							switch_only = false;
						}
						else if ( strcmp( rights, "" ) == 0 ) {
							cerr << gettext("WARNING: /etc/scpm.users has obsolete format") << endl;
							switch_only = false;
						}
						else {
							die( gettext("Error in /etc/scpm.users"), ERR_USER);
						}
						break;
					}
				}
				if ( !found ) {
					cerr << gettext("You are not root and not in /etc/scpm.users") << endl
							 << gettext("Running scpm in read-only mode") << endl;
					options |= scpm_flag_read_only;
				}
			}
		}
		else {
				// ok, no root, no fun, only read-only functions will work
			if ( !(options & scpm_flag_read_only) ) {
				cerr << gettext("not running as root, read-only mode enabled") << endl;
				options |= scpm_flag_read_only;
			}
		}
	}
	if ( options & scpm_flag_read_only ) {
		if ( (command != CMD_LIST) &&
				 (command != CMD_ACTIVE) &&
				 (command != CMD_STATUS) )
			die( gettext("this command is not available in read-only mode"), ERR_USER );
	}
	if ( switch_only ) {
		switch (command) {
			case CMD_LIST:
			case CMD_ACTIVE:
			case CMD_STATUS:
			case CMD_SWITCH:
			case CMD_SAVE:
			case CMD_RELOAD:
			case CMD_DIFF:
			case CMD_LIST_GROUPS:
			case CMD_SHOW_GROUP:
				break;
			default:
				die( gettext("you have no permission to execute this command"), ERR_USER );
		}
	}

	signal(SIGPIPE, sighand);
	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);

	if ( command != CMD_DB ) {
		if ( command == CMD_STATUS ) options |= scpm_flag_force_db;
		scpm = new SCPM( options, cout, cerr, scpm_root );
	}
	if ( journal_cmd_count > -1 ) scpm_journal_abort_steps = journal_cmd_count;

	retval=0;

	switch (command)
	{
		case CMD_ENABLE:
			if ( !scpm->Enable() ) die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_DISABLE:
			if ( !scpm->Disable() ) die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_ACTIVE:
		{
			string profile;
			if ( !scpm->Active( profile ) ) die( scpm_error, ERR_SUBSYS );
			cout << profile << endl;
			break;
		}
		case CMD_LIST:
		{
			vector<string> l;
			if ( !scpm->List( l ) ) die( scpm_error, ERR_SUBSYS );
			for ( unsigned int i=0 ; i<l.size(); i++ ) {
				cout << l[i];
				if ( list_descriptions ) {
					int j=l[i].size()-20;
					do {
						cout << " ";
					} while (j++<=0);
					string desc;
					scpm->Get( "description", desc, l[i] );
					cout << desc;
				}
				cout << endl;
			}
			break;
		}
		case CMD_ADD:
			if ( argc<1 ) die( gettext( "add needs one argument" ), ERR_USER );
			if ( !scpm->Add( argv[0], auto_switch ) ) die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_COPY:
			if ( argc<2 ) die( gettext( "copy needs two arguments" ), ERR_USER );
			if ( !scpm->Copy( argv[0], argv[1] ) ) die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_DELETE:
			if ( argc<1 ) die( gettext( "delete needs one argument" ), ERR_USER );
			if ( !scpm->Delete( argv[0] ) ) die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_RENAME:
			if ( argc<2 ) die( gettext( "rename needs two arguments" ), ERR_USER );
			if ( !scpm->Rename( argv[0], argv[1] ) ) die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_RELOAD:
				// the Reload() function is broken, we implement it high level
			if ( argc!=0 ) die( gettext( "reload must not have arguments" ), ERR_USER );
			{
				string profile;
				if ( !scpm->Active( profile ) ) die( scpm_error, ERR_SUBSYS );
				if ( !scpm->PrepareSwitch( profile, switch_info ) ) die( scpm_error, ERR_SUBSYS );
			}
		case CMD_SAVE:
			if ( command == CMD_SAVE ) {
				if ( argc!=0 ) die( gettext( "save must not have arguments" ), ERR_USER );
				{
					string profile;
					if ( !scpm->Active( profile ) ) die( scpm_error, ERR_SUBSYS );
					if ( !scpm->PrepareSwitch( profile, switch_info ) ) die( scpm_error, ERR_SUBSYS );
				}
			}
		case CMD_SWITCH:
			if ( command == CMD_SWITCH ) {
				if ( argc<1 ) die( gettext( "switch needs one argument" ), ERR_USER );
				if ( !scpm->PrepareSwitch( argv[0], switch_info ) ) die( scpm_error, ERR_SUBSYS );
			}
			if ( switch_info.profile_modified &&
					 ( ! (options & scpm_flag_force) ) &&
					 ( ! (options & scpm_flag_skip) ) ) {
				if ( switch_info.modified_resources.size() )
					if ( list_groups )
						query_by_group( switch_info );
					else
						query_by_resources( switch_info );
			}
			if ( !scpm->Save( switch_info )	 ) {
				if ( command == CMD_SWITCH ) {
					cerr << scpm_error << endl;
					cout << gettext("Saving resource failed. Do you want to proceed anyway (not recommended)? ");
					char c;
					cin >> c;
					if ( (c != 'y') && (c != 'j') ) {
						err_no = ERR_SUBSYS;
						cleanup_and_exit();
					}
				}
				else
					die( scpm_error, ERR_SUBSYS );
			}
			if ( command != CMD_SAVE )
				if ( !scpm->Switch( switch_info ) )
					die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_MODIFY:
			if ( argc<2 ) die( gettext( "modify needs two arguments" ), ERR_USER );
			if ( !scpm->Modify( argv[0], argv[1] ) ) die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_DIFF:
			if ( argc<2 ) die( gettext( "diff needs two arguments" ), ERR_USER );
			if ( argc==3 ) version = GetBackupVersionByNum(argv[0], argv[1], atoi(argv[2]) - 1);
			if ( !scpm->ShowChanges( cout, argv[0], argv[1], diff_backup, version ) ) die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_SET:
			if ( argc<2 ) die( gettext( "set needs at least two arguments" ), ERR_USER );
				// we abuse this function to test the config interface
			if ( strcmp( argv[0], "debug" ) == 0 ) {
				if ( !scpm->SetConfigOption( "DEBUG", argv[1] ) ) die( scpm_error, ERR_SUBSYS );
				break;
			}
			if ( strcmp( argv[0], "verbose" ) == 0 ) {
				if ( !scpm->SetConfigOption( "VERBOSE", argv[1] ) ) die( scpm_error, ERR_SUBSYS );
				break;
			}
			if ( strcmp( argv[0], "switchmode" ) == 0 ) {
				if ( !scpm->SetConfigOption( "SWITCH_MODE", argv[1] ) ) die( scpm_error, ERR_SUBSYS );
				break;
			}
			if ( strcmp( argv[0], "bootmode" ) == 0 ) {
				if ( !scpm->SetConfigOption( "BOOT_MODE", argv[1] ) ) die( scpm_error, ERR_SUBSYS );
				break;
			}
			if ( (strcmp( argv[1], "-d" ) == 0) ) argv[1][0]='\0';
			if ( argc==2 ) {
				if ( !scpm->Set( argv[0], argv[1] ) ) die( scpm_error, ERR_SUBSYS );
			}
			else
				if ( !scpm->Set( argv[0], argv[1], argv[2] ) ) die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_GET:
		{
			string result;
			if ( argc<1 ) die( gettext( "get needs at least one argument" ), ERR_USER );
			if ( strcmp( argv[0], "all" ) == 0 ) {
				string profile="";
				if ( argc>1 ) profile=argv[1];
				if ( !scpm->Get( "resource_set", result ) ) die( scpm_error, ERR_SUBSYS );
				cout << "resource set		 : " << result << endl;
				if ( !scpm->Get( "description", result, profile ) ) die( scpm_error, ERR_SUBSYS );
				cout << "description		 : " << result << endl;
				if ( !scpm->Get( "prestart", result, profile ) ) die( scpm_error, ERR_SUBSYS );
				cout << "prestart script : " << result << endl;
				if ( !scpm->Get( "prestop", result, profile ) ) die( scpm_error, ERR_SUBSYS );
				cout << "prestop script	 : " << result << endl;
				if ( !scpm->Get( "poststart", result, profile ) ) die( scpm_error, ERR_SUBSYS );
				cout << "poststart script: " << result << endl;
				if ( !scpm->Get( "poststop", result, profile ) ) die( scpm_error, ERR_SUBSYS );
				cout << "poststop script : " << result << endl;
				break;
			}
			if ( argc<=1 ) {
				if ( !scpm->Get( argv[0], result ) ) die( scpm_error, ERR_SUBSYS );
			}
			else
				if ( !scpm->Get( argv[0], result, argv[1] ) ) die( scpm_error, ERR_SUBSYS );
			cout << result << endl;
			break;
		}
		case CMD_STATUS:
		{
			struct scpm_status_t status;
			if ( !scpm->Status( status ) ) die( scpm_error, ERR_SUBSYS );
			cout << "initialized      : " << (status.initialized ? "yes" : "no") << endl
					 << "enabled          : " << (status.enabled ? "yes" : "no") << endl
					 << "db loadable      : " << (status.db_loaded ? "yes" : "no")	<< endl
					 << "db up to date    : " << (status.db_uptodate ? "yes" : "no") << endl
					 << "locked           : " << (status.locked ? "yes" : "no") << endl
					 << "needs reinit     : " << (status.needs_reinit ? "yes" : "no") << endl
					 << "needs recover    : " << (status.needs_recover ? "yes" : "no") << endl
					 << "active profile   : " << status.active_profile << endl
					 << "scpm version     : " << status.scpm_version << endl
					 << "db version       : " << status.db_version << endl
					 << "db format        : " << status.db_format << endl
					 << "db format version: " << status.db_format_version << endl;
			break;
		}
		case CMD_LIST_GROUPS:
			if ( argc>0 ) die( gettext("list_groups must not have arguments"), ERR_USER );
			{
				vector<resource_group_t> groups;
				if ( !scpm->ListResourceGroups( groups ) ) die( scpm_error, ERR_SUBSYS );
				for ( unsigned int i=0 ; i<groups.size(); i++ ) {
					if ( list_active_only && !groups[i].active ) continue;
					cout << groups[i].name;
					if ( list_descriptions ) {
						for ( unsigned int j=groups[i].name.size(); j-20>0; j++ ) cout << " ";
						cout << groups[i].description;
					}
					cout << endl;
				}
			}
			break;
		case CMD_ACT_GROUP:
			if ( argc<1 ) die( gettext("activate_group needs an argument"), ERR_USER );
			if ( !scpm->ActivateResourceGroup( argv[0], auto_group ) )
				die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_DEACT_GROUP:
			if ( argc<1 ) die( gettext("deactivate_group needs an argument"), ERR_USER );
			if ( !scpm->DeactivateResourceGroup( argv[0], auto_group ) )
				die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_SHOW_GROUP:
			if ( argc<1 ) die( gettext("show_group needs an argument"), ERR_USER );
			{
				vector<resource_entry_t> group;
				if ( !scpm->GetResourceGroup( argv[0], group, group_defaults  ) )
					die( scpm_error, ERR_SUBSYS );
				for(unsigned int i=0; i<group.size(); i++)
					cout << group[i].name << " (" << group[i].type << ") " <<
						(group[i].active ? "" : "(inactive)") << endl;
			}
			break;
		case CMD_REMOVE_GROUP:
			if ( argc<1 ) die( gettext("remove_group needs an argument"), ERR_USER );
			if ( !scpm->DeleteResourceGroup( argv[0] ) )
				die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_RESET_GROUP:
			if ( argc<1 ) die( gettext("reset_group needs an argument"), ERR_USER );
			if ( !scpm->ResetResourceGroup( argv[0] ) )
				die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_RESET_ALL:
			if ( argc>0 ) die( gettext("reset_all must not have arguments"), ERR_USER );
			if ( !scpm->ResetAllGroups( ) )
				die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_REBUILD:
			if ( argc>0 ) die( gettext("rebuild must not have arguments"), ERR_USER );
			if ( !scpm->RebuildDB() )
				die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_UPDATE_SCRIPT:
			if ( argc<1) die( gettext("update needs at least one argument"), ERR_USER );
			if ( strcmp( argv[0], "-" ) != 0 ) {
				struct stat s;
				if ( stat( argv[0], &s ) == -1 )
					die( gettext("script file is inaccessible"), ERR_USER );
			}
			{
				vector<pair<string, string> > resources;
				for (signed int i=1; i<argc; i+=2) {
					if (i+1>=argc) die( gettext("you need to specify resource types and names"), ERR_USER );
					resources.push_back( make_pair( argv[i], argv[i+1] ) );
				}
				if ( !scpm->Update( argv[0], resources ) )
					die( scpm_error, ERR_SUBSYS );
			}
			break;
		case CMD_UPDATE_RESOURCES:
			if ( argc>0 ) die( gettext("update_resources must not have arguments"), ERR_USER );
			if ( !scpm->UpdateResources() )
				die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_BACKUP:
			if ( argc<1) die( gettext("backup needs at least one argument"), ERR_USER );
			{
				vector<pair<string, string> > resources;
				vector<string> versions;
				
				// get resourcelist from args 
				for (signed int i=1; i<argc; i+=3) {
					if (i+1>=argc) die( gettext("you need to specify resource types and names"), ERR_USER );
					if ((strcmp(argv[0], "restore")==0 || strcmp(argv[0], "remove")==0 ) && i+2>=argc) die( gettext("you need to specify resource types, names and versions"), ERR_USER);
					resources.push_back( make_pair( argv[i], argv[i+1] ) );

					// if executing a restore command we have to resolve which backup to restore
					if (strcmp(argv[0], "restore")==0 || strcmp(argv[0], "remove")==0)
						versions.push_back(GetBackupVersionByNum(argv[i], argv[i+1], atoi(argv[i+2]) - 1));
				}

				// Process Backupcommand
				if (strcmp(argv[0], "restore") == 0) {
					// restore backup
					if ( !scpm->RestoreBackup( resources, versions, profile ) )
						die( scpm_error, ERR_SUBSYS );
				}
				else if (strcmp(argv[0], "remove") == 0) {
					// remove backup
					if ( !scpm->RemoveBackup( resources, versions, profile ) )
						die( scpm_error, ERR_SUBSYS );
				}
				else if (strcmp(argv[0], "add") == 0) {
					// add backup
					if ( !scpm->AddBackup( resources, profile ) )
						die( scpm_error, ERR_SUBSYS );
				}
				else if (strcmp(argv[0], "list") == 0) {
					vector<backup_info_t> bi;

					// get list from backend
					if (!scpm->ListBackups(profile, &bi))
						die(scpm_error, ERR_SUBSYS );
				
					// output a list of resources with available backups
					for (unsigned int i=0; i<bi.size(); i++) {
						// sort the different versions of backups descending 
						sort(bi[i].versions.begin(), bi[i].versions.end(), greater< pair<time_t, string> >());
							
						// ouput resource and backups
						cout << bi[i].name << " (" << bi[i].type << ")";
						if (profile=="all")
							cout << " (profile " << bi[i].profile << ")";
						cout << endl ;
						
						// output all available backup versions
						for (unsigned int j=1; j <= bi[i].versions.size(); j++) {
							cout << "  (" << j << ") " << ctime(&(bi[i].versions[j-1].first)); 
						}
						cout << endl;
					}
				}
			}
			break;
		case CMD_RECOVER:
			if (!scpm->Recover(rollback)) die( scpm_error, ERR_SUBSYS );
			break;
		case CMD_DB:
			db_command( argc, argv );
			break;
		default:
			die( gettext("Unknown command number. This is a bug."), ERR_SYSTEM );
			break;
	}

	cleanup_and_exit();
}
