
/*
 * $Id: db_commands.cc,v 1.3 2008/10/22 10:12:41 hschaa Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	SCDB manipulation utility
 * File				 :	db_commands.cc
 * Description :	SCDB Utility commands
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2002-2004 SuSE Linux AG
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "db_commands.h"
#include "help_texts.h"
#include <string>
#include <vector>
#include <scdb.h>
#include <scdb_util.h>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <libintl.h>

using namespace std;

bool check_if_valid_arg( string arg, string type )
{
	if ( type == "resource" ) {
		if ( (arg.find("deps") != 0) && (arg.find("data") != 0) ) return false;

			// resource args have the syntax {deps|data}:type=value

		int state=0;
		for ( unsigned int i=0; i<arg.size(); i++ ) {
			cerr << arg.at(i) ;
			switch (state) {
				case 0:
						// at begin
					if ( ! isalpha( arg.at(i) ) ) return false;
					state=1;
					break;
				case 1:
						// in first word
					if ( arg.at(i) == ':' ) { state=2; break; }
					if ( arg.at(i) == '=' ) return false;
					break;
				case 2:
						// at ':'
					if ( arg.at(i) == ':' ) return false;
					if ( arg.at(i) == '=' ) return false;
					state=3;
					break;
				case 3:
						// in second word
					if ( arg.at(i) == ':' ) return false;
					if ( arg.at(i) == '=' ) { state=4; break; }
					break;
				case 4:
						// at '='
					if ( arg.at(i) == ':' ) return false;
					if ( arg.at(i) == '=' ) return false;
					state=5;
					break;
				case 5:
						// in last word
					if ( arg.at(i) == ':' ) return false;
					if ( arg.at(i) == '=' ) return false;
					state=6;
					break;
			}
		}
		if ( state != 6 ) return false;
	}
	return true;
}


void cmd_load( vector<string> &args )
{
	try {
		db->DropDB();
		if ( args.empty() )
			db->LoadDB( );
		else
			db->LoadDB( args[0] );
	}
	catch( SCDB::db_action_failed &e ) {
		cerr << gettext( "Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

void cmd_save( vector<string> &args )
{
	try {
		if ( args.empty() )
			db->SaveDB( );
		else if ( args[0] == "-" )
			db->SaveDB( cout );
		else {
			ofstream dest( args[0].c_str() );
			if ( dest.bad() ) {
				cerr << "could not open " << args[0] << " for writing" << endl;
				err_no = ERR_SYSTEM;
				return;
			}
			char s[]	= "/tmp/scdb.tmp.XXXXXX";
			if ( mktemp( s ) == NULL ) {
				cerr << "mktemp() failed" << endl;
				err_no = ERR_SYSTEM;
				dest.close();
				return;
			}
			ofstream tmpdb( s );
			if ( tmpdb.bad() ) {
				cerr << "could not open " << s << " for writing" << endl;
				err_no = ERR_SYSTEM;
				dest.close();
				return;
			}
			db->SaveDB( tmpdb );
			tmpdb.close();
			ifstream src( s );
			dest << src.rdbuf();
			src.close();
			dest.close();
		}
	}
	catch( SCDB::db_action_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

void cmd_info( vector<string> &args )
{
	cout << "SCPM version: " << SCPM_VERSION << endl;
	cout << "SCDB version: " << db->GetSCDBVersion() << endl;
	cout << "SCDB format:	 " << db->GetFormat()
			 << "	 formatversion: " << db->GetFormatVersion() << endl;
}

void cmd_db_version( vector<string> &args )
{
	cout << db->GetSCDBVersion() << endl;
}

void cmd_scpm_version( vector<string> &args )
{
	cout << SCPM_VERSION << endl;
}

void cmd_create( vector<string> &args )
{
	if ( args.size() < 3 ) {
		cerr << gettext( "create needs at least 3 arguments" ) << endl;
		err_no = ERR_USER;
		return;
	}

	if ( args[0] == "resource" ) {
		if ( args.size() > 3 ) {
				// optional args given, check if valid
			for ( unsigned int i=3; i<args.size(); i++ ) 
				if ( ! check_if_valid_arg( args[i], "resource" ) ) {
					err_no=ERR_USER ;
					cerr << gettext( "invalid argument syntax" ) << endl;
					return;
				}
		}
		try {
			util->CreateResource( args[1], args[2] );
				/*
					db->ResourceAdd( args[1], args[2] );
					for ( unsigned int i=3; i<args.size(); i++ ) {
					if ( args[i].find( "deps:")	 == 0 )
					db->ResourceAddDep( args[1], args[2], 
					args[i].substr( args[i].find( "=" )+1 ),
					args[i].substr( args[i].find( ":" )+1, 
					args[i].find( "=" )-1 - args[i].find(":") ) );
					if ( args[i].find( "data:") == 0 )
					db->ResourceAddData( args[1], args[2], 
					args[i].substr( args[i].find( ":" )+1, 
					args[i].find( "=" )-1 - args[i].find(":") ),
					args[i].substr( args[i].find( "=" )+1 ) );
					}
				*/
		}
		catch( SCDB::db_action_failed &e ) {
			cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
			err_no = ERR_SUBSYS;
			return;
		}
		catch( SCDBUtil::operation_failed &e ) {
			cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
			err_no = ERR_SUBSYS;
			return;
		}
	}
}

void cmd_drop( vector<string> &args )
{
	if ( args.size() < 3 ) {
		cerr << gettext( "drop needs at least 3 arguments" ) << endl;
		err_no = ERR_USER;
		return;
	}

	try {
		if ( args[0] == "resource" ) {
			if ( args.size() > 3 ) {
				if ( args[3] == "only" ) {
					db->ResourceDelete( args[1], args[2] );
					return;
				}
				if ( args[3] != "all" ) {
					cerr << gettext( "unknown drop resource option" ) << " " << args[4] << endl;
					err_no = ERR_USER;
					return;
				}
			}
				// removing resource and all apperances in profiles
			util->DropResource( args[1], args[2] );
		}
		else {
			cerr << gettext( "drop: unknown object" ) << " " << args[0] << endl;
			err_no = ERR_USER;
		}
	}
	catch( SCDB::db_action_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
	catch( SCDBUtil::operation_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

void cmd_dump( vector<string> &args )
{
	try {
		if ( args.empty() )
			db->Dump( cout );
		else
			db->Dump( cout, args[0] );
	}
	catch( SCDB::db_action_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

void cmd_show( vector<string> &args )
{
	vector<string> names;
	vector<string> types;

	if ( args.size() != 1 ) {
		cerr << gettext( "show needs one argument" ) << endl;
		err_no = ERR_USER;
		return;
	}
	
	if ( (args[0] == "unused") ||
			 (args[0] == "unused_resources" ) ) {
		try {
			util->FindUnusedResources( names, types );
			for ( unsigned int i=0; i<names.size(); i++ )
				cout << names[i] << " (" << types[i] << ")" << endl;
		}
		catch( SCDBUtil::operation_failed &e ) {
			cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
			err_no = ERR_SUBSYS;
		}
	}
	else if ( args[0] == "unhandled" ||
						args[0] == "unhandled_resources" ) {
		try {
			util->FindUnhandledResources( names, types );
			for ( unsigned int i=0; i<names.size(); i++ )
				cout << names[i] << " (" << types[i] << ")" << endl;
		}
		catch( SCDBUtil::operation_failed &e ) {
			cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
			err_no = ERR_SUBSYS;
		}
	}
	else if ( args[0] == "deleted" ||
						args[0] == "deleted_resources" ) {
		try {
			util->FindDeletedResources( names, types );
			for ( unsigned int i=0; i<names.size(); i++ )
				cout << names[i] << " (" << types[i] << ")" << endl;
		}
		catch( SCDBUtil::operation_failed &e ) {
			cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
			err_no = ERR_SUBSYS;
		}
	}
	else if ( args[0] == "profiles" ) {
		try {
			vector<string> profiles = db->GetProfiles();
			for ( unsigned int i=0; i<profiles.size(); i++ ) {
				string desc = db->ProfileGetKey( profiles[i], "description" );
				cout << profiles[i] << "	" << desc << endl;
			}
		}
		catch( SCDB::db_action_failed &e ) {
			cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
			err_no = ERR_SUBSYS;
		}
	}
	else if ( args[0] == "resources" ) {
		try {
			vector<string> types = db->ResourceGetTypes();
			for ( unsigned int i=0; i<types.size(); i++ ) {
				vector<string> names = db->ResourceGetNames( types[i] );
				for ( unsigned int j=0; j<names.size(); j++ ) {
					cout << names[j];
					int spaces = 40 - names[j].size();
					if ( spaces<1 ) spaces = 1;
					for ( int k=0; k<spaces; k++ ) cout << " ";
					cout << "	 (" << types[i] << ")" << endl;
				}
			}
		}
		catch( SCDB::db_action_failed &e ) {
			cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
			err_no = ERR_SUBSYS;
		}
	}
	else if ( args[0] == "backups" ) {
		try {
			vector<string> profiles = db->GetProfiles();
			unsigned int longest_profile_name=1;
			for (unsigned int i=0;i<profiles.size();i++) 
				if (profiles[i].size()>longest_profile_name) longest_profile_name=profiles[i].size();
			vector<string> types = db->ResourceGetTypes();
			for ( unsigned int i=0; i<profiles.size(); i++ ) {
				for ( unsigned int j=0; j<types.size(); j++ ) {
					vector<string> names = db->ResourceGetNames( types[j] );
					for ( unsigned int k=0; k<names.size(); k++ ) {
						if (db->FileHasBackup( profiles[i], types[j], names[k] )) {
							cout << "profile " << profiles[i];
							int spaces = longest_profile_name - profiles[i].size();
							for ( int s=0; s<spaces; s++ ) cout << " ";
							cout << ": " << names[k];
							spaces = 30 - names[k].size();
							if ( spaces<1 ) spaces = 1;
							for ( int s=0; s<spaces; s++ ) cout << " ";
							cout << " (" << types[j] << ")" << endl;
						}
					}
				}
			}
		}
		catch( SCDB::db_action_failed &e ) {
			cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
			err_no = ERR_SUBSYS;
		}
	}
	else if (args[0] == "unreferenced" ||
					 args[0] == "unreferneced_files") {
		try {
			vector<string> files;
			util->FindUnreferencedFiles( files );
			for (unsigned int i=0;i<files.size();i++)
				cout << files[i] << "\n";
		}
		catch( SCDBUtil::operation_failed &e ) {
			cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
			err_no = ERR_SUBSYS;
		}
	}		 
	else {
		cerr << gettext( "unknown show option" ) << " " << args[0] << endl;
		err_no = ERR_USER;
	}
}

void cmd_minimize( vector<string> &args )
{
	try {
		util->MinimizeDB();
	}
	catch( SCDBUtil::operation_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

void cmd_help( vector<string> &args )
{
	if ( args.empty() ) {
		cerr << HELP_TEXT_DB_USAGE << endl;
		return;
	}
	if ( args[0] == "help" ) cerr << HELP_TEXT_HELP << endl;
	else if ( args[0] == "load" ) cerr << HELP_TEXT_LOAD << endl;
	else if ( args[0] == "save" ) cerr << HELP_TEXT_SAVE << endl;
	else if ( args[0] == "info" ) cerr << HELP_TEXT_INFO << endl;
	else if ( args[0] == "db_version" ) cerr << HELP_TEXT_DB_VERSION << endl;
	else if ( args[0] == "scpm_version" ) cerr << HELP_TEXT_SCPM_VERSION << endl;
	else if ( args[0] == "quit" ) cerr << HELP_TEXT_QUIT << endl;
	else if ( args[0] == "create" ) cerr << HELP_TEXT_CREATE << endl;
	else if ( args[0] == "drop" ) cerr << HELP_TEXT_DROP << endl;
	else if ( args[0] == "show" ) cerr << HELP_TEXT_SHOW << endl;
	else if ( args[0] == "minimize" ) cerr << HELP_TEXT_MINIMIZE << endl;
	else if ( args[0] == "maximize" ) cerr << HELP_TEXT_MAXIMIZE << endl;
	else if ( args[0] == "customize" ) cerr << HELP_TEXT_CUSTOMIZE << endl;
	else if ( args[0] == "rebuild_deps" ) cerr << HELP_TEXT_REBUILD_DEPS << endl;
	else if ( args[0] == "convert" ) cerr << HELP_TEXT_CONVERT << endl;
	else if ( args[0] == "get" ) cerr << HELP_TEXT_GET << endl;
	else if ( args[0] == "set" ) cerr << HELP_TEXT_SET << endl;
	else if ( args[0] == "dump" ) cerr << HELP_TEXT_DUMP << endl;
	else if ( args[0] == "import" ) cerr << HELP_TEXT_IMPORT << endl;
	else if ( args[0] == "export" ) cerr << HELP_TEXT_EXPORT << endl;
	else cerr << HELP_TEXT_UNKNOWN_CMD << endl;
}

void cmd_maximize( vector<string> &args )
{
	try {
		util->MaximizeDB();
	}
	catch( SCDBUtil::operation_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

void cmd_customize( vector<string> &args )
{
	cerr << gettext("not yet implemented") << endl;
}

void cmd_clean( vector<string> &args )
{
	try {
		vector<string> types;
		vector<string> names;
		util->FindDeletedResources( names, types );
		for( unsigned int i=0; i<names.size(); i++ )
			util->DropResource( names[i], types[i] );
	}
	catch( SCDBUtil::operation_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

void cmd_get( vector<string> &args )
{
	if ( args.size() != 1 ) {
		cerr << gettext( "get needs one argument" ) << endl;
		err_no = ERR_USER;
		return;
	}
	try {
		cout << db->GetValue( args[0] ) << endl;
	}
	catch( SCDB::db_action_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

void cmd_set( vector<string> &args )
{
	if ( args.size() != 2 ) {
		cerr << gettext( "set needs two arguments" ) << endl;
		err_no = ERR_USER;
		return;
	}
	try {
		db->SetValue( args[0], args[1] );
	}
	catch( SCDB::db_action_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

void cmd_rebuild_deps( vector<string> &args )
{
	try {
		util->RebuildDeps();
	}
	catch( SCDBUtil::operation_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

bool convert11( )
{
	try {
			// converting database to version 1.1
		
		vector<string> profiles = db->GetProfiles();
		
			// remove senseless packmagic/promagic entries
		db->DeleteNodes( "root*profiles*profile*promagic" );
		db->DeleteNodes( "root*profiles*profile*packmagic" );
		
			// the script names are now consistent
		db->RenameNodes( "root*profiles*profile*scripts*pre-start", "prestart" );
		db->RenameNodes( "root*profiles*profile*scripts*post-start", "poststart" );
		db->RenameNodes( "root*profiles*profile*scripts*pre-stop", "prestop" );
		db->RenameNodes( "root*profiles*profile*scripts*post-stop", "poststop" );

			// create new global magic node
		db->AddNodes( "root*status*magic" );

			// set new version number
		db->SetSCDBVersion( "1.1" );
	}
	catch( SCDB::db_action_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
		return false;
	}
	return true;
}

bool convert12( )
{
	try {
		vector<string> names = db->ResourceGetNames( "sysconfig" );
		for ( unsigned int i=0; i<names.size(); i++ ) {
				// we have no sysconfig resource anymore
			db->ResourceDelete( names[i], "sysconfig" );
			db->ResourceAdd( "/etc/sysconfig/"+names[i], "file" );
			vector<string> profiles = db->GetProfiles();
			for ( unsigned int j=0; j<profiles.size(); j++ ) {
				if ( db->ProfileHasResource( profiles[j], names[i], "sysconfig" ) ) {
					db->ProfileDeleteResource( profiles[j], names[i], "sysconfig" );
					db->ProfileAddResource( profiles[j], "/etc/sysconfig/"+names[i], "file" );
				}
			}
		}
			// set new version number
		db->SetSCDBVersion( "1.2" );
	}
	catch( SCDB::db_action_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
		return false;
	}
	return true;
}

void cmd_convert( vector<string> &args )
{
	try {
			// converting database to new version
		
		string scdb_version = db->GetSCDBVersion();

		if ( scdb_version != "1.0" && scdb_version != "1.1" ) {
			cerr << gettext("This tool can only convert SCDB Version 1.0 and 1.1 to 1.2") << endl
					 << gettext("Your SCDB has version") << " '" << scdb_version << "'" << endl;
			err_no = ERR_SYSTEM;
			return;
		}

		if ( args.empty() ) {
			if ( scdb_version == "1.2" ) {
				cerr << gettext("Database is already at version 1.2") << endl;
				return;
			}
			cout << gettext("Converting database to version 1.2") << endl;
			if ( scdb_version == "1.0" ) 
				if ( !convert11() ) return;
			if ( !convert12() ) return;
			cout << gettext("Running /lib/scpm/update_storage") << endl;
			system( "/lib/scpm/update_storage" );
			return;
		}
	

		if ( args[0] == "1.1" ) {
			if ( scdb_version == "1.1" ) {
				cerr << gettext("Database is already at version 1.1") << endl;
				err_no = ERR_USER;
				return;
			}
			cout << gettext("Converting database to version 1.1") << endl;
			if ( !convert11() ) return;
		}
		else if ( args[0] == "1.2" ) {
			if ( scdb_version == "1.2" ) {
				cerr << gettext("Database is already at version 1.2") << endl;
				err_no = ERR_USER;
				return;
			}
			cout << gettext("Converting database to version 1.2") << endl;
			if ( scdb_version == "1.0" )
				if ( !convert11() ) return;
			if ( !convert12() ) return;
			cout << gettext("Running /lib/scpm/update_storage") << endl;
			system( "/lib/scpm/update_storage" );
		}
		else {
			cerr << gettext( "Unknown target version number") << " " << args[0] << endl;
			err_no = ERR_USER;
			return;
		}
		
	}
	catch( SCDB::db_action_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

void cmd_export( vector<string> &args )
{
	if (args.size() < 2) {
		cerr << gettext( "export needs at least 2 argument" ) << endl;
		err_no = ERR_USER;
		return;
	}
	try {
		if (args[0] == "profile") {
			if (args.size() == 2)
				util->ExportProfile( args[1], db->GetActiveProfile() );
			else
				util->ExportProfile( args[1], args[2] );
		}
		else {
			cerr << gettext( "Unknown export command ") << " " << args[0] << endl;
			err_no = ERR_USER;
			return;
		}
	}
	catch( SCDB::db_action_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
	catch( SCDBUtil::operation_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}

void cmd_import( vector<string> &args )
{
	if (args.size() < 3) {
		cerr << gettext( "import needs 3 arguments" ) << endl;
		err_no = ERR_USER;
		return;
	}
	try {
		if (args[0] == "profile")
			util->ImportProfile( args[1], args[2] );
		else {
			cerr << gettext( "Unknown import command ") << " " << args[0] << endl;
			err_no = ERR_USER;
			return;
		}
	}
	catch( SCDB::db_action_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
	catch( SCDBUtil::operation_failed &e ) {
		cerr << gettext("Operation failed. Reason:") << " " << e.reason << endl;
		err_no = ERR_SUBSYS;
	}
}
