/*
 * $Id: journal.cc,v 1.3 2008/10/22 10:12:42 hschaa Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	Journal
 * File				 :	journal.cc
 * Description :	provides locking and operation logging
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2004 SuSE Linux AG
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "journal.hh"
#include "log.hh"
#include "scpm_conf.hh"
#include <sys/stat.h>
#include <signal.h>
#include <limits.h>
#include <fstream>
#include <sstream>

#include <iostream>

using namespace std;

#define module_name "journal"

static inline std::string itoa( int i )
{
	ostringstream o;
	o << i;
	return o.str();
}

const char *Journal::hdr="SCPM_JOURNAL_FORMAT_VERSION 1";
const char *Journal::sep="END_DATA_ENTRY";

Journal::Journal()
{
	lock_created = false;
	ready = false;
	done=0;
	struct stat st;
	if (stat("/var/lib/scpm/journal", &st)!=0)
		mkdir("/var/lib/scpm/journal", 0755);
	std::ifstream lockfile( "/var/lib/scpm/journal/lock");
	if (lockfile.good()) {
		int p;
		lockfile >> p;
		if (kill(p,0)==0) {
				// there is already a process running with an exclusive lock
			LOGDEBUG( "database is locked by process "+itoa(p));
			status = STATUS_LOCKED;
		}
		else {
				// lockfile there, but no process
			if (stat("/var/lib/scpm/journal/jobs", &st)==0) {
					// we have a journal
				LOGDEBUG( "dead database lock detected, consider recovery" );
				status = STATUS_RECOVER_NEEDED;
			}
			else {
				LOGDEBUG( "stale database lock detected" );
				status = STATUS_STALE_LOCK;
			}
		}
		lockfile.close();
	}
	else {
			// ok, no lock, but we may have a journal
		struct stat st;
		if (stat("/var/lib/scpm/journal/jobs", &st)==0) {
			LOGDEBUG( "old journal detected, consider recovery" );
			status = STATUS_RECOVER_NEEDED;
		}
		else {
			LOGDEBUG( "database is clean" );
			status = STATUS_CLEAN;
		}
	}
}

Journal::~Journal()
{
	if (lock_created) ::unlink("/var/lib/scpm/journal/lock");
}

bool Journal::MakeLock( )
{
	const char *lockfile = "/var/lib/scpm/journal/lock";
	struct stat statbuf;
	if ( stat( lockfile, &statbuf ) == 0 ) {
		LOGWARN( "SCPM is already locked." );
		return false;
	}
	ofstream lf(lockfile);
	if (lf.bad()) {
		LOGDEBUG( "could not create lock file" );
		return false;
	}
	lf << getpid();
	lf.close();
	lock_created = true;
	LOGDEBUG( "lock created" );
	return true;
}

void Journal::New( )
{
	::unlink( "/var/lib/scpm/journal/jobs" );
	::unlink( "/var/lib/scpm/journal/done" );
	stack_ptr=0;
	entry_stack.clear();
	ready=false;
}

void Journal::Ready( string magic )
{
	stack_ptr=0;
	if (entry_stack.size()==0) {
		LOGWARN( "preparing an empty journal, will not write journal data" );
		return;
	}
	ofstream os( "/var/lib/scpm/journal/jobs" );
	if (os.bad()) LOGERR( "could not write journal data" );
	os << hdr << std::endl
		 << magic << std::endl;
	for (unsigned int i=0; i<entry_stack.size(); i++ ) {
		os << entry_stack[i].command;
		for (unsigned int j=0; j<entry_stack[i].args.size(); j++ ) {
			string s = entry_stack[i].args[j];
				// escape newlines and whitespaces
			while (s.find("\n")!=string::npos)
				s.replace(s.find("\n"), 1, "\\n");
			while (s.find(" ")!=string::npos)
				s.replace(s.find(" "), 1, "\\w");
			os << " " << s;
		}
		os << endl;
	}
	os.close();
	ready=true;
}
		
void Journal::AddEntry( int command, string arg1, string arg2, string arg3, string arg4, string arg5 )
{
	if (stack_ptr == UINT_MAX)
		LOGERR( "journal stack exceeded" );
	entry e;
	e.command = command;
	if (arg1.size()) e.args.push_back(arg1);
	if (arg2.size()) e.args.push_back(arg2);
	if (arg3.size()) e.args.push_back(arg3);
	if (arg4.size()) e.args.push_back(arg4);
	if (arg5.size()) e.args.push_back(arg5);
	e.replay=false;
	entry_stack.push_back(e);
	stack_ptr++;
}

Journal::entry &Journal::NextEntry( bool mark_as_done ) throw( no_more_entries )
{
	if (stack_ptr>=entry_stack.size())
		throw no_more_entries();
	if (mark_as_done && (stack_ptr>=done)) {
			// we need to mark previous (if any) entry as done
		if (stack_ptr>0) this->Done(stack_ptr+1);
	}
	return(entry_stack[stack_ptr++]);
}

Journal::entry &Journal::PrevEntry( bool mark_as_undone ) throw( no_more_entries )
{
	if (stack_ptr==0) {
			// we have no more, but we may need to clear done flag
		if (mark_as_undone) this->Done( 0 );
		throw no_more_entries();
	}
	stack_ptr--;
	if (mark_as_undone) {
			// we need to push back dne mark
		if (stack_ptr>0) this->Done(stack_ptr);
	}
	return(entry_stack[stack_ptr]);
}

void Journal::Done( unsigned int mark )
{
	ofstream of( "/var/lib/scpm/journal/done", ios::trunc );
	of << mark;
	done = mark;
	of.close();
}

void Journal::Load( string magic ) throw( load_failed )
{
	ifstream in( "/var/lib/scpm/journal/jobs" );
	if (in.bad()) throw load_failed("could no open journal data file");

	ifstream in2( "/var/lib/scpm/journal/done" );
	if (in2.bad()) {
		LOGWARN( "done file missing or inaccessible, assuming nothing was done" );
		done=0;
	}
	else {
		in2 >> done;
		in2.close();
		LOGDEBUG( "journal done pointer is "+itoa(done) );
	}

		// read header
	string s;
	getline( in, s );
	if ( s != hdr ) {
		in.close();
		LOGDEBUG( "jobs header corrupt or journal incompatible");
		throw load_failed( "read error while loading journal data" );
	}
	getline( in, s );
	if ( s != magic ) {
		in.close();
		LOGDEBUG( "magic differs (saved: "+s+" current: "+magic+")" );
		throw load_failed( "wrong magic" );
	}
	while(in.good()) {
		entry e;
		in >> e.command;
		LOGDEBUG( "journal: read command "+itoa(e.command) );
		while(in.peek()!='\n' && in.good()) {
			string s;
			in >> s;
				// restore newlines and whitespaces
			while (s.find("\\n")!=string::npos)
				s.replace(s.find("\\n"), 2, "\n");
			while (s.find("\\w")!=string::npos)
				s.replace(s.find("\\w"), 2, " ");
			LOGDEBUG( "journal:			arg: "+s );
			e.args.push_back( s );
		}
		in.get(); // remove newline
		in.peek(); // may set eof bit
		if (done>entry_stack.size()+1)
			e.replay=true;
		else
			e.replay=false;
		entry_stack.push_back(e);
	}
	in.close();
	stack_ptr=0;
	LOGDEBUG( "journal loaded ("+itoa(entry_stack.size())+" entries)" );
}

void Journal::Delete( ) 
{
	::unlink( "/var/lib/scpm/journal/jobs" );
	::unlink( "/var/lib/scpm/journal/done" );
}

void Journal::Unlock( )
{
	::unlink( "/var/lib/scpm/journal/lock" );
}
