/*
 * $Id: log.cc,v 1.1.1.1 2005/05/11 09:03:12 gleissner Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	Log
 * File				 :	log.cc
 * Description :	implements error reporting facilities
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2001 SuSE GmbH
 *					 2002 SuSE Linux AG
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "log.hh"
#include "scpm_conf.hh"
#include <syslog.h>
#include <iostream>
#include <fstream>
#include <string.h>
#include <cstdlib>
#include <cstdio>
#include <time.h>
#include <clocale>
#include <libintl.h>

using namespace std;

// init const stuff
const Log::endlog_t Log::endlog = endlog_t();
const Log::debug_t Log::debug = debug_t();
const Log::info_t Log::info = info_t();
const Log::warning_t Log::warning = warning_t();
const Log::error_t Log::error = error_t();

// set default values
string Log::last_error="none";
int Log::log_type=Log::TYPE_VERBOSE;
int Log::log_mode=Log::MODE_NONE;
int Log::exit_mode=Log::EXIT_ERROR;

char *scpm_error;

stringstream Log::buf;
int Log::current_class;

Log::Log()
{
	scpm_error = strdup( "none" );
	SCPM_conf *conf = SCPM_conf::GetHandle();
	if ( conf->Debug() ) log_type = Log::TYPE_DEBUG;
	if ( conf->LogToSyslog() ) log_mode |= Log::MODE_SYSLOG;
	if ( conf->LogToStderr() ) log_mode |= Log::MODE_STDERR;
	if ( conf->LogToFile() )	 log_mode |= Log::MODE_FILE;
	if ( conf->ExitOnWarning() ) exit_mode = Log::EXIT_WARNING;
	aborted = false;
}

Log::~Log()
{
	free( scpm_error );
}

void Log::SetLog( logtype_t type , logmode_t mode )
{
	log_type = type;
	log_mode = mode;
}

void Log::WriteMessage( string module_name, logclass_t log_class, string message, 
												string action )
{
	int syslog_level;
	SCPM_conf *conf = SCPM_conf::GetHandle();

	if ( log_type < log_class ) return;

	switch (log_class) {
		case CLASS_ERROR: syslog_level=LOG_ERR; break;
		case CLASS_WARNING: syslog_level=LOG_WARNING; break;
		case CLASS_INFO: syslog_level=LOG_NOTICE; break;
		case CLASS_DEBUG: syslog_level=LOG_INFO; break;
		default:
			cerr << "scpm: Unkown log class " << log_class << endl;
			return;
	}

	if ( log_mode == MODE_SYSLOG ) {
		openlog( "scpm", 0, LOG_USER );
		syslog( syslog_level, untrans(message).c_str() );
	}
	if ( log_mode == MODE_STDERR ) {
		cerr << "scpm (" << module_name << ") " ;
		switch (log_class) {
			case CLASS_ERROR:		cerr << "ERROR " ; break;
			case CLASS_WARNING: cerr << "warning " ; break;
			case CLASS_INFO:		cerr << "info " ; break;
			case CLASS_DEBUG:		cerr << "debug " ; break;
		}
		cerr << untrans(message) << endl;
	}

	if ( log_mode == MODE_FILE ) {
		ofstream log_file( conf->GetLogfile().c_str(), ios::app );
		const time_t date = time( NULL );
		char *timestr = ctime( &date );
		timestr[ strlen(timestr)-1 ]=' ';
		log_file << timestr << "scpm (" << module_name << ") " ;
		switch (log_class) {
			case CLASS_ERROR:		log_file << "ERROR " ; break;
			case CLASS_WARNING: log_file << "warning " ; break;
			case CLASS_INFO:		log_file << "info " ; break;
			case CLASS_DEBUG:		log_file << "debug " ; break;
		}
		log_file << untrans(message) << endl;
		log_file.close();
	}

		// Save error message
	last_error = message;

		// maybe we have to terminate here
	if ( ( log_class == CLASS_ERROR ) || 
			 ( (log_class == CLASS_WARNING) && (exit_mode == EXIT_WARNING) )) {
		free( scpm_error );
		scpm_error = strdup( trans(message).c_str()	 );
		aborted = true;
		throw scpm_panic();
	}
}

void Log::WriteRawMessage( logclass_t log_class, string message, string action )
{
	int syslog_level;
	SCPM_conf *conf = SCPM_conf::GetHandle();

	if ( log_type < log_class ) return;

	switch (log_class) {
		case CLASS_ERROR: syslog_level=LOG_ERR; break;
		case CLASS_WARNING: syslog_level=LOG_WARNING; break;
		case CLASS_INFO: syslog_level=LOG_NOTICE; break;
		case CLASS_DEBUG: syslog_level=LOG_INFO; break;
		default:
			cerr << "scpm: Unkown log class " << log_class << endl;
			return;
	}

	if ( log_mode == MODE_SYSLOG ) {
		openlog( "scpm", 0, LOG_USER );
		syslog( syslog_level, message.c_str() );
	}
	if ( log_mode == MODE_STDERR ) {
		cerr << "scpm " << message << endl;
	}

	if ( log_mode == MODE_FILE ) {
		ofstream log_file( conf->GetLogfile().c_str(), ios::app );
		const time_t date = time( NULL );
		char *timestr = ctime( &date );
		timestr[ strlen(timestr)-1 ]=' ';
		log_file << timestr << "scpm " << message << endl;
		log_file.close();
	}
}

string Log::GetLastMessage()
{
	return last_error;
}

Log &Log::operator<<( string s)
{
	buf << s;
	return *this;
}

Log &Log::operator<<( char c )
{
	buf << c;
	return *this;
}

Log &Log::operator<<( long i )
{
	buf << i;
	return *this;
}

Log &Log::operator<<( unsigned long i )
{
	buf << i;
	return *this;
}

Log &Log::operator<<( endlog_t &a )
{
	char tmp[255];
	while ( buf.getline( tmp, 254 ) ) {
		this->WriteMessage( "test", current_class, tmp );
	}
	return *this;
}

Log &Log::operator<<( debug_t &a )
{
	current_class = CLASS_DEBUG;
	return *this;
}

Log &Log::operator<<( info_t &a )
{
	current_class = CLASS_INFO;
	return *this;
}

Log &Log::operator<<( warning_t &a )
{
	current_class = CLASS_WARNING;
	return *this;
}

Log &Log::operator<<( error_t &a )
{
	current_class = CLASS_ERROR;
	return *this;
}

string Log::untrans( string raw )
{
	bool escaped=false;
	string s;
	for ( unsigned int i=0; i<raw.size(); i++ ) {
		if ( (raw[i] != '&') || ( (raw[i] == '&') && escaped) )
			s += raw[i];
	}
	return s;
}

string Log::trans( string raw )
{
	if ( raw.find( "&" ) == string::npos ) {
			// has no variables in it, translating straight forward
		return dgettext( "scpm", raw.c_str() );
	}

		// prepare string
	string prepared;
	vector<string> variables;
	int argno=0;
	for ( unsigned int i=0; i<raw.size(); i++ ) {
		if ( raw[i]=='&' && i<raw.size()-1 && raw[i+1]!='&' ) {
				// found a variable
			string varname;
			i++;
			while( i<raw.size() && raw[i]!=' ' && raw[i]!=')' )
				varname+=raw[i++];
			argno++;
			if ( argno == 10 ) {
				this->WriteMessage( "log", Log::CLASS_WARNING, 
														"only up to 9 variables on translating are supported" );
				return raw;
			}
			char t[6];
			if ( i<raw.size() )
				sprintf( t, "VAR%i%c", argno, raw[i] );
			else
				sprintf( t, "VAR%i", argno );
			prepared+=t;
			variables.push_back( varname );
		}
		else
			prepared+=raw[i];
	}
	this->WriteMessage( "log", Log::CLASS_DEBUG, "prepared string: '"+prepared+"'" );
	
		// translate string
	string translated = dgettext( "scpm", prepared.c_str() );
	this->WriteMessage( "log", Log::CLASS_DEBUG, "translated string: '"+translated+"'" );

		// transform variables back
	string final;
	for ( unsigned int i=0; i<translated.size(); i++ ) {
		if ( translated[i]=='V' && i<translated.size()-3 && translated[i+1]=='A' && 
				 translated[i+2]=='R' ) {
			int argno = static_cast<int>(translated[i+3] - '0' );
			if ( argno < 1 || argno > 9 ) {
					// false alert
				final+=translated[i];
			}
			else {
				if ( argno-1 > static_cast<int>(variables.size()) ) {
					this->WriteMessage( "log", Log::CLASS_WARNING, "translator bug for string "+prepared );
					return raw;
				}
				final+=variables[argno-1];
				i+=3;
			}
		}
		else
			final+=translated[i];
	}
	this->WriteMessage( "log", Log::CLASS_DEBUG, "final string: '"+final+"'" );
	return final;
}
