/*
 * $Id: profile.cc,v 1.2 2005/05/23 13:26:02 gleissner Exp $
 *
 *
 * Project     :  SCPM (System Configuration Profile Management)
 * Module      :  Profile
 * File        :  profile.cc
 * Description :  implements profile specific highlevel actions
 * Author      :  Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2001 SuSE GmbH
 *           2002-2004 SuSE Linux AG
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "profile.hh"
#include "scdb.hh"
#include "log.hh"
#include "resource_helpers.hh"
#include "scpm_helpers.hh"
#include "progress.hh"
#include "storage.hh"
#include "executor.hh"
#include "modlib.hh"
#include "scpm_conf.hh"
#include "journal.hh"
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <sstream>

using namespace std;

Profile::Profile( string profile_name )
		: module_name( "profile" )
{
  name = profile_name;
	db	 = SCDB::GetHandle();
}

Profile::~Profile( )
{
	for (unsigned int i=0; i<resources_to_start.size(); i++) delete resources_to_start[i];
	for (unsigned int i=0; i<resources_to_stop.size(); i++) delete resources_to_stop[i];
	for (unsigned int i=0; i<resources_keep_running.size(); i++) delete resources_keep_running[i];
}

bool Profile::IsClean( )
{
	try {
		return db->ProfileGetFlag( name, "clean" );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "could not query clean flag for profile "+name+
						 " (database error "+e.reason+")" );
		LOGWARN( "assuming profile is not clean" );
		return false;
	}
}

void Profile::Remove( )
{
	SCDB *db = SCDB::GetHandle();
	try {
		vector<string> types = db->ProfileGetResourceTypes( name );
		for (unsigned int i=0; i<types.size(); i++) {
			vector<string> names = db->ProfileGetResources( name, types[i] );
			for (unsigned int j=0; j<names.size(); j++) {
				jrnl->AddEntry( Journal::cmd_file_delete, name, types[i], names[j] );
			}
		}
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGWARN( "could not mark resources of prfile &"+name+" for deletion" );
	}
}

void Profile::Restore( )
{
	Resource_helpers rh;
	SCDB *db = SCDB::GetHandle();
	try {
		vector<string> resource_types = db->ResourceGetTypes();
		for (unsigned int i=0; i<resource_types.size(); i++) {
			vector<string> resource_names = this->GetResources( resource_types[i] );
			for (unsigned int j=0; j<resource_names.size(); j++) {
				LOGDEBUG( "checking resource "+resource_names[j]+" ("+resource_types[i]+")" );
				Resource *r = rh.CreateObject( resource_types[i],
																			 resource_names[j] );
				r->SetProfile( name );
				if ( !r->UpToDate() ) {
					LOGDEBUG( "resource "+resource_names[j]+" ("+resource_types[i]+
										") needs to be restored" );
					jrnl->AddEntry( Journal::cmd_resource_restore, resource_types[i],
													resource_names[j], name );
				}
				delete r;
			}
		}
		return;
	}
	catch( Resource_helpers::type_invalid ) {
		LOGWARN( "used invalid resource type, this should not happen" );
	}
	catch( Resource::not_available ) {
		LOGWARN( "called missing resource handler, this is a bug" );
	}
	LOGERR( "could not restore profile &"+name );
}

vector<string> Profile::GetResources( string resource_type )
{
	static vector<string> res;
	try {
		res = db->ProfileGetResources( name, resource_type );
		if ( res.empty() ) {
			LOGWARN( "profile "+name+" has no "+resource_type+" resources" );
		}
		return res;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "could not determine resources for profile &"+name );
	return res;
}

vector<string> Profile::GetModifiedResources( string resource_type )
{
	Resource_helpers rh;
	Progress *pr = Progress::GetHandle();
	vector<string> resources;
	vector<string> resource_names = this->GetResources( resource_type );
	pr->SetSegmentSteps( resource_names.size() );
	try {
		for (unsigned int i=0; i<resource_names.size(); i++) {
			try {
				pr->BarStep();
				Resource *r = rh.CreateObject( resource_type, resource_names[i] );
				r->SetProfile( name );
				if ( r->Modified() ) {
					resources.push_back( resource_names[i] );
					Progress::GetHandle()->Newline( Progress::verbose );
					Progress::GetHandle()->Print( "  &"+resource_type+" &"+resource_names[i]+" is modified",
																				Progress::verbose );
				}
			}
			catch( Resource_helpers::type_invalid ) {
				LOGWARN( "used invalid resource type, this should not happen" );
				LOGERR( "could not determine modified resources for profile &"+name );
			}
			catch( SCDB::db_action_failed &e ) {
				LOGWARN( "scdb error ("+e.reason+")" );
				LOGERR( "could not determine modified resources for profile &"+name );
			}
		}
	}
	catch( Resource::not_available ) {
		LOGDEBUG( "called resource action not available for type ("+resource_type+")" );
	}
	return resources;
}

void Profile::ScriptRunPreStart ( )
{
	this->RunScript( "prestart" );
}

void Profile::ScriptRunPostStart( )
{
	this->RunScript( "poststart" );
}

void Profile::ScriptRunPreStop	( )
{
	this->RunScript( "prestop" );
}

void Profile::ScriptRunPostStop ( )
{
	this->RunScript( "poststop" );
}

void Profile::ScriptSetPreStart( istream &input )
{
	this->SaveScript( "prestart", input );
}

void Profile::ScriptSetPostStart( istream &input )
{
	this->SaveScript( "poststart", input );
}

void Profile::ScriptSetPreStop( istream &input )
{
	this->SaveScript( "prestop", input );
}

void Profile::ScriptSetPostStop( istream &input )
{
	this->SaveScript( "poststop", input );
}

void Profile::ScriptGetPreStart( ostream &output )
{
	this->DumpScript( "prestart", output);
}

void Profile::ScriptGetPostStart( ostream &output )
{
	this->DumpScript( "prestart", output);
}

void Profile::ScriptGetPreStop( ostream &output )
{
	this->DumpScript( "prestart", output);
}

void Profile::ScriptGetPostStop( ostream &output )
{
	this->DumpScript( "prestart", output);
}

void Profile::ScriptUnsetPreStart( )
{
	this->RemoveScript( "prestart" );
}

void Profile::ScriptUnsetPostStart( )
{
	this->RemoveScript( "poststart" );
}

void Profile::ScriptUnsetPreStop( )
{
	this->RemoveScript( "prestop" );
}

void Profile::ScriptUnsetPostStop( )
{
	this->RemoveScript( "poststop" );
}

void Profile::SaveScript( string scriptname, istream &input )
{
		// this is not supported anymore
	LOGERR( "BUG! Obosolete function called" );
	return;
}

void Profile::DumpScript( string scriptname, ostream &output )
{
	LOGERR( "BUG! Obosolete function called" );
	return;
}

void Profile::RemoveScript( string scriptname )
{
	try {
		db->ProfileSetScript( name, scriptname, "" );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGERR( "could not remove &"+scriptname+" script of profile &"+name );
	}
}

void Profile::RunScript( string scriptname )
{
	string script_location;
	try {
		script_location = db->ProfileGetScript( name, scriptname );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGERR( "could not run &"+scriptname+" script of profile &"+name );
		return;
	}

	if ( script_location == "yes" ) {
		LOGERR( "you have scripts in scpm 0.6.x style defined, read documentation" );
		return;
	}
	if ( (script_location.empty()) || (script_location == "no") ) return;

	Progress *p = Progress::GetHandle();
	vector<string> out;
	vector<string> err;

	p->Print( "Running &"+scriptname+" script of profile &"+name );

	try {
		Executor e( SCPM_conf::GetHandle()->GetLibDir()+"/run_script "+script_location );
		if ( e.Run( out, err ) != 0 ) {
			LOGDEBUG( "output of "+script_location );
			for ( unsigned int i=0; i<err.size(); i++ ) 
				LOGDEBUG( err[i] );
			p->Print( "failed" );
			p->Newline();
			if ( err.size() ) {
				for ( unsigned int i=1; i<err.size(); i++ ) {
					p->Print( " "+err[i] );
					p->Newline();
				}
			}
			LOGWARN( "bad exit status of run_script "+script_location );
			return;
		}
	}
	catch( Executor::exec_failed ) {
		LOGERR( "could not execute &"+scriptname+" script of profile &"+name );
		return;
	}

	p->Print( "done" );
	p->Newline();

		// analyse script logs
	if ( err.size() == 0 ) {
		LOGWARN( "could not locate script logs, this should not happen" );
		return;
	}
	string tmpdir = err[0];
	struct stat statbuf;
	if ( ::stat( (tmpdir+"/messages").c_str(), &statbuf ) == 0 ) {
			// we have to log messages
		ifstream logs( (tmpdir+"/messages").c_str() );
		string log_class;
		Log *log = Log::GetHandle();
		while ( logs >> log_class ) {
			Log::logclass_t lc;
			if ( log_class == "debug" ) lc = Log::CLASS_DEBUG;
			else if ( log_class == "info" ) lc = Log::CLASS_INFO;
			else if ( log_class == "warning" ) lc = Log::CLASS_WARNING;
			else if ( log_class == "ERROR" ) lc = Log::CLASS_ERROR;
			char buf[200];
			logs >> ws;
			logs.getline( buf, 199 );
			log->WriteRawMessage( lc, buf );
		}
		logs.close();
	}

	if ( ::stat( (tmpdir+"/exitcode").c_str(), &statbuf ) == 0 ) {
		ifstream exits( (tmpdir+"/exitcode").c_str() );
		string buf;
		exits >>	buf;
		if ( buf != "0" ) LOGWARN( "script exited with code "+buf );
		exits.close();
	}

	if ( out.size() ) {
		p->Print( "Output was" );
		p->Newline();
		for ( unsigned int i=1; i<out.size(); i++ ) { 
			p->Print( out[i] );
			p->Newline();
		}
	}

	bool stop_op=false;
	if ( ::stat( (tmpdir+"/aborted").c_str(), &statbuf ) ==0 ) {
			// uh, the script says STOP
		LOGWARN( scriptname+" script aborted SCPM operation" );
		LOGWARN( "this may result in an inconsistent system (read documenation)" );
		stop_op = true;
	}

	DIR *dp;
	struct dirent *ep;
	dp = ::opendir( tmpdir.c_str() );
	if ( dp ) {
		while ( (ep = readdir(dp)) ) 
			::unlink( (tmpdir+"/"+ep->d_name).c_str() );
		::closedir( dp );
	}
	::rmdir( tmpdir.c_str() );
	
	if (stop_op) LOGERR( "SCPM aborted by &"+script_location );
}

bool Profile::HasScript( string scriptname )
{
	SCDB *scdb = SCDB::GetHandle();
	try {
		if (scdb->ProfileGetScript(name,scriptname)=="") 
			return false;
		else return true;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "could not determine &"+scriptname+" script of profile &"+name );
	return false;
}
