/*
 * $Id: scpm.cc,v 1.12 2006/11/02 16:15:01 gleissner Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	SCPM main engine
 * File				 :	scpm.cc
 * Description :	implements the execution logic of SCPM
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2001 SuSE GmbH
 *					 2002-2004 SuSE Linux AG
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "scpm.hh"
#include "log.hh"
#include "scpm_conf.hh"
#include "scpm_helpers.hh"
#include "progress.hh"
#include "resource.hh"
#include "journal.hh"
#include <scdb_util.hh>
#include <tokenizer.h>
#include <fstream>
#include <sys/types.h>
#include <dirent.h>
#include <libintl.h>
#include <clocale>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

using namespace std;

#define _(x) dgettext ("scpm", x)

#define module_name "scpm"

Log *Log::log_handle = 0;
SCDB *SCDB::scdb_handle = 0;
Progress *Progress::progress = 0;
SCPM_conf *SCPM_conf::handle = 0;
Journal *Journal::journal_handle = 0;

SCPM::SCPM( int options, ostream &info_out, ostream &hash_out, string root )
{
	try {
		db_open					= true;
		switch_prepared = false;
		init_failed			= true;

		setlocale( LC_ALL, "" );
		bindtextdomain( "scpm", NULL );
			//textdomain( "scpm" );

		if (!root.empty()) {
			if ( chroot( root.c_str() ) == -1 ) {
				cerr << "Could not chroot to " << root << " (" << strerror( errno )
						 << " " << endl;
				db_open = false;
				return;
			}
		}

			// create static objects
			// it is important to init SCPM_conf first
		SCPM_conf::handle = new SCPM_conf( );
		SCPM_conf *conf = SCPM_conf::GetHandle();
		conf->SetOptions( options );
	
		Log::log_handle = new Log();
		SCDB::scdb_handle = new SCDB();
		Progress::progress = new Progress( info_out, hash_out );
		Journal::journal_handle = new Journal();

		Progress *p = Progress::GetHandle();
		p->SetOutputMode( ((conf->Quiet()) ? Progress::quiet : 0) +
											((conf->ProgressBar()) ? Progress::hash_marks : 0) +
											((conf->Verbose()) ? Progress::verbose : 0) );

		sh = new SCPM_helpers();
		log = Log::GetHandle();

		init_failed=false;
		db_open=false;
			// load configuration db from disk
		if ( !conf->SkipLoad() ) {
			try {
				int res = sh->LoadSCDB( );
				switch (res) {
					case SCDB::db_state_ok:
						db_open=true;
						LOGDEBUG( "database load successful" );
						break;
					case SCDB::db_state_open_failed:
						LOGERR( "could not open database" );
						break;
					case SCDB::db_state_corrupt:
						init_failed = true;
						LOGERR( "database corrupt" );
						break;
					case SCDB::db_state_version_unsupported:
						if ( conf->ForceDB() ) {
								// ok, the user wants it this way
							LOGDEBUG( "forced using old/unsupported database version" );
						}
						init_failed = true;
						LOGERR( "You have an old SCDB version. Read man page how to update it." );
						break;
					default:
						LOGERR( "unknown database state. This is a bug." );
						init_failed = true;
						break;
				}
				if ( sh->InstallationChanged() ) {
					init_failed = true;
					LOGERR( "Your operating system installation changed or is unknown. Read the man page.");
				}
			}
			catch( SCPM_helpers::db_corrupt ) {
				init_failed = true;
				db_open = false;
			}
			catch( SCPM_helpers::db_open_failed ) {
				db_open = false;
			}
		}
		if (!db_open) LOGERR( "could not load scdb" );

		Journal::status_t st = jrnl->GetStatus();
		switch (st) {
			case Journal::STATUS_CLEAN:
				LOGDEBUG( "journal is clean" );
				break;
			case Journal::STATUS_LOCKED:
				init_failed=true;
				LOGERR( "SCPM is locked by another process (read man page)" );
				break;
			case Journal::STATUS_STALE_LOCK:
				LOGWARN( "SCPM is locked by a dead process, lock ignored" );
				break;
			case Journal::STATUS_RECOVER_NEEDED:
				init_failed=true;
				LOGERR( "SCPM is locked and needs recover (read man page)" );
				break;
			default:
				LOGERR( "unknown journal state. This is a bug!" );
		}
		LOGDEBUG( "scpm main initialized" );
	}
	catch( ... ) {
	}
}

SCPM::~SCPM()
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	if (!conf->ReadOnly()) {
		if ( (!init_failed || conf->ForceDB()) && db_open && !conf->SkipLoad() ) {
			sh->SaveSCDB( );
		}
	}
	delete Journal::journal_handle;Journal::journal_handle=NULL;
	delete SCPM_conf::handle;SCPM_conf::handle=NULL;
	delete SCDB::scdb_handle;SCDB::scdb_handle=NULL;
	delete Progress::progress;Progress::progress=NULL;
	delete Log::log_handle; Log::log_handle=NULL;
	delete sh; sh=NULL;
}

void SCPM::Prepare( bool write_access )
{
	if (init_failed) throw Log::scpm_panic();
	if (!db_open) throw Log::scpm_panic();
	if (write_access) {
		if ( !sh->IsEnabled() ) LOGERR( "SCPM disabled" );
		if (jrnl->GetStatus() == Journal::STATUS_STALE_LOCK) {
			LOGDEBUG( "stale lock detected, removing" );
			jrnl->Unlock();
		}
		jrnl->MakeLock();
	}
}

bool SCPM::Active( string &profile )
{
	try {
		Prepare();
		if ( !sh->IsEnabled() ) LOGERR( "SCPM disabled" );
		profile = sh->GetActiveProfile();
	}
	catch( ... ) {
		return false;
	}
	return true;
}

bool SCPM::List( vector<string> &profiles )
{
	try {
		Prepare();
		profiles = sh->GetAllProfiles();
	}
	catch( ... ) {
		return false;
	}
	return true;
}

bool SCPM::Enable( bool force )
{
	Progress *pr = Progress::GetHandle();
	try {
		SCPM_conf *conf = SCPM_conf::GetHandle();
		if ( !force ) force = conf->Force();
		if ( init_failed && !force ) return false;
			// maybe a former operation set error state because scpm is disabled
			// we are going to change that, so clear abort state
		Log::GetHandle()->ClearAborted();
		jrnl->MakeLock();
		if ( db_open ) {
			if ( !force ) {
				if ( sh->IsEnabled() ) {
					LOGERR( "SCPM already active" );
					return false;
				}
				else {
					if ( sh->IsInitialized() ) {
						jrnl->New();
						jrnl->AddEntry(Journal::cmd_set_global_key, "enabled", "yes" );
						jrnl->AddEntry(Journal::cmd_write_boot_info);
						jrnl->Ready( "nomagic" );
						pr->SetBar(100);
						sh->ProcessJournal();
						return true;
					}
					LOGERR( "SCPM is enabled but uninitialized" );
					return false;
				}
			}
			else {
				jrnl->New();
					// force initialization, drop all old data
				pr->Print( "Enabling forced. Old database will be dropped." );
				SCDB *db = SCDB::GetHandle();
				db->DropDB();
				db_open = false;
				sh->stay_locked = false;
				pr->Newline();
					// we also remove old profile data
				jrnl->AddEntry( Journal::cmd_clear_all );
			}
		}
		Log::GetHandle()->ClearAborted();

		pr->Print( "Analyzing system" );

			// database is not open, so we have to do init
		init_failed = false;
		db_open = true;

		string exec_string;
		if ( force ) 
			exec_string = conf->GetLibDir()+"/prepare -f";
		else
			exec_string = conf->GetLibDir()+"/prepare";
		jrnl->AddEntry( Journal::cmd_message, "Preparing system" );
		jrnl->AddEntry( Journal::cmd_call, exec_string );
		jrnl->AddEntry( Journal::cmd_reload_db );

		jrnl->AddEntry( Journal::cmd_message, "Creating resource database" );
		sh->CreateResources();
		jrnl->AddEntry( Journal::cmd_profile_add, "default" );
		jrnl->AddEntry( Journal::cmd_set_global_key, "initialized", "yes" );
		jrnl->AddEntry( Journal::cmd_set_global_key, "enabled", "yes" );

		jrnl->AddEntry( Journal::cmd_message, "Creating default profile" );
		sh->AddProfileResources( "default" );
		jrnl->AddEntry( Journal::cmd_profile_set_key, "default", "clean", "no" );
		jrnl->AddEntry( Journal::cmd_profile_set_key, "default", "active", "yes" );

		jrnl->AddEntry( Journal::cmd_set_global_key, "active-profile", "default" );
		jrnl->AddEntry( Journal::cmd_save_install_info );
		jrnl->AddEntry( Journal::cmd_write_boot_info );

			// do the work
		pr->SetBar( 100 );
		pr->SetSegment( 100 );
		jrnl->Ready( "nomagic" );
		sh->ProcessJournal();
		pr->Newline();

		LOGINFO( "SCPM enabled" );
	}
	catch( ... ) {
			// we need this for the yast2 frontend
		pr->ResetBar();
		return false;
	}
	return true;
}

bool SCPM::Disable( )
{
	try {
		Prepare( true );

			// do disabling stuff
		sh->ClearBootInfo();
		sh->SetDisabled();

		LOGINFO( "SCPM disabled" );
	}
	catch( ... ) {
		return false;
	}
	return true;
}

bool SCPM::Create( string profile, string template_profile )
{
	Progress *pr = Progress::GetHandle();
	try {
		Prepare( true );

		if ( ! sh->ProfileExists( template_profile ) ) {
			LOGERR( "Profile &" + template_profile + " does not exist" );
			return false;
		}

		if ( sh->ProfileExists( profile ) ) {
			LOGERR( "Profile &" + profile + " already exists" );
			return false;
		}

		jrnl->New();
		pr->Print( "Copying profile" );
		sh->CloneProfile( profile, template_profile );
		jrnl->AddEntry( Journal::cmd_write_boot_info );

		jrnl->Ready( "nomagic" );
		pr->SetBar(100);
		sh->ProcessJournal();
		pr->Newline();

		LOGINFO( "profile "+profile+" created" );
	}
	catch( ... ) {
			// ensure full bar for the yast2 frontend
		pr->ResetBar();
		return false;
	}
	return true;
}

bool SCPM::Delete( string profile )
{
	try {
		Prepare(true);

		if ( ! sh->ProfileExists( profile ) ) {
			LOGERR( "Profile &"+profile+" does not exist" );
			return false;
		}		 

		if ( profile == sh->GetActiveProfile() ) {
			LOGERR( "active profile cannot be deleted" );
			return false;
		}		 

		jrnl->New();
		
		Profile p( profile );
		p.Remove();
		jrnl->AddEntry( Journal::cmd_profile_delete, profile );
		jrnl->AddEntry( Journal::cmd_write_boot_info );
		jrnl->Ready( "nomagic" );
		Progress::GetHandle()->SetBar(100);
		sh->ProcessJournal();
		
		LOGINFO( "profile "+profile+" deleted" );
	}
	catch( ... ) {
		return false;
	}
	return true;
}

bool SCPM::Rename( string profile, string newprofile )
{
	try {
		Prepare(true);

		if ( ! sh->ProfileExists( profile ) ) {
			LOGERR( "Profile &" + profile + " does not exist" );
			return false;
		}		 
		
		jrnl->New();

		if ( sh->ProfileExists( newprofile ) ) {
			SCPM_conf *conf = SCPM_conf::GetHandle();
			if ( conf->Force() ) {
				LOGINFO( "renaming forced, overwriting old profile "+newprofile );
				jrnl->AddEntry( Journal::cmd_profile_delete, newprofile );
			}
			else {
				LOGERR( "Profile &"+newprofile+" already exists" );
				return false;
			}
		}		 

		jrnl->AddEntry( Journal::cmd_message, "Copying profile data" );
		sh->CloneProfile( newprofile, profile );
			//Profile p( newprofile );
			//p.Save();

		if ( profile == sh->GetActiveProfile() ) {
				// renamed profile was the active one
			jrnl->AddEntry( Journal::cmd_set_global_key, "active-profile", newprofile );
		}
		else
			jrnl->AddEntry( Journal::cmd_profile_set_key, newprofile, "clean", "yes" );

		jrnl->AddEntry( Journal::cmd_message, "Removing profile &"+profile );
		jrnl->AddEntry( Journal::cmd_profile_delete, profile );
		jrnl->AddEntry( Journal::cmd_write_boot_info );

		jrnl->Ready( "nomagic" );
		Progress::GetHandle()->SetBar(100);
		Progress::GetHandle()->Print( "Creating profile &"+newprofile );
		sh->ProcessJournal();
		Progress::GetHandle()->Newline();
		
		LOGINFO( "profile "+profile+" renamed to "+newprofile );
	}
	catch( ... ) {
		return false;
	}
	return true;
}

bool SCPM::PrepareSwitch( string profile, switch_info_t &switch_info )
{
	Progress *p = Progress::GetHandle();
	try {
		Prepare(true);

		if ( !sh->ProfileExists( profile ) ) {
			LOGERR( "Profile &"+profile+" does not exist" );
			return false;
		}

		string active = sh->GetActiveProfile();

		if ( active.find( "(none)" ) != string::npos ) {
			LOGERR( "could not determine active profile" );
			return false;
		}


			// check for new resources here

		switch_info.profile_name = profile;

		Profile active_profile( active );

		Resource_helpers rh;
		vector<string> types = rh.GetResourceTypes();

		p->SetBar( 100 );
		p->Print( "Checking for modified resources" );
		switch_info.profile_modified = false;
		switch_info.modified_resources.clear();

		int bsize=99;
		if ( sh->SystemChanged() ) bsize=50;

		for (unsigned int i=0; i<types.size(); i++) {
			p->SetSegment( bsize/types.size() );
			vector<string> names = active_profile.GetModifiedResources( types[i] );
			for (unsigned int j=0; j<names.size(); j++) {
				resource_info_t ri;
				ri.resource_name=names[j];
				ri.resource_type=types[i];
				ri.is_new = false;
				ri.is_deleted = false;
				if ( SCPM_conf::GetHandle()->Skip() )
					ri.save = false;
				else
					ri.save = true;
				ri.save_mode = normal;
				ri.groups = this->GetGroupsByResource( names[j], types[i] );
				switch_info.modified_resources.push_back( ri );
				switch_info.profile_modified = true;
			}
		}

		if ( sh->SystemChanged() ) {
			p->SetSegment( 49 );
			SCDB *db = SCDB::GetHandle();
			p->Newline();
			p->Print( "Checking for new/obsolete resources" );
			int count=0;
			for ( unsigned int i=0; i<types.size(); i++ ) {
				vector<string> bla = rh.GetResourceNames( types[i] );
				count+=bla.size();
			}
			p->SetSegmentSteps( count );
			for ( unsigned int i=0; i<types.size(); i++ ) {
				vector<string> names = rh.GetResourceNames( types[i] );
				for ( unsigned int j=0; j<names.size(); j++ ) {
					if ( !db->ResourceExists( names[j], types[i] ) ) {
						p->BarStep();
						p->Newline( Progress::verbose );
						p->Print( "  &"+types[i]+" &"+names[j]+" is new", Progress::verbose );
						resource_info_t ri;
						ri.resource_name=names[j];
						ri.resource_type=types[i];
						ri.is_new = true;
						ri.is_deleted = false;
						if ( SCPM_conf::GetHandle()->Skip() )
							ri.save = false;
						else
							ri.save = true;
						switch_info.modified_resources.push_back( ri );
						switch_info.profile_modified = true;
					}
				}
			}
			for ( unsigned int i=0; i<types.size(); i++ ) {
				vector<string> names = db->ResourceGetNames( types[i] );
				for ( unsigned int j=0; j<names.size(); j++ ) {
					bool found = false;
					vector<string> avail = rh.GetResourceNames( types[i] );
					for ( unsigned int k=0; k<avail.size(); k++ )
						if ( avail[k] == names[j] ) { found = true; break; }
					if ( !found ) {
						p->Newline( Progress::verbose );
						p->Print( "  &"+types[i]+" &"+names[j]+" is deleted", Progress::verbose );
						resource_info_t ri;
						ri.resource_name=names[j];
						ri.resource_type=types[i];
						ri.is_new = false;
						ri.is_deleted = true;
						if ( SCPM_conf::GetHandle()->Skip() )
							ri.save = false;
						else
							ri.save = true;
						switch_info.modified_resources.push_back( ri );
						switch_info.profile_modified = true;
					}
				}
			}
		}
	
		p->Newline();
		p->ResetBar();

		switch_prepared=true;
	}
	catch( ... ) {
			// ensure full bar for the yast2 frontend
		p->ResetBar();
		return false;
	}
	return true;
}

bool SCPM::Switch( switch_info_t &switch_info )
{
	Progress *pr = Progress::GetHandle();
	try {
		Prepare(true);

		if ( !sh->ProfileExists( switch_info.profile_name ) ) {
			LOGERR( "Profile &"+switch_info.profile_name+" does not exist" );
			return false;
		}

		string active = sh->GetActiveProfile();

		if ( active.find( "(none)" ) != string::npos ) {
			LOGERR( "could not determine active profile" );
			return false;
		}

		SCPM_conf *conf = SCPM_conf::GetHandle();
		SCDB *db = SCDB::GetHandle();
		jrnl->New();

		Profile active_profile( active );
		Profile dest_profile( switch_info.profile_name );

		vector<Resource*> resources_to_stop;
		vector<Resource*> resources_to_start;
		vector<Resource*> resources_keep_running;
		if ( ! (conf->BootMode()) ) {
				// to determine which resources have to be stopped or
				// started, we must compare the active resources of
				// the two profiles
			sh->DetermineResourceActions( active_profile, dest_profile, resources_to_start,
																		resources_to_stop, resources_keep_running );
				// check out dependencies, maybe some resources have to be restarted
				// because of modified dependend resources
			vector<Resource*> resources_to_restart;
			sh->ResolveDeps( active_profile, dest_profile, resources_keep_running, resources_to_restart );

				// add the resources to _to_start and to _to_stop
			for (unsigned int i=0; i<resources_to_restart.size(); i++) {
				LOGDEBUG( resources_to_restart[i]->GetName()+ " has to be restarted" );
				resources_to_start.push_back( resources_to_restart[i] );
				resources_to_stop.push_back( resources_to_restart[i] );
			}
				// kill dupes and order resources respective to their dependencies
			sh->OrderResources( resources_to_stop, true );

				// doing OrderResources() for the to_start resources here
				// because they may need the restored data for priority determination
			sh->OrderResources( resources_to_start, false );

				// add resources_to_stop to journal
			for (unsigned int i=0;i<resources_to_stop.size();i++)
				jrnl->AddEntry( Journal::cmd_resource_stop,
												resources_to_stop[i]->GetType(),
												resources_to_stop[i]->GetName() );
		}
		else {
			LOGDEBUG( "operating in boot mode, skipping start/stop actions" );
		}

		string p = db->ProfileGetScript(active,"poststop");
		if (p!="") jrnl->AddEntry( Journal::cmd_script, p );

		jrnl->AddEntry( Journal::cmd_profile_set_key,
										switch_info.profile_name, "clean", "no" );
		dest_profile.Restore();
		jrnl->AddEntry( Journal::cmd_profile_set_key,
										switch_info.profile_name, "active", "yes" );
		jrnl->AddEntry( Journal::cmd_set_global_key, "active-profile",
										switch_info.profile_name );
		jrnl->AddEntry( Journal::cmd_profile_set_key, active, "active", "no" );

		p=db->ProfileGetScript( switch_info.profile_name, "prestart" );
		if (p!="") jrnl->AddEntry( Journal::cmd_script, p );

		if ( !(conf->BootMode()) ) {
			for(unsigned int i=0;i<resources_to_start.size();i++)
				jrnl->AddEntry( Journal::cmd_resource_start, resources_to_start[i]->GetType(),
												resources_to_start[i]->GetName() );
			p=db->ProfileGetScript( switch_info.profile_name, "poststart" );
			if (p!="") jrnl->AddEntry( Journal::cmd_script, p );

			vector<Resource*> to_free;
			for (unsigned int i=0; i<resources_to_stop.size(); i++) 
				to_free.push_back( resources_to_stop[i] );
			for (unsigned int i=0; i<resources_to_start.size(); i++) 
				to_free.push_back( resources_to_start[i] );
			for (unsigned int i=0; i<resources_keep_running.size(); i++) 
				to_free.push_back( resources_keep_running[i] );

			unsigned int size = to_free.size();
			for (unsigned int i=0; i<size; i++) {
				bool free_it=true;
				for (unsigned int j=i+1; j<size; j++) {
					if ( to_free[i] == to_free[j] ) free_it=false;
				}
				if ( free_it ) delete to_free[i];
			}
		}
		
			// journal is done
		jrnl->Ready( "nomagic" );
		pr->SetSegment(74);
		if ( !(conf->BootMode()) )
			pr->Print( "Switching to profile" );
		else
			pr->Print( "Restoring profile" );
		pr->Print( switch_info.profile_name );
		sh->ProcessJournal();
		pr->Newline();

		if ( !(conf->BootMode()) )
			LOGINFO( "restored profile "+ switch_info.profile_name );
		else
			LOGINFO( "switched to profile "+ switch_info.profile_name );

			// disabled for testing purposes
		sh->SaveMagic();
		sh->SaveSCDB();
		sh->WriteBootInfo();
	}
	catch( ... ) {
		pr->ResetBar();
		return false;
	}
	return true;
}

bool SCPM::Modify( string profile, string action )
{
	try {
		Prepare(true);

		if ( action == "begin" ) {
			LOGDEBUG( "Modify() called, action start" );
		}
		else if ( action == "end" )
		{
			LOGDEBUG( "Modify() called, action start" );
		}
		else {
			LOGDEBUG( "Modify() called with unknown action command" );
		}

		LOGINFO( "modify "+profile+" "+action );
	}
	catch( ... ) {
		return false;
	}
	return true;
}	
	

bool SCPM::Add( string profile, bool auto_switch )
{
	Progress *pr = Progress::GetHandle();
	try {
		Prepare(true);

		if ( sh->ProfileExists( profile ) ) {
			LOGERR( "Profile &"+profile+" already exists" );
			return false;
		}
	
		// check if invalid characters are in use
		if (profile.find(' ') != string::npos || profile.find('*') != string::npos || profile.find('<') != string::npos || profile.find('>') != string::npos)
		{
			LOGERR( "You may not use space, *, < and > in the profile name" );
			return false;
		}

		pr->Print( "Creating new profile &"+profile );
		jrnl->New();
		jrnl->AddEntry( Journal::cmd_profile_add, profile );
		jrnl->AddEntry( Journal::cmd_message, "Adding resources" );
		sh->AddProfileResources( profile );

		if ( auto_switch ) {
			jrnl->AddEntry( Journal::cmd_message, "Setting profile &"+profile+" to 'active'" );
			jrnl->AddEntry( Journal::cmd_profile_set_key, sh->GetActiveProfile(), "active", "no" );
			jrnl->AddEntry( Journal::cmd_profile_set_key, sh->GetActiveProfile(), "clean", "yes" );
			jrnl->AddEntry( Journal::cmd_profile_set_key, profile, "active", "yes" );
			jrnl->AddEntry( Journal::cmd_profile_set_key, profile, "clean", "no" );
			jrnl->AddEntry( Journal::cmd_set_global_key, "active-profile", profile );
		}

		jrnl->AddEntry( Journal::cmd_write_boot_info );

		jrnl->Ready( "nomagic" );
		pr->SetBar(100);
		sh->ProcessJournal();
		pr->Newline();

		LOGINFO( "profile "+profile+" added" );
	}
	catch(...) {
		pr->ResetBar();
		return false;
	}
	return true;
}

bool SCPM::Copy( string source_profile, string profile )
{
	try {
		Prepare(true);
		return this->Create( profile, source_profile );
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::Reload( )
{
	try {
		Prepare(true);
		string active_profile = sh->GetActiveProfile();

		if ( active_profile == "(none)" ) {
			LOGERR( "no profile active" );
			return false;
		}

		switch_info_t switch_info;

		if ( !this->PrepareSwitch( active_profile, switch_info ) ) return false;

		SCPM_conf *conf = SCPM_conf::GetHandle();
		if ( conf->Skip() ) {
				// in skip mode we have to set all modified resources to "not save"
				// (default is save)
			for ( unsigned int i=0; i<switch_info.modified_resources.size(); i++ )
				switch_info.modified_resources[i].save=false;
		}
		return( this->Switch( switch_info ) );
	}
	catch( ... ) {
		return false;
	}
}

bool SCPM::Set( string command, string argument, string profile )
{
	try {
		Prepare(true);

		if ( command == "resource_set" ) return this->SetResourceSet( argument );
		if ( profile.empty() ) profile=sh->GetActiveProfile();
		if ( !sh->ProfileExists( profile ) ) {
			LOGERR( "profile &"+profile+" does not exist" );
			return false;
		}
		if ( command == "description" ) sh->SetDescription( profile, argument );
		else if ( command == "prestart" || command == "poststart" || command == "prestop" ||
							command == "poststop" ) sh->SetScript( profile, command, argument );
		else {
			LOGERR( "unknown set command &"+command );
			return false;
		}
	}
	catch( ... ) {
		return false;
	}
	if ( argument.empty() ) 
		LOGINFO( "unset "+command+" for profile " +profile );
	else
		LOGINFO( "set "+command+" to "+argument+" for profile " +profile );
	return true;
}

bool SCPM::Get( string command, string &result, string profile )
{
	try {
		Prepare();

		if ( command == "resource_set" ) return this->GetResourceSet( result );
		if ( command == "resource_sets" ) {
			vector<string> pre;
			vector<string> ind;
			this->ListResourceSets( pre, ind );
			for ( unsigned int i=0; i<pre.size(); i++ ) result+=pre[i]+"\n";
			result+="----\n";
			for ( unsigned int i=0; i<ind.size(); i++ ) result+=ind[i]+"\n";
			return true;
		}
		if ( profile.empty() ) profile=sh->GetActiveProfile();
		if ( !sh->ProfileExists( profile ) ) {
			LOGERR( "profile &"+profile+" does not exist" );
			return false;
		}
		if ( command == "description" )
			result = sh->GetDescription( profile );
		else if ( command == "prestart" || command == "poststart" || command == "prestop" ||
							command == "poststop" ) 
			result = sh->GetScript( profile, command );
		else {
			LOGERR( "unknown get command &"+command );
			return false;
		}
	}
	catch( ... ) {
		return false;
	}
	return true;

}

bool SCPM::Save( switch_info_t &switch_info )
{
	Progress *pr = Progress::GetHandle();
	SCDB *db = SCDB::GetHandle();
	try {
		Prepare();

		if ( !sh->ProfileExists( switch_info.profile_name ) ) {
			LOGERR( "Profile &"+switch_info.profile_name+" does not exist" );
			return false;
		}

		string active = sh->GetActiveProfile();

		if ( active.find( "(none)" ) != string::npos ) {
			LOGERR( "could not determine active profile" );
			return false;
		}

			// init journal
		jrnl->New();
		
		for ( unsigned int i=0; i<switch_info.modified_resources.size(); i++ ) {
			if ( switch_info.modified_resources[i].is_new && switch_info.modified_resources[i].save )
				jrnl->AddEntry( Journal::cmd_resource_create, 
												switch_info.modified_resources[i].resource_type,
												switch_info.modified_resources[i].resource_name, "all" );
			if ( switch_info.modified_resources[i].is_deleted && 
					 switch_info.modified_resources[i].save )
				jrnl->AddEntry( Journal::cmd_resource_drop, 
												switch_info.modified_resources[i].resource_type,
												switch_info.modified_resources[i].resource_name, "all" );
		}
		
		Profile active_profile( active );
		Profile dest_profile( switch_info.profile_name );
		
		if (active_profile.HasScript("prestop"))
			jrnl->AddEntry( Journal::cmd_script, db->ProfileGetScript(active, "prestop") );
	
		if ( !active_profile.IsClean() ) {
			LOGDEBUG( "profile "+active+" is not clean" );

			for ( unsigned int i=0; i<switch_info.modified_resources.size(); i++ ) {
				switch(switch_info.modified_resources[i].save_mode) {
					case save_all:
						sh->ApplyAll( switch_info.modified_resources[i].resource_type,
													switch_info.modified_resources[i].resource_name );
						break;
					case patch_all:
						sh->PatchAll( switch_info.modified_resources[i].resource_type,
													switch_info.modified_resources[i].resource_name );
						break;
					case normal:
							// do nothing, normal will be saved using the profile handler
						break;
				}
			}
				// check modified resources/services and
			if ( switch_info.profile_modified ) {
				for (unsigned int i=0;i<switch_info.modified_resources.size();i++)
					if (switch_info.modified_resources[i].save)
						jrnl->AddEntry( Journal::cmd_resource_save,
														switch_info.modified_resources[i].resource_type,
														switch_info.modified_resources[i].resource_name,
														active );
			}
		}
		else {
				// unusal case, may be happen at boot time
			LOGDEBUG( "profile "+active+" is clean" );
				// no action required
		}

			// journal is ready now
		jrnl->Ready( "nomagic" );

		pr->SetBar( 100 );
			// we give it 26% bar space
		pr->SetSegment( 26 );
		pr->Print( "Saving resources" );
		sh->ProcessJournal();
		pr->Newline();

			// ensuring full bar (HACK!)
		pr->SetSegment( 1 );
		pr->SetSegmentSteps( 1 );
		pr->BarStep();
	}
	catch(...) {
			// ensuring full bar (HACK!)
		pr->ResetBar();
		return false;
	}
	return true;
}

bool SCPM::Status( scpm_status_t &status )
{
		// fill in defaults
	status.enabled = false;
	status.initialized = false;
	status.db_loaded = false;
	status.db_uptodate = false;
	status.locked = false;
	status.needs_reinit = false;
	status.needs_recover = false;
	status.active_profile = "(none)";
	status.scpm_version = SCPM_VERSION;
	status.db_version = "(n/a)";
	status.db_format = "(n/a)";
	status.db_format_version = "(n/a)";
	status.needs_reinit = sh->InstallationChanged();
	try {
		SCDB *db = SCDB::GetHandle();
		if ( db_open ) {
			status.db_loaded = true;
			status.enabled = sh->IsEnabled();
			status.initialized = sh->IsInitialized();
			status.db_uptodate = db->UpToDate();
			status.needs_reinit = sh->InstallationChanged();
		}
		status.locked = ( jrnl->GetStatus() == Journal::STATUS_LOCKED ) ? true : false;
		status.needs_recover = ( jrnl->GetStatus() == Journal::STATUS_RECOVER_NEEDED ) ? true : false;
		status.active_profile = sh->GetActiveProfile();
		status.scpm_version = SCPM_VERSION;
		status.db_version = db->GetSCDBVersion();
		status.db_format = db->GetFormat();
		status.db_format_version = db->GetFormatVersion();
	}
	catch(...) {
		LOGWARN( "could not query scpm status" );
			// but this is not a real abortion situation
		Log::GetHandle()->ClearAborted();
	}
	return true;
}

bool SCPM::SetResourceSet( string set )
{
	SCPM_conf *cfg = SCPM_conf::GetHandle();
	if ( set == "none" )
		cfg->SetResourceSet( "" );
	else if ( set == "auto" )
		cfg->SetResourceSet( "" );
	else
		cfg->SetResourceSet( set );
		// in this case the config file has to be written immediatly, because the
		// handlers read this value from there
	cfg->SaveConfig();
	LOGINFO( "Set resource set "+set );
	return true;
}

bool SCPM::GetResourceSet( string &set )
{
	set = SCPM_conf::GetHandle()->GetResourceSet();
	if ( set.empty() ) set = "auto";
	return true;
}

bool SCPM::ListResourceSets( vector<string> &predefined, vector<string> &individual )
{
	DIR *dp;
	struct dirent *ep;
	SCPM_conf *conf = SCPM_conf::GetHandle();

	individual.clear();
	predefined.clear();
		//predefined.push_back( "auto" );
	try {
		dp = opendir( conf->GetUserSetDir().c_str() );
		if ( dp == NULL ) {
			LOGWARN( "could not open dir " + conf->GetUserSetDir() );
			Log::GetHandle()->ClearAborted();
		}
		else {
			individual.clear();
			while( (ep = readdir( dp )) ) {
				if ( ep->d_name[0]=='.' ) continue;
				individual.push_back( ep->d_name );
			}
		}
		
			// individual done
		closedir( dp );
		dp = opendir( conf->GetSetDir().c_str() );
		if ( dp == NULL ) {
			LOGWARN( "could not open dir " + conf->GetSetDir() );
			Log::GetHandle()->ClearAborted();
		}
		else {
			while( (ep = readdir( dp )) ) {
				if ( ep->d_name[0]=='.' ) continue;
				bool found=false;
				for ( unsigned int i=0; i<individual.size(); i++ )
					if ( ep->d_name == individual[i] ) found = true;
				if ( !found ) predefined.push_back( ep->d_name );
			}
			closedir( dp );
		}
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::CopyResourceSet( string set, string newset )
{
	try {
		LOGERR( "not yet implemented" );
	}
	catch(...) {
	}
	return false;
}

bool SCPM::DeleteResourceSet( string set )
{
	try {
		LOGERR( "not yet implemented" );
	}
	catch(...) {
	}
	return false;
}

bool SCPM::ShowChanges( ostream &output, string resource_type, string resource_name, bool backup, string version )
{
	Resource *r;
	try {
		Prepare(true);
		Resource_helpers rh;
		r = rh.CreateObject( resource_type, resource_name );
		r->SetProfile( sh->GetActiveProfile() );
		r->ShowChanges( output, backup, version );
		delete r;
	}
	catch( Resource::not_available ) {
		delete r;
		output 
			<< LOGTRANS( "resource type &"+resource_type+" does not support show changes feature" )
			<< endl;
	}
	catch( Resource::resource_unknown ) {
		output << LOGTRANS( "unknown resource &"+resource_name+" (&"+resource_type+")" ) << endl;
		return false;
	}
	catch( Resource_helpers::type_invalid ) {
		output << LOGTRANS( "unknown resource type &"+resource_type ) << endl;
		return false;
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::RebuildDB( rebuild_mode_t mode )
{
	SCDBUtil util;
	try {
		Prepare(true);
		switch( mode ) {
			case REBUILD_MODE_QUICK:
			{
				vector<string> names;
				vector<string> types;
				util.FindDeletedResources( names, types );
				for ( unsigned int i=0; i<names.size(); i++ ) 
					util.DropResource( names[i], types[i] );
				util.MaximizeDB( );
				break;
			}
			case REBUILD_MODE_RESOURCES_ONLY:
				util.RebuildResources();
				break;
			case REBUILD_MODE_DATA_ONLY:
					//util.RebuildData();
				LOGERR( "not yet implemented" );
				break;
			case REBUILD_MODE_FULL:
					//util.RebuildResources();
					//util.RebuildData();
				LOGERR( "not yet implemented" );
				break;
			default:
				LOGERR( "unknown rebuild mode specified" );
		}
	}
	catch( SCDBUtil::operation_failed &e ) {
		LOGWARN( "scdb error "+e.reason );
		LOGERR( "scdb rebuild failed" );
	}
	catch( ... ) {
		return false;
	}
	LOGINFO( "Rebuilt database" );
	return true;
}

bool SCPM::ListResourceGroups( vector<resource_group_t> &groups )
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	groups.clear();

	try {
		DIR *dp;
		struct dirent *ep;
		dp = opendir( conf->GetUserGroupDir().c_str() );
		if ( dp == NULL ) {
			LOGWARN( "could not open dir " + conf->GetUserGroupDir() );
			Log::GetHandle()->ClearAborted();
		}
		else {
			while( (ep = readdir( dp )) ) {
				if ( ep->d_name[0]=='.' ) continue;
				if ( sh->GroupDeleted( ep->d_name ) ) continue;
				resource_group_t grp;
				grp.name = ep->d_name;
				grp.user_modified = true;
				grp.user_defined	= true;
				grp.description = sh->GetGroupDescription( ep->d_name );
				grp.active = sh->GroupActive( ep->d_name );
				groups.push_back( grp );
			}
		}
		
			// user groups done
		closedir( dp );
		dp = opendir( conf->GetGroupDir().c_str() );
		if ( dp == NULL ) {
			LOGWARN( "could not open dir " + conf->GetSetDir() );
			Log::GetHandle()->ClearAborted();
		}
		else {
			unsigned int user_groups_count = groups.size();
			while( (ep = readdir( dp )) ) {
				if ( ep->d_name[0]=='.' ) continue;
				if ( sh->GroupDeleted( ep->d_name ) ) continue;
				bool found=false;
				for ( unsigned int i=0; i<user_groups_count; i++ )
					if ( ep->d_name == groups[i].name ) {
							// group is also user group, but not user defined
						groups[i].user_defined = false;
						found = true;
					}
				if ( !found ) {
						// normal group found
					resource_group_t grp;
					grp.name = ep->d_name;
					grp.user_modified = false;
					grp.user_defined = false;
					grp.description = sh->GetGroupDescription( ep->d_name );
					grp.active = sh->GroupActive( ep->d_name );
					groups.push_back( grp );
				}
			}
			closedir( dp );
		}
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::GetResourceGroup( const string &name, vector<resource_entry_t> &group,
														 bool nouser )
{
	try {
		string foo;
		sh->GetResourceGroup( name, group, foo, nouser );
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::GetResourceGroup( const string &name, vector<resource_entry_t> &group,
														 string &description, bool nouser )
{
	try {
		sh->GetResourceGroup( name, group, description, nouser );
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::SetResourceGroup( const string &name, const vector<resource_entry_t> &group,
														 const string &description )
{
	if (init_failed) return false;
	try {
		sh->SetResourceGroup( name, group, description );
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::DeleteResourceGroup( const string &group )
{
	if (init_failed) return false;
	try {
		sh->DeactivateResourceGroup( group );
		sh->DeleteResourceGroup( group );
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::RenameResourceGroup( const string &group, const string &newgroup )
{
	try {
		if (init_failed) return false;
		sh->RenameResourceGroup( group, newgroup );
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::SetActiveResourceGroups( const vector<string> &groups )
{
	if (init_failed) return false;
	sh->SetActiveResourceGroups( groups );
	return true;
}

bool SCPM::GetActiveResourceGroups( vector<string> &groups )
{
	try {
		sh->GetActiveResourceGroups( groups );
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::ActivateResourceGroup( const string &group, bool auto_add )
{
	try {
		if (auto_add)
			Prepare(true);
		else
			if (init_failed) return false;
		sh->ActivateResourceGroup( group, auto_add );
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::DeactivateResourceGroup( const string &group, bool auto_remove )
{
	try {
		if (auto_remove)
			Prepare(true);
		else if (init_failed) return false;
		sh->DeactivateResourceGroup( group, auto_remove );
	}
	catch(...) {
		return false;
	}
	return true;
}

vector<resource_group_t> &SCPM::GetGroupsByResource( const string &resname,
																										 const string &restype )
{
	static vector<resource_group_t> groups;
	vector<resource_group_t> allgroups;
	
	groups.clear();
	this->ListResourceGroups( allgroups );
	for ( unsigned int i=0; i<allgroups.size(); i++ ) {
		vector<resource_entry_t> entries;
		if ( !this->GetResourceGroup( allgroups[i].name, entries ) ) 
			LOGERR( "could not query resource group &"+allgroups[i].name );
		for (unsigned int j=0; j<entries.size(); j++ ) {
			if ( (entries[j].name == resname) && (entries[j].type == restype) )
				groups.push_back( allgroups[i] );
		}
	}
	if ( groups.empty() ) {
		resource_group_t grp;
		grp.name = "other";
		grp.description = "Resources without groups";
		grp.user_defined = false;
		grp.user_modified = false;
		grp.active = true;
		groups.push_back( grp );
	}
	return groups;
}

bool SCPM::SetConfigOption( const string &option, const string &value )
{
	try {
		SCPM_conf::GetHandle()->SaveOption( option, value );
	}
	catch( SCPM_conf::io_error ) {
		return false;
	}
	catch( SCPM_conf::no_such_option ) {
		return false;
	}
	return true;
}

bool SCPM::ResetResourceGroup( const string &group )
{
	try {
		if (init_failed) return false;
		sh->ResetResourceGroup( group );
	}
	catch(...) {
		Log::GetHandle()->ClearAborted();
		return false;
	}
	return true;
}	 

bool SCPM::ResetAllGroups( )
{
	try {
		if (init_failed) return false;
		sh->ResetAllGroups();
	}
	catch(...) {
		Log::GetHandle()->ClearAborted();
		return false;
	}
	return true;
}

bool SCPM::GetConfigOption( const string &option, string &value )
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	if ( option == "DEBUG" )
		value = conf->Debug() ? "yes" : "no";
	else if ( option == "VERBOSE" )
		value = conf->Verbose() ? "yes" : "no";
	else if ( option == "SWITCHMODE" ) {
		if ( conf->Force() ) value = "force";
		else if ( conf->Skip() ) value = "skip";
		else value = "normal";
	}
	else if ( option == "BOOTMODE" ) {
		value = conf->GetBootMode();
	}
	return true;
}

bool SCPM::Update( const string &scriptname, vector<pair<string, string> > resources )
{
	string scriptfile;
	if ( scriptname=="-" ) {
		char tmpfile[40] = "/tmp/scpm.update_script.XXXXXX";
		if ( mktemp( tmpfile ) == NULL ) {
			LOGERR( "could not create file &"+string(tmpfile) );
			return false;
		}
		ofstream o( tmpfile );
		o << cin.rdbuf();
		o.close();
		scriptfile = tmpfile;
	}
	else {
		scriptfile = scriptname;
	}

	if (init_failed) {
		LOGWARN( "SCPM initialization failed, just running the update script" );
		::setenv( "SCPM_UPDATE", "yes", 1 );
		::setenv( "SCPM_INACTIVE", "yes", 1 );
		sh->RunScript( scriptfile );
		return true;
	}

	try {
			// we use the boot mode so we can use the Switch command without
			// resources being started/stopped and without switch scripts being
			// executed
		SCPM_conf::GetHandle()->SetBootMode( true );

			// some vars
		Progress *pr = Progress::GetHandle();
		vector<string> profiles = sh->GetAllProfiles();
		string active_profile = sh->GetActiveProfile();
		switch_info_t switch_info;
		::setenv( "SCPM_UPDATE", "yes", 1 );
		
		if ( resources.empty() ) {
				// no resources specified, we need to do full switches
				// so we restore every profile and run the update script
				// save all changes afterwards

			pr->Print( "Updating profile &"+active_profile );
			pr->Newline();
			::setenv( "SCPM_PROFILE", active_profile.c_str(), 1 );
			sh->RunScript( scriptfile );
			for (unsigned int i=0; i<profiles.size(); i++) {
				if ( profiles[i] == active_profile ) continue;
				this->PrepareSwitch( profiles[i], switch_info );
				pr->Print( "Updating profile &"+profiles[i] );	
				pr->Newline();
				this->Switch( switch_info );
				::setenv( "SCPM_PROFILE", profiles[i].c_str(), 1 );
				sh->RunScript( scriptfile );
				for (unsigned int j=0; j<switch_info.modified_resources.size(); j++)
					switch_info.modified_resources[j].save = true;
			}
				// switch back to default profile
			this->PrepareSwitch( active_profile, switch_info );
			for (unsigned int j=0; j<switch_info.modified_resources.size(); j++)
				switch_info.modified_resources[j].save = true;
			this->Switch( switch_info );
		}
		else {
				// we only need to restore and save given resources
			sh->CheckResources( resources );
			if ( resources.empty() ) {
				LOGWARN( "no (valid) resource to update, just running update script" );
				::setenv( "SCPM_UPDATE", "yes", 1 );
				sh->RunScript( scriptfile );
				return true;
			}	
			::setenv( "SCPM_PROFILE", active_profile.c_str(), 1 );
			pr->Print( "Updating profile &"+active_profile );
			pr->Newline();
			sh->RunScript( scriptfile );
			sh->SaveResources( active_profile, resources );
			for (unsigned int i=0; i<profiles.size(); i++) {
				if ( profiles[i] == active_profile ) continue;
				pr->Print( "Updating profile &"+profiles[i] );
				pr->Newline();
				sh->RestoreResources( profiles[i], resources );
				::setenv( "SCPM_PROFILE", profiles[i].c_str(), 1 );
				sh->RunScript( scriptfile );
				sh->SaveResources( profiles[i], resources );
			}
			sh->RestoreResources( active_profile, resources );
		}
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::UpdateResources( )
{
	if (init_failed) return false;
	try {
		sh->UpdateResources();
		sh->WriteBootInfo();
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::RestoreBackup( const vector<pair<string,string> > &resources, const vector<string> &versions, const string &profile) 
{
	if (init_failed) return false;
	try {
		// we need as much version-strings as resources
		if (resources.size() != versions.size()) {
			LOGERR("number of versions does not match number of resources");
			return false;
		}

		// does the specified profile exist
		if (!profile.empty() && profile != "all") {
			if (!SCDB::GetHandle()->ProfileExists( profile) ) {
				LOGERR( "Profile &"+profile+" does not exist" );
				return false;
			}
		}
	
		// get resources to restore 	
		vector<Resource*> res;
		vector<string> ver;

		if (resources.empty()) {
			res = sh->GetResourcesWithBackup( profile );
			ver.insert(ver.end(), res.size(), "");
		}
		else {
			res = sh->GetResourcesByName( profile, resources );
			ver = versions;
		}

		// restore backup(s)
		sh->RestoreBackup( res, ver );

		// cleanup
		for(unsigned int i=0; i<res.size(); i++) delete res[i];
	}
	catch(...) {
		return false;
	}
	return true;

}

bool SCPM::RemoveBackup( const vector<pair<string,string> > &resources, const vector<string> &versions, const string &profile)
{
	if (init_failed) return false;
	try {
		// we need as much version-strings as resources
		if (resources.size() != versions.size()) {
			LOGERR("number of versions does not match number of resources");
			return false;
		}

		// does the specified profile exist
		if (!profile.empty() && profile != "all") {
			if (!SCDB::GetHandle()->ProfileExists( profile) ) {
				LOGERR( "Profile &"+profile+" does not exist" );
				return false;
			}
		}
		// get resources to remove
		vector<Resource*> res;
		vector<string> ver;
		if (resources.empty()) {
			res = sh->GetResourcesWithBackup( profile );
			ver.insert(ver.end(), res.size(), "");
		}
		else {
			res = sh->GetResourcesByName( profile, resources );
			ver = versions;
		}

		// remove backup(s)
		sh->RemoveBackup( res, ver );
		}
	catch(...) {
		return false;
	}
	return true;

}

bool SCPM::ListBackups( const string &profile, vector<backup_info_t> *reslist)
{
	if (init_failed) return false;
	try {
		if (!profile.empty() && profile != "all") {
			if (!SCDB::GetHandle()->ProfileExists( profile) ) {
				LOGERR( "Profile &"+profile+" does not exist" );
				return false;
			}
		}
	
		// is reslist ok?
		if (reslist==NULL) {
			LOGDEBUG( "uh? reslist is NULL" );
			LOGERR( "could not query profiles for backup resources" );
			return false;
		}

		// get all resources with backups
		vector<Resource*> res = sh->GetResourcesWithBackup( profile );
		reslist->clear();
		
		// fill up reslist-vector with found backups
		for (unsigned int i=0; i<res.size(); i++) {
			backup_info_t bi;
			bi.name = res[i]->GetName();
			bi.type = res[i]->GetType();
			bi.profile = res[i]->GetProfile();
			bi.versions = res[i]->GetBackupVersions(); 
			reslist->push_back( bi );
			delete res[i];
		}
	}
	catch(...) {
		return false;
	}
	return true;

}

bool SCPM::AddBackup( const vector<pair<string,string> > &resources, const string &profile )
{
	if (init_failed) return false;
	try {
		if (!profile.empty() && profile != "all") {
			if (!SCDB::GetHandle()->ProfileExists( profile) ) {
				LOGERR( "Profile &"+profile+" does not exist" );
				return false;
			}
		}
		sh->AddBackup( resources );
	}
	catch(...) {
		return false;
	}
	return true;
}

bool SCPM::Recover( bool rollback )
{
	Progress *pr = Progress::GetHandle();
	try {
		if (!db_open) LOGERR( "scdb not available" );
		if (jrnl->GetStatus() != Journal::STATUS_RECOVER_NEEDED ) {
			LOGINFO( "no recover needed" );
			return true;
		}
		jrnl->Load( "nomagic" );
		pr->SetBar(100);
		pr->SetSegmentSteps( jrnl->Steps() );
		if (rollback) {
			pr->Print( "Rolling back journal" );
			for(;;) {
				Journal::entry e=jrnl->PrevEntry();
				sh->PerformCommand( e );
				pr->BarStep();
			}
		}
		else {
			pr->Print( "Replaying journal" );
			for(;;) {
				pr->BarStep();
				Journal::entry e=jrnl->NextEntry( );
				LOGDEBUG( "performing journal cmd "+itoa(e.command));
				sh->PerformCommand( e );
			}
		}
	}
	catch( Journal::no_more_entries ) {
			// should have sane state back
		init_failed=false;
		Log::GetHandle()->ClearAborted();
		jrnl->Unlock();
		jrnl->Delete();
		pr->Newline();
		return true;
	}
	catch(...) {
		pr->ResetBar();
		return false;
	}
	pr->ResetBar();
	return true;
}

bool SCPM::Commit( )
{
	try {
		if (jrnl->IsReady()) sh->ProcessJournal( true );
		else LOGWARN( "journal not ready, ignoring commit" );
	}
	catch(...) {
		return false;
	}
	return true;
}
