/*
 * $Id: scpm_conf.cc,v 1.4 2008/10/22 10:12:42 hschaa Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	SCPM main engine
 * File				 :	scpm_conf.cc
 * Description :	Interface for SCPM configuration facilities
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2001 SuSE GmbH
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "scpm_conf.hh"
#include <scpm_types.h>
#include <fstream>
#include <iostream>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>

using namespace std;

SCPM_conf::SCPM_conf( )
{
	struct stat buf;

	changed = false;

		// set some defaults
	log_to_syslog= false;
	log_to_file	 = false;
	log_to_stderr= false;
	log_debug		 = false;
	log_warning	 = true;
	log_error		 = true;
	log_to_stderr= false;
	skip_mode		 = false;
	force_mode	 = false;
	verbose			 = false;
	quiet				 = false;
	progress_bar = false;
	boot_mode		 = false;
	exit_warn		 = false;
	skip_load		 = false;
	force_db		 = false;
	write_boot_config = false;
	read_only		 = false;
	max_backups	 = 4;
	wait_commit	 = false;

	cfg_fname = "/etc/sysconfig/scpm";
	
	if (stat("/etc/sysconfig/scpm",&buf)!=0) {
		if (stat("/etc/scpm.conf",&buf)==0)
			cfg_fname = "/etc/scpm.conf";
		else {
			cerr << "could not open /etc/scpm.conf, using defaults" << endl;
			return;
		}
	}
		
	ifstream cfg_file( cfg_fname.c_str() );

	string s;
	string token;
	string value;
	int line=0;
	while ( cfg_file.good() ) {
		line++;
		getline( cfg_file, s );
		if ( s.empty() ) continue;
		if ( s.at(0) == '#' ) continue;
		if ( s.find("=") == string::npos ) {
			cerr << "syntax error in /etc/scpm.conf (line " << line << ")" << endl;
			continue;
		}
		token = s.substr( 0, s.find("=") );
		value = s.substr( s.find("=")+1 );
		if ( value == "" ) {
			cerr << "variable " << token << " has empty value (ignored)" << endl;
			continue;
		}
		if ( value[0] == '"' ) {
				// strip quotes
			value = value.substr( 1, value.find_last_of("\"")-1 );
		}
		if ( token == "DBFILE" ) dbfile = value;
		if ( token == "LIBDIR" ) libdir = value;
		if ( token == "RPMDB" ) rpmfile = value;
		if ( token == "RESOURCE_SET" ) resource_set = value;
		if ( token == "PROFILEDIR" ) profiledir = value;
		if ( token == "ACTIVE_GROUPS_FILE" ) groupsfile = value;
		if ( token == "LOGFILE" ) {
			if ( value == "syslog" )
				log_to_syslog = true;
			else if ( value == "stderr" )
				log_to_stderr = true;
			else {
				log_to_file = true;
				logfile = value;
			}
		}
		if ( token == "DEBUG" ) {
			if ( value == "yes" ) log_debug=true;
			else log_debug=false;
		}
		if ( token == "VERBOSE" ) {
			if ( value == "yes" ) verbose=true;
			else verbose=false;
		}
		if ( token == "SWITCH_MODE" ) {
			if ( value == "force" ) force_mode = true;
			if ( value == "skip" ) skip_mode = true;
		}
		if ( token == "BOOT_MODE" ) {
			if ( value == "force" || value == "skip" || value == "normal" )
				boot_mode_str = value;
		}
		if ( token == "WRITE_BOOT_CONFIG" ) {
			if ( value == "yes" )
				write_boot_config = true;
		}
	}
		// use internal defaults in case some data is missing
	if ( dbfile.empty() ) dbfile = "/var/lib/scpm/scdb/scdb.db";
	if ( libdir.empty() ) libdir = "/lib/scpm";
	if ( rpmfile.empty() ) rpmfile = "/var/lib/rpm/packages.rpm";
	if ( profiledir.empty() ) profiledir = "/var/lib/scpm/profiles";
	if ( groupsfile.empty() ) groupsfile = "/var/lib/scpm/active_groups";
	
		// check out init directory (FIXME: this can surely be done better)
	if ( stat( "/etc/init.d", &buf ) == 0 )
		initdir = "/etc/init.d";
	else
		initdir = "/sbin/init.d";

}

SCPM_conf::~SCPM_conf()
{ 
	this->SaveConfig();
}

void SCPM_conf::SetOptions( int options )
{
	if ( options & scpm_flag_force )	 force_mode = true;
	if ( options & scpm_flag_skip )		 skip_mode = true;
	if ( options & scpm_flag_quiet )	 { quiet = true; verbose=false; }
	if ( options & scpm_flag_verbose ) verbose = true;
	if ( options & scpm_flag_hash )		 progress_bar = true;
	if ( options & scpm_flag_debug )	 log_debug = true;
	if ( options & scpm_flag_boot )		 boot_mode = true;
	if ( options & scpm_flag_exit_on_warning ) exit_warn = true;
	if ( options & scpm_flag_skip_load ) skip_load = true;
	if ( options & scpm_flag_force_db ) force_db = true;
	if ( options & scpm_flag_read_only ) read_only = true;
	if ( options & scpm_flag_wait_commit ) wait_commit = true;
}

void SCPM_conf::SaveConfig( )
{
	if ( changed ) {
		char tmpfile[] ="/tmp/scpm.conf.XXXXXX";
		if ( mktemp( tmpfile ) == NULL ) {
			cerr << "Could not create " << tmpfile << endl;
			return;
		}
		ifstream ins( cfg_fname.c_str() );
		ofstream outs( tmpfile );
		char buf[255];
		while ( ins.getline( buf, 254 ) ) {
			string s( buf );
			if ( s.find( "LIBDIR" ) == 0 ) s = "LIBDIR=\""+libdir+"\"";
			if ( s.find( "DBFILE" ) == 0 ) s = "DBFILE=\""+dbfile+"\"";
			if ( s.find( "LOGFILE" ) == 0 ) s = "LOGFILE=\""+logfile+"\"";
			if ( s.find( "RESOURCE_SET" ) == 0 ) s = "RESOURCE_SET=\""+resource_set+"\"";
			outs << s << endl;
		}
		ins.close();
		outs.close();
		outs.open( cfg_fname.c_str(), ios::trunc );
		ins.open( tmpfile );
		outs << ins.rdbuf();
		outs.close();
		ins.close();
		unlink( tmpfile );
	}
}

void SCPM_conf::SaveOption( const string &option, const string &value )
	throw( io_error, no_such_option )
{
	char tmpfile[] ="/tmp/scpm.conf.XXXXXX";
	if ( mktemp( tmpfile ) == NULL )
		throw io_error();

	ifstream ins( cfg_fname.c_str() );
	ofstream outs( tmpfile );
	char buf[255];
	bool found=false;
	while ( ins.getline( buf, 254 ) ) {
		string s( buf );
		if ( s.find( option ) == 0 ) {
			s = option+"=\""+value+"\"";
			found = true;
			if ( option == "DEBUG" ) {
				if ( value == "yes" ) log_debug=true;
				else if ( value == "no" ) log_debug=false;
			}
			else if ( option == "VERBOSE" ) {
				if ( value == "yes" ) verbose=true;
				else if ( value == "no" ) verbose=false;
			}
			else if ( option == "SWITCH_MODE" ) {
				if ( value == "normal" ) {
					force_mode = false;
					skip_mode = false;
				}
				else if ( value == "force" ) {
					force_mode = true;
					skip_mode = false;
				}
				else if ( value == "skip" ) {
					force_mode = false;
					skip_mode = true;
				}
			}
		}
		outs << s << endl;
	}
	ins.close();
	outs.close();
	outs.open( cfg_fname.c_str(), ios::trunc );
	ins.open( tmpfile );
	outs << ins.rdbuf();
	outs.close();
	ins.close();
	unlink( tmpfile );
	if (!found) throw no_such_option();
}
