/*
 * $Id: scpm_helpers.cc,v 1.14 2006/09/27 13:07:17 hschaa Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	SCPM main engine
 * File				 :	scpm_helpers.cc
 * Description :	implements helper functions for SCPM main module
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2001 SuSE GmbH
 *					 2002-2004 SuSE Linux AG
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "scpm_helpers.hh"
#include "scpm_conf.hh"
#include "log.hh"
#include <scdb.hh>
#include <scdb_util.hh>
#include <tokenizer.h>
#include "executor.hh"
#include "progress.hh"
#include "modlib.hh"
#include <fstream>
#include <sstream>
#include <cstdio>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <dirent.h>
#include <errno.h>

using namespace std;

#define module_name "scpm_helpers"

int scpm_journal_abort_steps = -1;

SCPM_helpers::SCPM_helpers( )
{
	db				 = SCDB::GetHandle();
	log				 = Log::GetHandle();
	stay_locked = false;
}

SCPM_helpers::~SCPM_helpers( )
{
}

int SCPM_helpers::LoadSCDB( )
{
	return db->LoadDB( );
}

void SCPM_helpers::SaveSCDB( )
{
	if ( !db->Modified() ) {
		LOGDEBUG( "scdb up to date" );
		return;
	}
		// update magic
	this->SaveMagic();

	SCPM_conf *cfg = SCPM_conf::GetHandle();

	string dbfile = cfg->GetDBFile();
	char tmpfile[255];
	std::sprintf( tmpfile, "%s.XXXXXX", dbfile.c_str() );

	if ( mktemp( tmpfile ) == NULL ) {
		LOGDEBUG( "mktemp failed" );
		LOGERR( "could not save scdb" );
		throw db_open_failed();
	}

	ofstream dbout( tmpfile );

	if ( dbout.bad() ) {
		LOGDEBUG( (string)"could not open "+tmpfile+" for writing" );
		LOGERR( "could not save scdb" );
		throw db_open_failed();
	}

	db->SaveDB( dbout );
	dbout.close();

	if ( log->Aborted() && !cfg->ForceDB() ) {
		LOGINFO( (string)"scpm has been aborted, (possibly corrupt) scdb has been saved to "
						 +tmpfile );
	}
	else {
		LOGDEBUG( "saved scdb" );
		dbout.open( dbfile.c_str(), ios::trunc );
		if ( dbout.bad() ) {
			LOGDEBUG( "cannot open "+dbfile+" for writing" );
			LOGERR( "could not save scdb" );
		}
		ifstream dbin( tmpfile );

		dbout << dbin.rdbuf();
		dbout.close();
		dbin.close();
		unlink( tmpfile );
	}
}

void SCPM_helpers::SaveMagic( )
{
	SCPM_conf *cfg = SCPM_conf::GetHandle();
	Modlib ml;
	string md5_1 = ml.GetMD5sum( cfg->GetRPMFile() );
	string md5_2;
	string setfile = cfg->GetResourceSet();
	if ( !setfile.empty() ) {
		ifstream ins( (cfg->GetUserSetDir()+"/"+setfile).c_str() );
		if ( !ins.good() ) {
			ins.clear();
			ins.open( (cfg->GetSetDir()+"/"+setfile).c_str() );
		}
		if ( !ins.good() )
			log->WriteMessage( "scpm_helpers", Log::CLASS_DEBUG,
												 "could not open resource set file ("+cfg->GetResourceSet()+")" );
		else {
			md5_2 = ml.GetMD5sum( ins );
			ins.close();
		}
	}
	try {
		db->SetStatusKey( "magic", md5_1+md5_2 );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "could not write magic ("+ e.reason+")" );
	}
}

bool SCPM_helpers::IsEnabled( )
{
	try {
		return db->GetStatusFlag( "enabled" );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGWARN( "status operation failed" );
		return false;
	}
}

void SCPM_helpers::SetDisabled( )
{
	try {
		db->SetStatusFlag( "enabled", false );
		return;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "status operation failed" );
}

bool SCPM_helpers::IsInitialized( )
{
	try {
		return db->GetStatusFlag( "initialized" );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "status operation failed" );
	return false;
}


bool SCPM_helpers::SystemChanged( )
{
	Modlib ml;
	SCPM_conf *conf = SCPM_conf::GetHandle();
	string md5_saved;
	string md5_actual1;
	string md5_actual2;

	try {
		md5_saved = db->GetStatusKey( "magic" );
		md5_actual1 = ml.GetMD5sum( conf->GetRPMFile() );
		string setname = conf->GetResourceSet();
		if ( !setname.empty() ) {
			ifstream ins( (conf->GetUserSetDir()+"/"+setname).c_str() );
			if ( !ins.good() ) {
				ins.clear();
				ins.open( (conf->GetSetDir()+"/"+setname).c_str() );
			}
			if ( ins.good() ) {
				md5_actual2 = ml.GetMD5sum( ins );
				ins.close();
			}
		}
	}
	catch( SCDB::db_action_failed ) {
		LOGWARN( "could not determine magic key" );
		return true;
	}
	if ( md5_saved == md5_actual1+md5_actual2 )
		return false;
	return true;
}

bool SCPM_helpers::ProfileExists( string profile )
{
	try {
		return db->ProfileExists( profile );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "status operation failed" );
	return false;
}

const string SCPM_helpers::GetActiveProfile( )
{
	try {
		return db->GetActiveProfile();
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "status operation failed" );
	return "";
}

const vector<string> SCPM_helpers::GetAllProfiles( )
{
	vector<string> v;

	try {
		v = db->GetProfiles();
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGERR( "could not determine defined profiles" );
		v.push_back( "(n/a)" );
	}

	if ( v.empty() ) {
		LOGWARN( "no profiles defined" );
		v.push_back( "(none)" );
	}

	return v;
}

void SCPM_helpers::CloneProfile( string profile, string source_profile )
{
	try {
		jrnl->AddEntry( Journal::cmd_profile_add, profile );
		
			// copy profile attributes
		string tmp;
		tmp = db->ProfileGetKey( source_profile, "description" );
		if (tmp.size()) jrnl->AddEntry( Journal::cmd_profile_set_key, profile, "description", tmp );
	
			// copy profile resources
		vector<string> restypes = db->ResourceGetTypes();
		for ( unsigned int i=0; i<restypes.size(); i++ ) {
			vector<string> resnames = db->ProfileGetResources( source_profile, restypes[i] );
			for ( unsigned int j=0; j<resnames.size(); j++ ) {
				jrnl->AddEntry( Journal::cmd_profile_add_resource, profile, restypes[i], resnames[j] );
				jrnl->AddEntry( Journal::cmd_file_copy, source_profile, profile, restypes[i], resnames[j] );
			}
		}

			// copy scripts
		tmp = db->ProfileGetScript( source_profile, "prestart" );
		if (tmp.size()) jrnl->AddEntry( Journal::cmd_profile_set_script, profile, "prestart", tmp );
		tmp = db->ProfileGetScript( source_profile, "poststart" );
		if (tmp.size()) jrnl->AddEntry( Journal::cmd_profile_set_script, profile, "poststart", tmp );
		tmp = db->ProfileGetScript( source_profile, "prestop" );
		if (tmp.size()) jrnl->AddEntry( Journal::cmd_profile_set_script, profile, "prestop", tmp );
		tmp = db->ProfileGetScript( source_profile, "poststop" );
		if (tmp.size()) jrnl->AddEntry( Journal::cmd_profile_set_script, profile, "poststop", tmp );

		return;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "could not copy profile &"+source_profile );
}

void SCPM_helpers::AddProfileResources( string profile )
{
	Resource_helpers rh;
	vector<string> types = rh.GetResourceTypes();
	for (unsigned int i=0;i<types.size();i++) {
		vector<string> names = rh.GetResourceNames( types[i] );
		for (unsigned int j=0;j<names.size();j++)
			jrnl->AddEntry( Journal::cmd_profile_add_resource, profile, types[i], names[j] );
	}
	for (unsigned int i=0;i<types.size();i++) {
		vector<string> names = rh.GetResourceNames( types[i] );
		for (unsigned int j=0;j<names.size();j++)
			jrnl->AddEntry( Journal::cmd_resource_save, types[i], names[j], profile );
	}
}

void SCPM_helpers::SetDescription( string profile, string description )
{
	try {
		db->ProfileSetKey( profile, "description", description );
		return;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "could not set description of profile &"+profile );
}

string SCPM_helpers::GetDescription( string profile )
{
	try {
		return db->ProfileGetKey( profile, "description" );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "could not get description of profile &"+profile );
	return "";
}

void SCPM_helpers::SetScript( string profile, string script_type, string script_name )
{
	try {
		return db->ProfileSetScript( profile, script_type, script_name );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "could not set &"+script_type+" script of profile &"+profile );
}

string SCPM_helpers::GetScript( string profile, string script_type )
{
	try {
		return db->ProfileGetScript( profile, script_type );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "could not get &"+script_type+" script of profile &"+profile );
	return "";
}

void SCPM_helpers::CreateResources( )
{
	Resource_helpers rh;
	vector<string> resource_types = rh.GetResourceTypes();

	if ( resource_types.size() == 0 ) {
		LOGERR( "no resource types defined" );
		return;
	}

	for ( unsigned int i=0; i < resource_types.size(); i++ ) {
		jrnl->AddEntry( Journal::cmd_resource_add_type, resource_types[i] );
		vector<string> resource_names = rh.GetResourceNames( resource_types[i] );
		for ( unsigned int j=0; j < resource_names.size(); j++ )
			jrnl->AddEntry( Journal::cmd_resource_add, resource_types[i], resource_names[j] );
	}

		// add dependencies
	Resource *r;
	for ( unsigned int i=0; i < resource_types.size(); i++ ) {
		vector<string> resource_names = rh.GetResourceNames( resource_types[i] );
		for ( unsigned int j=0; j < resource_names.size(); j++ ) {
			try {
				r = rh.CreateObject( resource_types[i], resource_names[j], true );
			}
			catch( Resource_helpers::type_invalid ) {
				LOGWARN( "used invalid resource type, this should not happen" );
				LOGERR( "could not create resources" );
				return;
			}
			for (unsigned int l=0; l<resource_types.size(); l++) {
				vector<string> deps;
				try {
					deps = r->GetDependencies( resource_types[l] );
				}
				catch( Resource::not_available ) {
					deps.clear();
				}
				if ( deps.size() > 0 ) {
					for (unsigned int k=0; k<deps.size(); k++)
						jrnl->AddEntry( Journal::cmd_resource_add_dep, resource_types[i], resource_names[j],
														resource_types[l], deps[k] );
				}
				else {
					LOGDEBUG( "no "+resource_types[l]+" dependencies for resource "+resource_names[j]
										+" ("+resource_types[i]+")" );
				}
			}
			delete r;
		}
	} 
}

void SCPM_helpers::DetermineResourceActions( Profile &active, Profile &dest,
																						 vector<Resource*> &to_start,
																						 vector<Resource*> &to_stop,
																						 vector<Resource*> &keep_running )
{
	Resource_helpers rh;
	vector<string> types = rh.GetResourceTypes();
	Progress::GetHandle()->Print( "Checking for resources to start or shut down",
																Progress::verbose );
	Progress::GetHandle()->Newline( Progress::verbose );
	for (unsigned int i=0; i<types.size(); i++) {
		try {
			vector<string> resources = dest.GetResources( types[i] );
			for (unsigned int j=0; j<resources.size(); j++) {
				Resource *r = rh.CreateObject( types[i], resources[j] );
				r->SetProfile( dest.GetName() );
				if ( r->Active() ) {
					if ( r->Running() ) {
						LOGDEBUG( "adding resource "+resources[j]+" ("+types[i]+") to keep_running" );
						keep_running.push_back( r );
					}
					else {
						LOGDEBUG( "adding resource "+resources[j]+" ("+types[i]+") to to_start" );
						to_start.push_back( r );
						Progress::GetHandle()->Print( "  &"+types[i]+" &"+resources[j]+" must be started",
																					Progress::verbose );
						Progress::GetHandle()->Newline( Progress::verbose );
					}
				}
				else {
						// the resource has to be shut down for the upcoming profile
						// so add it to to_start in case it is up (as it should)
					if ( r->Running() ) {
						LOGDEBUG( "adding resource "+resources[j]+" ("+types[i]+") to to_stop" );
						to_stop.push_back( r );
						Progress::GetHandle()->Print( "  &"+types[i]+" &"+resources[j]+" must be stopped",
																					Progress::verbose );
						Progress::GetHandle()->Newline( Progress::verbose );
					}
				}
			}
		}
		catch( Resource::not_available) {
				// a called resource method was not available for the current
				// resource type, we skip this resource type
		}
		catch( Resource_helpers::type_invalid ) {
			LOGWARN( "used invalid resource type, this should not happen" );
			LOGERR( "could not determine resource actions for profile" );
			throw;
		}
	}
}

void SCPM_helpers::ResolveDeps( Profile &active_profile,
																Profile &dest_profile,
																vector<Resource*> &resources,
																vector<Resource*> &affected)
{
	Resource_helpers rh;
	Progress *pr = Progress::GetHandle();
	vector<string> resource_types = rh.GetResourceTypes();
	bool resolved = false;
	unsigned int found = affected.size();
	pr->Print( "Checking for dependencies", Progress::verbose );
	pr->Newline( Progress::verbose );
	pr->SetSegmentSteps( resources.size() * resource_types.size() );
	for ( unsigned int j =0; j<resources.size(); j++) {
		try {
			for ( unsigned int i =0; i<resource_types.size(); i++) {
				pr->BarStep();
				try {
					vector<string> dep = resources[j]->GetDependencies( resource_types[i] );
					for ( unsigned int k=0; k<dep.size(); k++ ) {
						try {
							Resource *r = rh.CreateObject( resource_types[i], dep[k] );
							r->SetProfile( dest_profile.GetName() );
							if ( ! r->UpToDate() ) {
								LOGDEBUG( "resource "+resources[j]->GetName()+" )"+resources[j]->GetType()+
													" has modified dependencies" );
								affected.push_back( resources[j] );
								resolved = true;
								Progress::GetHandle()->Print( "  &"+resources[j]->GetType()+" &"
																							+resources[j]->GetName()+" must be restarted",
																							Progress::verbose );
								Progress::GetHandle()->Newline( Progress::verbose );
							}
							delete r;
							if ( resolved ) throw 1;
						}
						catch( Resource::resource_unknown ) {
								// maybe a dependend resource is not installed,
								// this is harmless
							LOGDEBUG( "dependend resource "+dep[k]+" ("+resource_types[i]+") for "
												+resources[j]->GetName()+" ("+resources[j]->GetType()
												+") not found, ignoring" );
						}
						catch( Resource_helpers::type_invalid ) {
							LOGWARN( "used invalid resource type, this should not happen" );
							LOGERR( "could not resolve dependencies" );
						}
					}
				}
				catch ( Resource::not_available ) {
						// resource has no deps, fine
				}
			}
		}
		catch( int z ) {
		}
	}
	while (found<affected.size()) {
		unsigned int affcount = affected.size();
		for ( unsigned int i=0; i<resources.size(); i++) {
				// check for all resources whether they have a dependency equal to one of
				// the (new) affected ones
			for ( unsigned int j=found; j<affcount; j++) {
					// dependend resources of affected[j] are usually not equal to affected[j]
				if (resources[i] == affected[j]) continue;
				vector<string> dep;
				try {
					dep = resources[i]->GetDependencies( affected[j]->GetType() );
				}
				catch( Resource::not_available ) {
						// resource has no deps, fine
					continue;
				}
				for ( unsigned int k=0; k<dep.size(); k++ ) {
					if ( dep[k] == affected[j]->GetName()) {
							// found a candidate
						bool found=false;
						for (unsigned int l=0; l<affected.size(); l++)
							if ( (resources[i]->GetName() == affected[l]->GetName()) &&
									 (resources[i]->GetType() == affected[l]->GetType()) ) {
									// already in list
								found=true;
								continue;
							}
						if (!found) {
							try {
								affected.push_back(resources[i]);
								Progress::GetHandle()->Print( "  &"+resources[i]->GetType()+" &"
																							+resources[i]->GetName()+" has to be restarted",
																							Progress::verbose );
								Progress::GetHandle()->Newline( Progress::verbose );
							}
							catch( Resource::resource_unknown ) {
									// maybe a dependend resource is not installed,
									// this is harmless
								LOGDEBUG( "dependend resource "+dep[k]+" ("+resource_types[i]+") for "
													+resources[j]->GetName()+" ("+resources[j]->GetType()
													+") not found, ignoring" );
							}
							catch( Resource_helpers::type_invalid ) {
								LOGWARN( "used invalid resource type, this should not happen" );
								LOGERR( "could not resolve dependencies" );
							}
						}
					}	 
				}
			}
		}
		found=affected.size();
	}
}

void SCPM_helpers::OrderResources( vector<Resource*> &resources, bool stop )
{
	unsigned int size = resources.size();
	if ( size == 0 ) return;
	for ( unsigned int i=0; i<size-1; i++) {
		vector<Resource*>::iterator it = resources.begin();
		for ( unsigned int j=0; j<=i; j++) it++;
		while( it != resources.end() ) {
			if ( *it == resources[i] ) {
				resources.erase( it );
				size--;
			}
			else it++;
		}
	}
		// dupes should be killed, now try to order the remaining
		// resources in a useful way
	try {
		vector<int> prio;
		size = resources.size();
		for ( unsigned int i=0; i<size; i++) {
			if ( !stop ) prio.push_back( resources[i]->GetStartPriority() );
			else prio.push_back( resources[i]->GetStopPriority() );
		}
			// doing a bubble sort
		for ( unsigned int i=0; i<size-1; i++ ) {
			for ( unsigned int j=0; j<size-i-1; j++) {
				if ( prio[j] > prio[j+1] ) {
					int priobuf = prio[j];
					prio[j] = prio[j+1];
					prio[j+1] = priobuf;
					Resource *resbuf = resources[j];
					resources[j] = resources[j+1];
					resources[j+1] = resbuf;
				}
			}
		}
	}
	catch( Resource::not_available ) {
			// Resource type does not support priorities, skip sorting
	}
}

const vector<string> SCPM_helpers::GetResourceTypes( )
{
	static vector<string> types;
	try {
		types = db->ResourceGetTypes();
		return types;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "could not determine resource types" );
	return types;
}

const vector<string> SCPM_helpers::GetResourceNames( string &resource_type )
{
	static vector<string> names;
	try {
		names = db->ResourceGetNames( resource_type );
		return names;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	LOGERR( "could not determine resource names for resource type &"+resource_type );
	return names;
}

void SCPM_helpers::CopyResources( const string &src_profile, const string &dest_profile )
{
	SCDB *db = SCDB::GetHandle();

	try {
		vector<string> resource_types = db->ResourceGetTypes();
		for( unsigned int i=0; i<resource_types.size(); i++ ) {
			vector<string> resource_names = db->ResourceGetNames( resource_types[i] );
			for( unsigned int j=0; j<resource_names.size(); j++ )
				jrnl->AddEntry( Journal::cmd_file_copy, src_profile, dest_profile,
												resource_types[i], resource_names[j] );
		}
		return;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database error ("+e.reason+")" );
	}
	LOGERR( "could not copy resources (database error)" );
}

void SCPM_helpers::ApplyAll( const string &resource_type, const string &resource_name )
{
	SCDB *db = SCDB::GetHandle();

	try {
		vector<string> profiles = db->GetProfiles();
		for( unsigned int i=0; i<profiles.size(); i++ ) {
			if ( profiles[i] == db->GetActiveProfile() ) continue;
			jrnl->AddEntry( Journal::cmd_resource_save, resource_type, resource_name, profiles[i] );
		}
		return;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error ("+e.reason+")" );
	}
	LOGERR( "could not save resource &"+resource_name+" (&"+resource_type+")" );
}

void SCPM_helpers::PatchAll( const string &resource_type, const string &resource_name )
{
	SCDB *db = SCDB::GetHandle();

	try {
		vector<string> profiles = db->GetProfiles();
		for( unsigned int i=0; i<profiles.size(); i++ ) {
			if ( profiles[i] == db->GetActiveProfile() ) continue;
			jrnl->AddEntry( Journal::cmd_resource_patch, resource_type, resource_name, profiles[i] );
		}
		return;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error ("+e.reason+")" );
	}
	LOGERR( "could not save resource &"+resource_name+" (&"+resource_type+")" );
}

bool SCPM_helpers::InstallationChanged( )
{
	SCDB *db = SCDB::GetHandle();
	
	char current_system[50];
	ifstream in( "/etc/SuSE-release" );
	if ( in.bad() ) {
		LOGWARN( "could not determine installed system" );
		return true;
	}
	in.getline( current_system, 49 );
	in.close();
	try {
		if ( string(current_system) == db->GetStatusKey( "system" ) )
			return false;
		else
			return true;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error ("+e.reason+")" );
		LOGWARN( "could not determine installed system" );
	}
	return true;
}

void SCPM_helpers::SaveInstallInfo( )
{
	SCDB *db = SCDB::GetHandle();
	
	char current_system[50];
	ifstream in( "/etc/SuSE-release" );
	if ( in.bad() ) {
		LOGWARN( "could not determine installed system" );
	}
	in.getline( current_system, 49 );
	in.close();
	try {
		db->SetStatusKey( "system", current_system );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error ("+e.reason+")" );
		LOGWARN( "could save installed system info" );
	}
}

string &SCPM_helpers::GetGroupDescription( const string &group )
{
	static string descr;
	SCPM_conf *conf = SCPM_conf::GetHandle();
	string fname = conf->GetUserGroupDir()+"/"+group;
	struct stat statbuf;
	if ( stat( fname.c_str(), &statbuf ) != 0 ) {
		fname = conf->GetGroupDir()+"/"+group;
		if ( stat( fname.c_str(), &statbuf ) != 0 ) {
			descr="no such resource group";
			return descr;
		}
	}
	ifstream in( fname.c_str() );
	char buf[255];
	try {
		while ( in.getline( buf, 254 ) ) {
			Tokenizer toker(buf);
			if ( toker.GetToken() == "description" ) {
				descr = toker.GetToken();
				try {
					for (;;) descr += " "+toker.GetToken();
				}
				catch( Tokenizer::end_of_tokens ) {
					descr = LOGTRANS( descr );
				}
				return descr;
			}
		}
	}
	catch( Tokenizer::end_of_tokens ) {
	}
	descr.clear();
	return descr;
}

bool SCPM_helpers::GroupActive( const string &group )
{
	ifstream in( "/var/lib/scpm/active_groups" );
	string s;
	while ( in >> s ) if ( s == group ) return true;
	return false;
}

bool SCPM_helpers::GroupDeleted( const string &group )
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	ifstream group_file( (conf->GetUserGroupDir()+"/"+group).c_str() );
	if ( group_file.bad() ) return false;
	string s;
	group_file >> s;
	group_file.close();
	if ( s == "DELETED" ) return true;
	return false;
}

void SCPM_helpers::WriteBootInfo( )
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	if ( !conf->WriteBootConfig() ) return;
	vector<string> profiles = this->GetAllProfiles();
	string active = this->GetActiveProfile();

	string profstr = conf->GetLibDir()+"/write_boot_message "+active;

	for ( unsigned int i=0; i<profiles.size(); i++ )
		profstr += " "+profiles[i];

	Executor e( profstr );
	vector<string> out;
	vector<string> err;
	try {
		if ( e.Run( out, err ) != 0 )
			LOGWARN( "Writing boot information failed" );
	}
	catch( Executor::exec_failed ) {
		LOGDEBUG( "execution of '"+profstr+"' failed" );
		if ( !err.empty() ) {
			LOGDEBUG( "error messages" );
			for (unsigned int i=0; i<err.size(); i++)
				LOGDEBUG( err[i] );
		}
		LOGWARN( "Writing boot information failed" );
	}
}

void SCPM_helpers::ClearBootInfo( )
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	string cmd = conf->GetLibDir()+"/write_boot_message clear";

	Executor e( cmd );
	vector<string> out;
	vector<string> err;
	try {
		if ( e.Run( out, err ) != 0 )
			LOGWARN( "Clearing boot information failed" );
	}
	catch( Executor::exec_failed ) {
		LOGDEBUG( "execution of '"+cmd+"' failed" );
		if ( !err.empty() ) {
			LOGDEBUG( "error messages" );
			for (unsigned int i=0; i<err.size(); i++)
				LOGDEBUG( err[i] );
		}
		LOGWARN( "Clearing boot information failed" );
	}
}

void SCPM_helpers::RunScript( const string &scriptfile )
{
	Script s( scriptfile );
	s.Exec();
}

void SCPM_helpers::RestoreResources( const string &profile, const vector<pair<string, string> > &resources )
{
	Resource_helpers rh;
	unsigned int i;
	try {
		for (i=0; i<resources.size(); i++) {
			Resource *r = rh.CreateObject( resources[i].first, resources[i].second );
			r->SetProfile( profile );
			r->Restore();
		}
	}
	catch( Resource_helpers::type_invalid ) {
		LOGERR( "resource type &"+resources[i].first+" unknown" );
	}
	catch ( Resource::not_available ) {
		LOGDEBUG( "called resource handler not available (type "+
							resources[i].first+")" );
	}
}

void SCPM_helpers::SaveResources( const string &profile, const vector<pair<string, string> > &resources )
{
	Resource_helpers rh;
	unsigned int i;
	try {
		for (i=0; i<resources.size(); i++) {
			Resource *r = rh.CreateObject( resources[i].first, resources[i].second );
			r->SetProfile( profile );
			if ( r->Modified() )
				r->Save();
		}
	}
	catch( Resource_helpers::type_invalid ) {
		LOGERR( "resource type &"+resources[i].first+" unknown" );
	}
	catch ( Resource::not_available ) {
		LOGDEBUG( "called resource handler not available (type "+
							resources[i].first+")" );
	}
}

void SCPM_helpers::UpdateResources(	 )
{
	Resource_helpers rh;
	SCDB *db = SCDB::GetHandle();
	Progress *pr = Progress::GetHandle();
	
	try {
		string active_profile = db->GetActiveProfile();
		vector<string> resource_types = db->ResourceGetTypes();
		vector<string> profiles = db->GetProfiles();
		
		for (unsigned int i=0; i<resource_types.size(); i++) {
			try {
				vector<string> resource_names = db->ResourceGetNames( resource_types[i] );
				for (unsigned int j=0; j<resource_names.size(); j++) {
					Resource *r = rh.CreateObject( resource_types[i], resource_names[j] );
						// maybe resource has system installed alternative (.rpmorig)
						// tell that to the user
					if (r->HasSystemAlternative()) {
						LOGINFO( "resource "+resource_names[j]+" ("+resource_types[i]+") "+
										 "has alternative(s) data" );
						pr->Print( "Resource &"+resource_names[j]+" (&"+resource_types[i]
											 +") has .rpmorig files; you may want to have a look at it" );
						pr->Newline( Progress::verbose );
					}
						// first check: look whether resource is modified at all in
						// relationship to the one in the active profile
					r->SetProfile( active_profile );
					if ( r->Modified() ) {
						pr->Print( "Updating resource &"+resource_names[j]+" (&"+resource_types[i]+")" );
						for (unsigned int k=0; k<profiles.size(); k++) {
								// active profile is done at last
							if (profiles[k] == active_profile) continue;
							r->SetProfile( profiles[k] );

								// resource was backup by system update (.rpmsave)
								// backup the resource and just save the new version to the profile
							if (r->HasSystemBackup()) {
								LOGDEBUG( "resource "+resource_names[j]+" ("+resource_types[i]+") "+
													"has backup data, updating resource" );
								pr->Newline( Progress::verbose );
								pr->Print( "  Resource has .rpmsave files; saving to profile &"
													 +profiles[k]+" (backup will be made)" );
								r->Save();
								LOGINFO( "saved resource "+resource_names[j]+" ("+resource_types[i]+
												 ") to profile "+profiles[k] );
							}
							else {
									// resource was replaced by system update
									// this may mean that rpm has replaced an untouched config file
									// with the supplied version
									// that resource in other profiles may be touched, so we
									// create a backup from the current version in case
									// they do differ from the version in the active profile
									// in all other cases, a backup should not be needed
								if ( r->ModifiedInProfile( active_profile ) ) {
									LOGDEBUG( "resource "+resource_names[j]+" ("+resource_types[i]+") "+
														"differs in active profile and in profile "+profiles[k] );
									LOGDEBUG( "trying to patch" );
									pr->Newline( Progress::verbose );
									pr->Print( "  Resource in modified in profile &"+profiles[k] );
									if ( !r->Patch( active_profile ) ) {
										LOGDEBUG( "could not patch resource "+resource_names[j]+" ("+resource_types[i]+") "+
															"saving it to profile "+profiles[k] );
										r->Save();
										pr->Print( "patch failed; resource saved (backup made)" );
										LOGINFO( "saved resource "+resource_names[j]+" ("+resource_types[i]+
														 ") to profile "+profiles[k] );
									}
									else {
										pr->Print( "resource patched (backup made)" );
										LOGINFO( "patched resource "+resource_names[j]+" ("+resource_types[i]+
														 ") of profile "+profiles[k] );
									}
								}
								else {
										// resource modified, but is same is same in this profile and the active on
										// just saving it
									LOGDEBUG( "resource "+resource_names[j]+" ("+resource_types[i]+") "+
														"is same in active profile and in profile "+profiles[k] );
									LOGDEBUG( "just saving" );
									pr->Newline( Progress::verbose );
									pr->Print( "  Resource is modified in profile &"+profiles[k]+"; resource saved (backup made)" );
									r->Save();
								}
							}
						}
							// now do this also for the current profile
						r->SetProfile( active_profile );
						pr->Newline( Progress::verbose );
						pr->Print( "  Resource is modified in profile &"+active_profile+", resource saved (backup made)" );
						r->Save();
						LOGINFO( "saved resource "+resource_names[j]+" ("+resource_types[i]+
										 ") to profile "+active_profile );
						pr->Newline();
					}
				}
			}
			catch( Resource::not_available ) {
				LOGDEBUG( "resource type "+resource_types[i]+" does not support update, skipping" );
			}
		}
	}
	catch( Resource_helpers::type_invalid ) {
			// huh?
		LOGERR( "Bug! invalid resource type exception" );
	}
	catch( Resource::resource_unknown ) {
			// huh?
		LOGERR( "Bug! resource unknown exception" );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error ("+e.reason+")" );
		LOGERR( "could not determine resources to be updated" );
	}
}

void SCPM_helpers::CheckResources( vector<pair<string,string> > &resources )
{
	if ( resources.empty() ) return;
	SCDB *db = SCDB::GetHandle();
	Progress *pr = Progress::GetHandle();
	vector<pair<string,string> >::iterator it = resources.begin();
	do {
		if ( !db->ResourceExists( it->second, it->first ) ) {
			LOGWARN( "Can not update resource "+it->second+" ("+
							 it->first+") (does not exist)" );
			pr->Print( "WARNING: resource &"+it->second+" (&"+it->first
								 +") does not exist; ignoring", Progress::verbose );
			pr->Newline( Progress::verbose );
			resources.erase( it );
		}
		else it++;
	} while( it != resources.end() );
}

void SCPM_helpers::AddBackup( const vector<pair<string,string> > &resources )
{
	vector<Resource*> res;
	Progress *pr = Progress::GetHandle();
	if ( resources.empty() ) return;

	pr->Print("Adding backup files");
	jrnl->New();
	for ( unsigned int i=0; i<resources.size(); i++ ) {
		try {
			if ( db->ResourceExists( resources[i].second, resources[i].first ) ) {
				jrnl->AddEntry(Journal::cmd_resource_add_backup, resources[i].first, resources[i].second, this->GetActiveProfile());
			}
			else {
				LOGWARN( "resource "+resources[i].second+" ("+resources[i].first+
								 " ) does not exist, ignoring" );
				pr->Print( "WARNING: resource &"+resources[i].second+" (&"+resources[i].first
									 +") does not exist; ignoring", Progress::verbose );
				pr->Newline( Progress::verbose );
			}
		}
		catch( Resource::resource_unknown ) {
			LOGERR( "unknown resource &"+resources[i].second+" (&"+resources[i].first+")" );
		}
	}
	jrnl->Ready("nomagic");
	pr->SetBar(100);
	this->ProcessJournal();
	pr->Newline();
}

void SCPM_helpers::RemoveBackup( const vector<Resource*> &resources, const vector<string> &versions )
{
	Progress *pr = Progress::GetHandle();
	if (resources.size()) {
		pr->Print( "Removing backup files" );
		jrnl->New();
		for (unsigned int i=0; i<resources.size(); i++) {
			jrnl->AddEntry( Journal::cmd_resource_remove_backup, resources[i]->GetType(), 
											resources[i]->GetName(), resources[i]->GetProfile(), versions[i] );
			
			delete resources[i];
		}
		jrnl->Ready( "nomagic" );
		pr->SetBar(100);
		this->ProcessJournal();
		pr->Newline( );
	}
}

void SCPM_helpers::RestoreBackup( const vector<Resource*> &resources, const vector<string> &versions )
{
	Progress *pr = Progress::GetHandle();
	if (resources.size() && resources.size()==versions.size() ) {
		pr->Print("Restoring resources" );
		jrnl->New();
		for (unsigned int i=0; i<resources.size(); i++) {
			jrnl->AddEntry( Journal::cmd_resource_restore_backup, resources[i]->GetType(),
											resources[i]->GetName(), resources[i]->GetProfile(), versions[i] );
		}
		jrnl->Ready( "nomagic" );
		pr->SetBar(100);
		this->ProcessJournal();
		pr->Newline();
	}
}

vector<Resource*> &SCPM_helpers::GetResourcesWithBackup( string profile, bool verbose )
{
	static vector<Resource*> resources;
	Progress *pr = Progress::GetHandle();
	resources.clear();
	if (profile.empty()) profile=this->GetActiveProfile();
	try {
		vector<string> profiles;
		if ( profile == "all" )
			profiles=this->GetAllProfiles();
		else 
			profiles.push_back(profile);
		vector<string> resource_types = db->ResourceGetTypes();
		Resource_helpers rh;
		for (unsigned int i=0; i<resource_types.size(); i++) {
			try {
				vector<string> resource_names = db->ResourceGetNames( resource_types[i] );
				for (unsigned int j=0; j<resource_names.size(); j++) {
					try {
						for (unsigned int k=0; k<profiles.size(); k++) {
							Resource *r = rh.CreateObject( resource_types[i], resource_names[j] );
							r->SetProfile( profiles[k] );
							if ( r->HasBackup() ) {
								LOGDEBUG( "Resource &"+resource_names[j]+" ("+resource_types[i]+
													") has backup data in profile "+profiles[k] );
								if ( verbose ) {
									pr->Newline( Progress::verbose );
									pr->Print( "  Resource &"+resource_names[j]+" (&"+resource_types[i]+
														 ") has backup data in profile &"+profiles[k], Progress::verbose );
								}
								resources.push_back( r );
							}
							else
								delete r;
						}
					}
					catch( Resource::resource_unknown ) {
						LOGERR( "unknown resource &"+resource_names[j]+" (&"+resource_types[i]+")" );
					}
				}
			}
			catch( Resource::not_available ) {
				LOGDEBUG( "resource type "+resource_types[i]+" does not support backups" );
			}
			catch( Resource_helpers::type_invalid ) {
				LOGERR( "unknown resource type &"+resource_types[i] );
			}
		}
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error ("+e.reason+")" );
		LOGERR( "could not determine resources with backup data" );
	}
	return resources;
}

vector<Resource*> &SCPM_helpers::GetResourcesByName( const string &profile,
																										 vector<pair<string,string> > resources )
{
	static vector<Resource*> res;
	Resource_helpers rh;
	res.clear();
	for (unsigned int i=0; i<resources.size(); i++) {
		vector<string> profiles;
		if (profile == "all") 
			profiles = this->GetAllProfiles();
		else
			profiles.push_back( this->GetActiveProfile() );
		for (unsigned int j=0; j<profiles.size(); j++) {
			try {
				Resource *r = rh.CreateObject( resources[i].first, resources[i].second );
				r->SetProfile( profiles[j] );
				res.push_back( r );
			}
			catch( Resource::resource_unknown ) {
				LOGERR( "unknown resource &"+(resources[j]).second+" (&"+(resources[i]).first+")" );
			}
			catch( Resource_helpers::type_invalid ) {
				LOGERR( "unknown resource type &"+(resources[i]).first );
			}
		}
	}
	return res;
}

void SCPM_helpers::PerformCommand( Journal::entry &e )
{
	SCDB *db = SCDB::GetHandle();
	SCPM_conf *conf = SCPM_conf::GetHandle();
	Progress *p = Progress::GetHandle();
	Resource_helpers rh;
	Resource *r;
	int argcount = e.args.size();
	string * version = NULL;
	try {
		switch (e.command) {
			case Journal::cmd_set_global_key:
				if (argcount!=2) LOGERR( "journal corrupt" );
				db->SetStatusKey( e.args[0], e.args[1] );
				break;
			case Journal::cmd_save_install_info:
				this->SaveInstallInfo();
				break;
			case Journal::cmd_write_boot_info:
				this->WriteBootInfo();
				break;
			case Journal::cmd_clear_boot_info:
				this->ClearBootInfo();
				break;
			case Journal::cmd_resource_save:
				if (argcount!=3) LOGERR( "journal corrupt" );
				r=rh.CreateObject( e.args[0], e.args[1] );
				r->SetProfile(e.args[2]);
				if (e.replay) {
						// only database needs to be replayed
					try {
						if (!r->Refresh()) {
							delete r;
							LOGERR( "could not refresh resource &"+e.args[1]+" (&"+e.args[0]+")" );
						}
					}
					catch( Resource::not_available ) {
							// ok, nothing to do
					}
				}
				else {
					p->Newline( Progress::verbose );
					p->Print( " ", Progress::verbose );
					p->Print( "saving &"+e.args[1]+" (&"+e.args[0]+") to profile &"+e.args[2],
										Progress::verbose );
					if (!r->Save()) p->Print("[failed]", Progress::verbose);
				}
				delete r;
				break;
			case Journal::cmd_resource_patch:
				if (argcount!=3) LOGERR( "journal corrupt" );
				r=rh.CreateObject( e.args[0], e.args[1] );
				r->SetProfile(e.args[2]);
				if (e.replay) {
						// only md5 sum in database needs to be replayed
					try {
						if (!r->Refresh()) {
							delete r;
							LOGERR( "could not refresh resource &"+e.args[1]+" (&"+e.args[0]+")" );
						}
					}
					catch( Resource::not_available ) {
							// ok, nothing to do
					}
				}
				else {
					p->Newline( Progress::verbose );
					p->Print( " ", Progress::verbose );
					p->Print( "patching &"+e.args[1]+" (&"+e.args[0]+") of profile &"+e.args[2]
										, Progress::verbose );
					if (!r->Patch( this->GetActiveProfile() ) )
						p->Print( "[failed]", Progress::verbose );
				}
				delete r;
				break;
			case Journal::cmd_resource_restore:
				if (argcount!=3) LOGERR( "journal corrupt" );
				if (e.replay) {
					LOGDEBUG( "cmd_resource_restore does not need to be replayed" );
					break;
				}
				r=rh.CreateObject( e.args[0], e.args[1] );
				p->Newline( Progress::verbose );
				p->Print( " ", Progress::verbose );
				p->Print( "restoring &"+e.args[1]+" (&"+e.args[0]+")", Progress::verbose );
				r->SetProfile(e.args[2]);
				if (!r->Restore()) p->Print("[failed]", Progress::verbose);
				delete r;
				break;
			case Journal::cmd_resource_stop:
				if (argcount!=2) LOGERR( "journal corrupt" );
				if (e.replay) {
					LOGDEBUG( "cmd_resource_stop does not need to be replayed" );
					break;
				}
				r=rh.CreateObject( e.args[0], e.args[1] );
				p->Newline( Progress::verbose );
				p->Print( " ", Progress::verbose );
				p->Print( "stopping &"+e.args[1]+" (&"+e.args[0]+")", Progress::verbose );
				if (r->Stop())
					p->Print( "[ok]", Progress::verbose );
				else
					p->Print( "[failed]", Progress::verbose );
				delete r;
				break;
			case Journal::cmd_resource_start:
				if (argcount!=2) LOGERR( "journal corrupt" );
				if (e.replay) {
					LOGDEBUG( "cmd_resource_start does not need to be replayed" );
					break;
				}
				r=rh.CreateObject( e.args[0], e.args[1] );
				p->Newline( Progress::verbose		 );
				p->Print( " ", Progress::verbose );
				p->Print( "starting &"+e.args[1]+" (&"+e.args[0]+")", Progress::verbose );
				if (r->Start())
					p->Print( "[ok]", Progress::verbose );
				else
					p->Print( "[failed]", Progress::verbose );
				delete r;
				break;
			case Journal::cmd_resource_add:
				if (argcount!=2) LOGERR( "journal corrupt" );
				db->ResourceAdd( e.args[1], e.args[0] );
				try {
					p->Newline( Progress::verbose		 );
					p->Print( " ", Progress::verbose );
					p->Print( "adding resource &"+e.args[1]+" (&"+e.args[0]+")", Progress::verbose );
					r=rh.CreateObject( e.args[0], e.args[1] );
					r->UpdateData();
					r->WriteData();
					delete r;
					break;
				}
				catch( Resource_helpers::type_invalid ) {
					LOGWARN( "used invalid resource type, this should not happen" );
					break;
				}
				catch( Resource::not_available ) {
						// resource probably does not support UpdateData or something
						// this is harmless
					delete r;
					break;
				}
				LOGERR( "could not create resource &"+e.args[1]+" (&"+e.args[0]+")" );
				break;
			case Journal::cmd_resource_add_type:
				if (argcount!=1) LOGERR( "journal corrupt" );
				db->ResourceAddType( e.args[0] );
				break;
			case Journal::cmd_resource_delete:
				if (argcount!=2) LOGERR( "journal corrupt" );
				db->ResourceDelete( e.args[1], e.args[0] );
				break;
			case Journal::cmd_resource_add_dep:
				if (argcount!=4) LOGERR( "journal corrupt" );
				db->ResourceAddDep( e.args[1], e.args[0], e.args[3], e.args[2] );
				break;
			case Journal::cmd_resource_clean_deps:
				if (argcount!=2) LOGERR( "journal corrupt" );
				db->ResourceCleanDeps( e.args[1], e.args[0] );
				break;
			case Journal::cmd_resource_add_data:
				if (argcount!=4) LOGERR( "journal corrupt" );
				db->ResourceAddData( e.args[1], e.args[0], e.args[2], e.args[3] );
				break;
			case Journal::cmd_resource_clear_data:
				if (argcount!=2) LOGERR( "journal corrupt" );
				db->ResourceClearData( e.args[1], e.args[0] );
				break;
			case Journal::cmd_resource_create:
			case Journal::cmd_resource_drop:
				LOGERR( "journal command not implemented" );
				break;
			case Journal::cmd_resource_add_backup:
				if (argcount<3 || argcount >4) LOGERR( "journal corrupt" );
				if (e.replay) {
					LOGDEBUG( "cmd_resource_backup does not need to be replayed" );
					break;
				}
				r=rh.CreateObject( e.args[0], e.args[1] );
				r->SetProfile(e.args[2]);
				p->Newline( Progress::verbose );
				p->Print( " ", Progress::verbose );
				p->Print( " &"+r->GetName()+" (&"+r->GetType()+") to profile &"+
								 r->GetProfile(), Progress::verbose );
				if (!r->AddBackup()) p->Print("[failed]", Progress::verbose);
				delete r;
				break;
			case Journal::cmd_resource_restore_backup:
				if (argcount<3 || argcount >5) LOGERR( "journal corrupt" );
				if (e.replay) {
					LOGDEBUG( "cmd_resource_restore does not need to be replayed" );
					break;
				}
				r=rh.CreateObject( e.args[0], e.args[1] );
				p->Newline( Progress::verbose );
				p->Print( " ", Progress::verbose );
				p->Print( "restoring backup of &"+e.args[1]+" (&"+e.args[0]+")", Progress::verbose );
				r->SetProfile(e.args[2]);
				if (e.args.size() == 4)
					version = new string(e.args[3]);
				else
					version = NULL;
				if (!r->RestoreBackup(version)) p->Print("[failed]", Progress::verbose);
				if (version) delete version;
				delete r;
				break;
			case Journal::cmd_resource_remove_backup:
				if (argcount<3 || argcount >5) LOGERR( "journal corrupt" );
				if (e.replay) {
					LOGDEBUG( "cmd_resource_remove_backup does not need to be replayed" );
					break;
				}
				r=rh.CreateObject( e.args[0], e.args[1] );
				r->SetProfile(e.args[2]);
				if (e.args.size() == 4)
					version = new string(e.args[3]);
				else
					version = NULL;
				try {
					p->Newline( Progress::verbose );
					p->Print( " &"+r->GetName()+" (&"+r->GetType()+") from profile &"
										 +r->GetProfile(), Progress::verbose );
					if (!r->RemoveBackup( version ) ) {
						p->Print( " ", Progress::verbose );
						p->Print( "[failed]", Progress::verbose );
					}
				}
				catch( Resource::not_available ) {
					// should not happen
					LOGWARN( "resource type "+r->GetType()+" does not support backup removal");
					p->Print( " ", Progress::verbose );
					p->Print( "[failed]", Progress::verbose );
				}			
				break;
			case Journal::cmd_profile_add:
				if (argcount!=1) LOGERR( "journal corrupt" );
				db->ProfileAdd( e.args[0] );
				break;
			case Journal::cmd_profile_delete:
				if (argcount!=1) LOGERR( "journal corrupt" );
				db->ProfileDelete( e.args[0] );
				break;
			case Journal::cmd_profile_add_resource:
				if (argcount!=3) LOGERR( "journal corrupt" );
				db->ProfileAddResource( e.args[0], e.args[2], e.args[1] );
				break;
			case Journal::cmd_profile_delete_resource:
				if (argcount!=3) LOGERR( "journal corrupt" );
				db->ProfileDeleteResource( e.args[0], e.args[2], e.args[1] );
				break;
			case Journal::cmd_profile_clear_resources:
				if (argcount!=1) LOGERR( "journal corrupt" );
				db->ProfileClearResources( e.args[0] );
				break;
			case Journal::cmd_profile_set_key:
				if (argcount!=3) LOGERR( "journal corrupt" );
				db->ProfileSetKey( e.args[0], e.args[1], e.args[2] );
				break;
			case Journal::cmd_profile_set_script:
				if (argcount!=3) LOGERR( "journal corrupt" );
				db->ProfileSetScript( e.args[0], e.args[1], e.args[2] );
				break;
			case Journal::cmd_file_save:
			case Journal::cmd_file_restore:
				LOGERR( "journal command not implemented" );
				break;
			case Journal::cmd_file_copy:
				if (argcount!=4) LOGERR( "journal corrupt" );
				p->Newline( Progress::verbose );
				p->Print( " ", Progress::verbose );
				p->Print( "copying resource &"+e.args[3]+" (&"+e.args[2]+")", Progress::verbose );
				db->FileCopy( e.args[0], e.args[1], e.args[2], e.args[3] );
				break;
			case Journal::cmd_file_delete:
				if (argcount!=3) LOGERR( "journal corrupt" );
				db->FileDelete( e.args[0], e.args[1], e.args[2] );
				break;
			case Journal::cmd_call:
				if (argcount!=1) LOGERR( "journal corrupt" );
				this->Call( e.args[0] );
				break;
			case Journal::cmd_script:
				if (argcount!=1) LOGERR( "journal corrupt" );
				this->SetScriptEnv();
				this->RunScript( e.args[0] );
				break;
			case Journal::cmd_reload_db:
				db->DropDB();
				this->LoadSCDB();
				break;
			case Journal::cmd_clear_all:
			{
				pid_t pid = fork();
				if (pid == -1) {
					LOGWARN("could not cleanup profile dir (fork failed)");
					break;
				}
				if (pid==0) {
					setuid(0);
					execlp( "rm", "rm", "-rf", conf->GetProfileDir().c_str(), NULL );
				}
				else wait(NULL);
				mkdir( conf->GetProfileDir().c_str(), 0755 );
				break;
			}
			case Journal::cmd_message:
				if (argcount!=1) LOGERR( "journal corrupt" );
				p->Newline();
				p->Print( e.args[0] );
				break;
			default: 
				LOGERR( "unknown journal command. This is a bug!" );
		}
	}
	catch( SCDB::db_action_failed &e ) {
		LOGERR( "database error (&"+e.reason+")" );
	}
	catch( Resource_helpers::type_invalid ) {
		LOGERR( "unknown type used, this is a bug!" );
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "tried to access non existing file" );
		if (e.command == Journal::cmd_file_delete)
			LOGWARN( "tried to delete missing file, ignoring" );
		else
			LOGERR( "file operation failed" );
	}
	catch( SCDB::file_exists ) {
		LOGWARN( "tried to create an existing file" );
		LOGERR( "file operation failed" );
	}
}

void SCPM_helpers::RollbackCommand( Journal::entry &e )
{
	SCDB *db = SCDB::GetHandle();
	Progress *p = Progress::GetHandle();
	Resource_helpers rh;
	Resource *r;
	int argcount = e.args.size();
	try {
		switch (e.command) {
			case Journal::cmd_set_global_key:
			case Journal::cmd_save_install_info:
			case Journal::cmd_write_boot_info:
			case Journal::cmd_clear_boot_info:
					// no rollback needed or possible
				break;
			case Journal::cmd_resource_save:
			case Journal::cmd_resource_patch:
				if (argcount!=3) LOGERR( "journal corrupt" );
				r=rh.CreateObject( e.args[0], e.args[1] );
				r->SetProfile(e.args[2]);
				p->Newline( Progress::verbose );
				p->Print( " ", Progress::verbose );
				p->Print( "rolling back &"+e.args[1]+" (&"+e.args[0]+") of profile &"+e.args[2]
									, Progress::verbose );
				r->RestoreBackup();
				break;
			case Journal::cmd_resource_restore:
				LOGDEBUG( "cmd_resource_restore can not be rolled back" );
				p->Newline( Progress::verbose );
				p->Print( "  previous restore of resource &"+e.args[1]+" (&"+e.args[0]+
									") cannot be undone (yet)", Progress::verbose );
				break;
			case Journal::cmd_resource_stop:
				if (argcount!=2) LOGERR( "journal corrupt" );
				r=rh.CreateObject( e.args[0], e.args[1] );
				p->Newline( Progress::verbose );
				p->Print( " ", Progress::verbose );
				p->Print( "starting &"+e.args[1]+" (&"+e.args[0]+")", Progress::verbose );
				r->SetProfile(e.args[2]);
				if (r->Start())
					p->Print( "[ok]", Progress::verbose );
				else
					p->Print( "[failed]", Progress::verbose );
				delete r;
				break;
			case Journal::cmd_resource_start:
				if (argcount!=2) LOGERR( "journal corrupt" );
				r=rh.CreateObject( e.args[0], e.args[1] );
				p->Newline( Progress::verbose		 );
				p->Print( " ", Progress::verbose );
				p->Print( "stopping &"+e.args[1]+" (&"+e.args[0]+")", Progress::verbose );
				r->SetProfile(e.args[2]);
				if (r->Stop())
					p->Print( "[ok]", Progress::verbose );
				else
					p->Print( "[failed]", Progress::verbose );
				delete r;
				break;
			case Journal::cmd_resource_add:
			case Journal::cmd_resource_add_type:
			case Journal::cmd_resource_delete:
			case Journal::cmd_resource_add_dep:
			case Journal::cmd_resource_clean_deps:
			case Journal::cmd_resource_add_data:
			case Journal::cmd_resource_clear_data:
			case Journal::cmd_resource_create:
			case Journal::cmd_resource_drop:
					// no need, database only
				break;
			case Journal::cmd_resource_restore_backup:
				LOGDEBUG( "cmd_resource_restore_backup can not be rolled back" );
				break;
			case Journal::cmd_profile_add:
			case Journal::cmd_profile_delete:
			case Journal::cmd_profile_add_resource:
			case Journal::cmd_profile_delete_resource:
			case Journal::cmd_profile_clear_resources:
			case Journal::cmd_profile_set_key:
			case Journal::cmd_profile_set_script:
					// no need, database only
				break;
			case Journal::cmd_file_save:
			case Journal::cmd_file_restore:
				LOGERR( "journal command not implemented" );
				break;
			case Journal::cmd_file_copy:
				LOGDEBUG( "cmd_file_copy can not be rolled back" );
				break;
			case Journal::cmd_file_delete:
				LOGDEBUG( "cmd_file_delete can not be rolled back" );
				break;
			case Journal::cmd_call:
				LOGDEBUG( "cmd_call can not be rolled back" );
				break;
			case Journal::cmd_script:
				LOGDEBUG( "cmd_script can not be rolled back" );
				break;
			case Journal::cmd_reload_db:
				LOGDEBUG( "cmd_reload_db can not be rolled back" );
				break;
			case Journal::cmd_clear_all:
				LOGDEBUG( "cmd_reload_db can not be rolled back" );
				break;
			case Journal::cmd_message:
				break;
			default: 
				LOGERR( "unkown journal command. This is a bug!" );
		}
	}
	catch( SCDB::db_action_failed &e ) {
		db->Dump(cerr);
		LOGERR( "database error (&"+e.reason+")" );
	}
	catch( Resource_helpers::type_invalid ) {
		LOGERR( "unknown type used, this is a bug!" );
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "tried to access non existing file" );
		LOGERR( "file operation failed" );
	}
	catch( SCDB::file_exists ) {
		LOGWARN( "tried to create an existing file" );
		LOGERR( "file operation failed" );
	}
}

void SCPM_helpers::ProcessJournal( bool nowait )
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	if ( !nowait && conf->WaitCommit() ) return;
	Progress *pr = Progress::GetHandle();
	pr->SetSegmentSteps( jrnl->Steps() );
	try {
		int ix=1;
		for(;;) {
			if ( scpm_journal_abort_steps>-1 && ix>scpm_journal_abort_steps )
				LOGERR( "Journal processing aborted due to user request" );
			Journal::entry &e = jrnl->NextEntry();
			pr->BarStep();
			string debugstr;
			debugstr="cmd #"+itoa(++ix)+": "+itoa(e.command);
			for (unsigned int i=0;i<e.args.size();i++) debugstr+=" "+e.args[i];
			LOGDEBUG( debugstr );
			this->PerformCommand(e);
		}
	}
	catch( Journal::no_more_entries ) {
		LOGDEBUG( "journal done" );
			// before we delete the journal, we save the database, so our data is in sync
		this->SaveSCDB();
		jrnl->Delete();
	}
}

void SCPM_helpers::SetScriptEnv( )
{
		// FIXME: environment does not contain all information it did before journalization
	SCPM_conf *conf = SCPM_conf::GetHandle();
	::setenv( "SCPM_PROFILE", this->GetActiveProfile().c_str(), 1 );
	if ( conf->Verbose() ) ::setenv( "SCPM_VERBOSE", "yes", 1 );
	if ( conf->Debug() ) ::setenv( "SCPM_DEBUG", "yes", 1 );
	if ( conf->Force() ) ::setenv( "SCPM_FORCE", "yes", 1 );
	if ( conf->Skip() ) ::setenv( "SCPM_SKIP", "yes", 1 );
	if ( conf->Quiet() ) ::setenv( "SCPM_QUIET", "yes", 1 );
	if ( conf->BootMode() ) ::setenv( "SCPM_BOOT", "yes", 1 );	
}

void SCPM_helpers::Call( string command )
{
	Executor ex( command );
	vector<string> out;
	vector<string> err;
	try {
		if ( ex.Run( out, err ) != 0 )
			LOGWARN( "Script exitcode is nonzero" );
	}
	catch( Executor::exec_failed ) {
		LOGDEBUG( "execution of '"+command+"' failed" );
		if ( !err.empty() ) {
			LOGDEBUG( "error messages" );
			for (unsigned int i=0; i<err.size(); i++)
				LOGDEBUG( err[i] );
		}
		LOGWARN( "Script could not be executed" );
	}
}

void SCPM_helpers::SetResourceGroup( const string &name, const vector<resource_entry_t> &group,
																		 const string &description )
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	ofstream out( (conf->GetUserGroupDir()+"/"+name).c_str() );
	if ( description.size() ) out << "description " << description << "\n";
	for ( unsigned int i=0; i<group.size(); i++ ) {
		out << group[i].type << " " << group[i].name;
		if ( group[i].active == false ) out << " inactive";
		out << "\n";
	}
	out.close();
}

void SCPM_helpers::DeleteResourceGroup( const string &group )
{
	this->DeactivateResourceGroup( group );
	SCPM_conf *conf = SCPM_conf::GetHandle();
	ofstream group_file( (conf->GetUserGroupDir()+"/"+group).c_str(), ios::trunc );
	if ( group_file.bad() ) {
		LOGWARN( "could not open "+conf->GetUserGroupDir()+"/"+group+
						 "for writing" );
		LOGERR( "could not remove resource group &"+group );
	}
	group_file << "DELETED" << endl;
	group_file.close();
}

void SCPM_helpers::RenameResourceGroup( const string &group, const string &newgroup )
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	if ( rename( (conf->GetUserGroupDir()+"/"+group).c_str(), 
							 (conf->GetUserGroupDir()+"/"+newgroup).c_str()) != 0 ) {
		LOGWARN( string("rename failed, ")+strerror( errno ) );
		LOGERR( "could not rename resource group &"+group );
	}
	vector<string> groups;
	GetActiveResourceGroups( groups );
	for( unsigned int i=0; i<groups.size(); i++ ) {
		if ( groups[i] == group ) {
			groups[i] = newgroup;
			SetActiveResourceGroups( groups );
		}
	}
}

void SCPM_helpers::SetActiveResourceGroups( const vector<string> &groups )
{
	ofstream out( "/var/lib/scpm/active_groups", ios::trunc );
	for( unsigned int i=0; i<groups.size(); i++ )
		out << groups[i] << "\n";
	out.close();
}

void SCPM_helpers::GetActiveResourceGroups( vector<string> &groups )
{
	ifstream in( "/var/lib/scpm/active_groups", ios::trunc );
	if ( in.bad() ) LOGERR( "could not open active groups file" );
	groups.clear();
	for( unsigned int i=0; i<groups.size(); i++ ) {
		char buf[255];
		in.getline( buf, 254 );
		groups.push_back( buf );
	}
	in.close();
}

void SCPM_helpers::ActivateResourceGroup( const string &group, bool auto_add )
{
	SCDB *scdb = SCDB::GetHandle();
	SCDBUtil scdb_util;

	ifstream in( "/var/lib/scpm/active_groups" );
	if ( in.bad() ) LOGERR( "could not open active groups file" );
	bool found=false;
	string groupname;
	while ( in >> groupname ) {
		if ( groupname == group ) {
			found=true;
			break;
		}
	}
	in.close();
	if ( found )
		LOGWARN( "Resource group "+group+" is already active" );
	else {
		ofstream out( "/var/lib/scpm/active_groups", ios::app );
		out << group << "\n";
		out.close();
		LOGINFO( "Resource Group &"+group+" activated" );
		if ( auto_add ) {
			vector<resource_entry_t> resources;
			string foo;
			this->GetResourceGroup( group, resources, foo, false );
			for ( unsigned int i=0; i<resources.size(); i++ ) {
				try {
					if ( !scdb->ResourceExists( resources[i].name, resources[i].type ) )
						scdb_util.CreateResource( resources[i].name, resources[i].type );
				}
				catch( SCDB::db_action_failed &e ) {
					LOGERR( "could not create resource &"+resources[i].name+" (&"+e.reason+")" );
				}
				catch( SCDBUtil::operation_failed &e ) {
					LOGERR( "could not create resource &"+resources[i].name+" (&"+e.reason+")" );
				}
			}
		}
	}
}

void SCPM_helpers::DeactivateResourceGroup( const string &group, bool auto_remove )
{
	SCDB *scdb = SCDB::GetHandle();
	SCDBUtil scdb_util;

	ifstream in( "/var/lib/scpm/active_groups" );
	if ( in.bad() ) LOGERR( "could not open active groups file" );
	stringstream ss;
	bool modified=false;
	string groupname;
	while ( in >> groupname ) {
		if ( groupname == group )
			modified=true;
		else
			ss << groupname << "\n";
	}
	in.close();
	if ( modified ) {
		ss.seekg( 0, ios::beg );
		ofstream out( "/var/lib/scpm/active_groups", ios::trunc );
		out << ss.rdbuf();
		out.close();
		LOGINFO( "Resource Group &"+group+" deactivated" );
		if ( auto_remove ) {
			vector<resource_entry_t> resources;
			string foo;
			this->GetResourceGroup( group, resources, foo, false );
			for ( unsigned int i=0; i<resources.size(); i++ ) {
				try {
					if ( scdb->ResourceExists( resources[i].name, resources[i].type ) )
						scdb_util.DropResource( resources[i].name, resources[i].type );
				}
				catch( SCDB::db_action_failed &e ) {
					LOGERR( "could not drop resource &"+resources[i].name+" (&"+e.reason+")" );
				}
				catch( SCDBUtil::operation_failed &e ) {
					LOGERR( "could not drop resource &"+resources[i].name+" (&"+e.reason+")" );
				}
			}
		}
	}
	else LOGWARN( "Resource group "+group+" is already inactive" );
}

void SCPM_helpers::ResetResourceGroup( const string &group )
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	struct stat statbuf;
	if ( stat( (conf->GetGroupDir()+"/"+group).c_str(), &statbuf ) == -1 )
		LOGERR( "no such default group &"+group );
	string groupfile = conf->GetUserGroupDir()+"/"+group;
	if ( stat( groupfile.c_str(), &statbuf ) == 0 ) {
		if ( unlink( groupfile.c_str() ) != 0 ) {
			LOGWARN( string("unlink failed, ")+strerror( errno ) );
			LOGERR( "could not reset resource group &"+group );
		}
	}
}

void SCPM_helpers::ResetAllGroups( )
{
	DIR *dp;
	struct dirent *ep;
	SCPM_conf *conf = SCPM_conf::GetHandle();


	dp = opendir( (conf->GetUserGroupDir()).c_str() );
	if ( dp == NULL ) LOGERR( "could not access directory &"+conf->GetUserGroupDir() );
	while( (ep = readdir(dp)) ) {
		if ( (ep->d_name[0] == '.') ) continue;
		if ( unlink( (conf->GetUserGroupDir()+"/"+ep->d_name).c_str() ) != 0 ) {
			LOGWARN( string("unlink failed, ")+strerror( errno ) );
			LOGERR( "could not remove resource group &"+string(ep->d_name) );
		}
		LOGDEBUG( "removed resource group "+string(ep->d_name) );
	}
	ifstream default_file( (conf->GetLibDir()+"/default_groups").c_str() );
	if ( default_file.bad() ) LOGERR( "could not open default groups file" );
	ofstream active_file( "/var/lib/scpm/active_groups", ios::trunc );
	if ( active_file.bad() ) {
		default_file.close();
		LOGERR( "could not open active groups file for writing" );
	}
	active_file << default_file.rdbuf();
	active_file.close();
	default_file.close();
}

void SCPM_helpers::GetResourceGroup( const string &name, vector<resource_entry_t> &group,
																		 string &description, bool nouser )
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	struct stat statbuf;
	bool user = true;
	string groupfile;

	if (nouser) {
		user=false;
		groupfile = conf->GetGroupDir()+"/"+name;
	}
	else
		groupfile = conf->GetUserGroupDir()+"/"+name;
	
	if ( stat( groupfile.c_str(), &statbuf ) == -1 ) {
		if (user) {
			user = false;
			groupfile = conf->GetGroupDir()+"/"+name;
			if ( stat( groupfile.c_str(), &statbuf ) == -1 ) {
				LOGERR( "could not open resource group &"+name );
			}
		}
		else 
			LOGERR( "could not open resource group &"+name );
	}

	ifstream in( groupfile.c_str() );
	group.clear();
	char buf[255];
	while ( in.getline( buf, 254 ) ) {
		Tokenizer toker( buf );
		resource_entry_t re;
		try {
			string &token = toker.GetToken();
			if (token == "description") {
				try {
					description = toker.GetToken();
					for (;;) description += " "+toker.GetToken();
				}
				catch( Tokenizer::end_of_tokens ) {
					description = LOGTRANS( description );
				}
			}
			else {
				re.type = token;
				re.name = toker.GetToken();
				try {
					string &flag = toker.GetToken();
					if ( flag == "inactive" ) re.active=false;
				}
				catch( Tokenizer::end_of_tokens ) {
						// no flags, so we set active to true
					re.active=true;
				}
				group.push_back( re );
			}
		}
		catch( Tokenizer::end_of_tokens ) {
			LOGERR( "Resource group &"+name+" is corrupt" );
		}
	}
}
