/*
 * $Id: file.cc,v 1.12 2008/10/22 10:12:42 hschaa Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	file resource
 * File				 :	file.cc
 * Description :	handling module for file resource type
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2001 SuSE GmbH
 *					 2002-2004 SuSE Linux AG
 *					 2004 SUSE LINUX Products GmbH
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include <file.hh>
#include <log.hh>
#include <fstream>
#include <sstream>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <utime.h>
#include <errno.h>
#include <modlib.hh>
#include <scdb.hh>
#include <progress.hh>
#include <executor.hh>
#include <cstring>

using namespace std;

#define module_name "file"

string File::type = "file";

File::File( string name_string, string type_string, bool allow_missing ) :
		Resource( name_string, type_string, allow_missing )
{
	for (unsigned int i=0; i<resource_data.size(); i++) {
		if ( resource_data[i].data_name == "postrestore" )
			postrestore = resource_data[i].data_value;
		else
			LOGDEBUG( "unkown file resource data "+resource_data[i].data_name );
	}
	backup_state=0;
}

File::~File( )
{
}

bool File::Active( ) throw( Resource::not_available )
{
	return true;
}

bool File::Modified( ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;

	SCDB *db = SCDB::GetHandle();
	SCDBFile *fileh;
	try {
		fileh = db->FileGetHandle( profile, type, name, false );

			// compare type
		if (this->GetType( name ) != fileh->GetSubtype()) {
			LOGDEBUG( "subtype differs for file "+name );
			delete fileh;
			return true;
		}

		Modlib ml;
		this->ResolveFiles();

			// maybe we have an external handler
		if ( ml.HasHandler( "file", name, "status" ) ) {
			vector<string> out;
			vector<string> err;
			char tmpfile[30] = "/tmp/scpm_tmpfile.XXXXXX";
			try {
				if ( files.size() > 1 ) {
						// FIXME: file resources with subfiles do not work properly with
						// external handles yet, they do not have access to the saved files
					ml.CallHandler( "file", name, "status", "/dev/null", out, err );
				}
				else {
					if ( this->GetType( files[0] ) != "regular" ) {
							// non regular files also have bad luck
						ml.CallHandler( "file", name, "status", "/dev/null", out, err );
					}
					else {
						if ( ::mktemp( tmpfile ) != NULL ) {
							ofstream of( tmpfile );
							try {
								fileh->Restore( of );
								delete fileh;
								of.close();
								ml.CallHandler( "file", name, "status", tmpfile, out, err );
							}
							catch( SCDBFile::read_error ) {
								LOGWARN( "could not restore resource "+name+" (file) (read error)" );
								LOGWARN( "assuming resource is modified to be sure" );
								::unlink( tmpfile );
								delete fileh;
								return true;
							}
							catch( SCDBFile::write_error ) {
								LOGWARN( "could not restore resource "+name+" (file) (write error)" );
								LOGWARN( "assuming resource is modified to be sure" );
								::unlink( tmpfile );
								delete fileh;
								return true;
							}
						}
						else {
							LOGWARN( "could not make temp file" );
							LOGWARN( "assuming resource "+name+" (file) is modified to be sure" );
							::unlink( tmpfile );
							delete fileh;
							return true;
						}
						::unlink( tmpfile );
					}
				}
				if ( out.size() && out[0] == "modified" ) {
					LOGDEBUG( "file "+name+" differs (external handler)" );
					return true;
				}
				else
					return false;
			}
			catch( Modlib::handler_not_available ) {
				::unlink( tmpfile );
			}
		}		 
		if (this->FileModified( fileh )) {
			delete fileh;
			return true;
		}

		if ( fileh->GetNumContents() != files.size()-1 ) {
			LOGDEBUG( "number of contents differs for file "+name );
			delete fileh;
			return true;
		}

			// type and md5 are same, but maybe resource contents are different
			// (in case they have content)
		if ( fileh->GetNumContents() == 0 ) {
				// no, resource must be unchanged
			delete fileh;
			return false;
		}

		for (unsigned int i=1; i<files.size(); i++) {
			pair<string, string> &p = fileh->GetContent(i-1);
				// check type and name for equality
			if ( (p.second != files[i]) || (p.first != this->GetType(files[i])) ) {
				LOGDEBUG( "subfile "+files[i]+" of file "+name+" differs in name/type" );
				LOGDEBUG( "current: "+files[i]+" ("+this->GetType(files[i])+
									") saved: "+p.second+" ("+p.first+")" );
				delete fileh;
				return true;
			}
			try {
				SCDBFile *fileh2 = db->FileGetHandle( profile, "file", p.second, false );
				if (this->FileModified( fileh2 ) ) {
					delete fileh2;
					delete fileh;
					return true;
				}
				delete fileh2;
			}
			catch( SCDB::no_such_file ) {
				LOGWARN( "could not find subfile "+p.second+" of resoure "+name+" (file) in data storage" );
				LOGWARN( "assuming resource is modified" );
				return true;
			}
		}
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "could not determine modification state for "+name+" (file) ("+e.reason+")" );
		LOGWARN( "assuming resource is modified" );
		return true;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "could not find resoure "+name+" (file) in data storage" );
		LOGWARN( "assuming resource is modified" );
		return true;
	}
		// all tests failed, resource is unchanged
	return false;
}

bool File::FileModified( SCDBFile *fileh )
{
	try {
			// for links check destination
		if ( fileh->GetSubtype() == "link" ) {
			stringstream ss;
			fileh->Restore( ss );
			string tmp;
			ss >> tmp;
			if ( tmp != this->ReadLink( fileh->GetName() ) ) {
				LOGDEBUG( "subfile "+fileh->GetName()+" of file "+name+" differs in link destination" );
				LOGDEBUG( "current: "+this->ReadLink( fileh->GetName() )+"  saved: "+tmp );
				return true;
			}
		}
		else if (fileh->GetSubtype() == "dir") {
				// for directories we check only id/gid/mode
			struct stat &saved_attribs = fileh->GetAttribs();
			struct stat local_attribs;
			if (stat((fileh->GetName().c_str()), &local_attribs) == -1) {
				LOGDEBUG( "could not stat file "+fileh->GetName() );
			}
			else {
				if ( ((saved_attribs.st_mode | S_IFDIR) != (local_attribs.st_mode | S_IFREG)) ||
						 (saved_attribs.st_uid != local_attribs.st_uid) ||
						 (saved_attribs.st_gid != local_attribs.st_gid) ) {
					LOGDEBUG( "attributes differ for file "+name );
					return true;
				}
			}
		}			 
		else if (fileh->GetSubtype() == "regular") {
				// before checking md5 sums we compare attribs
				// this is faster and can even detect changes in permissions
			struct stat &saved_attribs = fileh->GetAttribs();
			struct stat local_attribs;
			if (stat((fileh->GetName().c_str()), &local_attribs) == -1) {
				LOGDEBUG( "could not stat file "+fileh->GetName() );
			}
			else {
				if ( (saved_attribs.st_mode != local_attribs.st_mode) ||
						 (saved_attribs.st_uid != local_attribs.st_uid) ||
						 (saved_attribs.st_gid != local_attribs.st_gid) ||
						 (saved_attribs.st_size != local_attribs.st_size) ) {
					LOGDEBUG( "attributes differ for file "+name );
					return true;
				}
				else if (saved_attribs.st_mtim.tv_sec != local_attribs.st_mtim.tv_sec) {
						// mod time differs, checking md5 if there was a change actually
					Modlib ml;
					string md5 = ml.GetMD5sum( fileh->GetName() );
					if (fileh->MD5sum() != ml.GetMD5sum( fileh->GetName() )) {
						LOGDEBUG( "subfile "+fileh->GetName()+" of file "+name+" differs in md5" );
						return true;
					}
				}
			}
		}
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "could not determine modification state for "+name+" (file) ("+e.reason+")" );
		LOGWARN( "assuming resource is modified" );
		return true;
	}
	catch( SCDBFile::read_error ) {
		LOGWARN( "could not determine modification state for "+name+" (file) (read_error)" );
		LOGWARN( "assuming resource is modified" );
		return true;
	}
	catch( SCDBFile::write_error ) {
		LOGWARN( "could not determine modification state for "+name+" (file) (write_error)" );
		LOGWARN( "assuming resource is modified" );
		return true;
	}
	return false;
}

bool File::UpToDate( )	throw( Resource::not_available )
{
	SCDB *db = SCDB::GetHandle();
	SCDBFile *fileh;
	try {
		fileh = db->FileGetHandle( profile, type, name, false );

		Modlib ml;
		this->ResolveFiles();

			// maybe we have an external handler
		if ( ml.HasHandler( "file", name, "status" ) ) {
			vector<string> out;
			vector<string> err;
			char tmpfile[30] = "/tmp/scpm_tmpfile.XXXXXX";
			try {
				if ( files.size() > 1 ) {
						// FIXME: file resources with subfiles do not work properly with
						// external handles yet, they do not have access to the saved files
					ml.CallHandler( "file", name, "status", "/dev/null uptodate", out, err );
				}
				else {
					if ( this->GetType( files[0] ) != "regular" ) {
							// non regular files also have bad luck
						ml.CallHandler( "file", name, "status", "/dev/null uptodate", out, err );
					}
					else {
						if ( ::mktemp( tmpfile ) != NULL ) {
							ofstream of( tmpfile );
							try {
								fileh->Restore( of );
								delete fileh;
								of.close();
								ml.CallHandler( "file", name, "status", string(tmpfile)+" uptodate", out, err );
							}
							catch( SCDBFile::read_error ) {
								LOGWARN( "could not restore resource "+name+" (file) (read error)" );
								LOGWARN( "assuming resource is not uptodate to be sure" );
								::unlink( tmpfile );
								delete fileh;
								return false;
							}
							catch( SCDBFile::write_error ) {
								LOGWARN( "could not restore resource "+name+" (file) (write error)" );
								LOGWARN( "assuming resource is not uptodate to be sure" );
								::unlink( tmpfile );
								delete fileh;
								return false;
							}
						}
						else {
							LOGWARN( "could not make temp file" );
							LOGWARN( "assuming resource "+name+" (file) is not uptodate to be sure" );
							::unlink( tmpfile );
							delete fileh;
							return false;
						}
						::unlink( tmpfile );
					}
				}
				if ( out.size() && out[0] == "modified" ) {
					LOGDEBUG( "file "+name+" differs (external handler)" );
					return false;
				}
				else
					return true;
			}
			catch( Modlib::handler_not_available ) {
				::unlink( tmpfile );
			}
		}		 
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "could not determine modification state for "+name+" (file) ("+e.reason+")" );
		LOGWARN( "assuming resource is not update" );
		return false;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "could not find resoure "+name+" (file) in data storage" );
		LOGWARN( "assuming resource is not uptodate" );
		return false;
	}
		// for File resources without external handlers, UpToDate is invers to Modified
	return !this->Modified();
}

bool File::Save( )	throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;

	this->ResolveFiles();
	if (files.size() == 0 ) {
			// should not be possible
		LOGWARN( "Could not resolve files for "+name+" (file)" );
		return false;
	}
	SCDB *db = SCDB::GetHandle();

	try {
		SCDBFile *fileh;
		vector<pair<string, string> > old_contents;
		string suffix="";
		if (!db->FileExists(profile, type, files[0])) {
			LOGDEBUG( "resource "+name+" (file) does not exist, no backup will be made" );
			fileh = db->FileGetHandle( profile, type, files[0], true );
		}
		else 
			fileh = db->FilePushBackup( profile, type, files[0] );

		// get old contents for checking if files were removed
		for(unsigned int i=0; i< fileh->GetNumContents(); i++)
			old_contents.push_back( (*fileh)[i] );

		fileh->ClearContent();
		this->SaveFile( files[0], fileh );
		suffix = fileh->GetBackupVersion();
		for (unsigned int i=1; i<files.size(); i++) {
			pair<string,string> p= make_pair( this->GetType( files[i] ), files[i] );
			fileh->AddContent( p );
			SCDBFile *tmp;
			if (!db->FileExists( profile, type, files[i] )) {
				LOGDEBUG( "file "+files[i]+" seems to be new in resource "+name+" (file), creating it" );
				tmp = db->FileGetHandle( profile, type, files[i], true );
			}
			else 
				tmp = db->FilePushBackup( profile, type, files[i] );
			this->SaveFile( files[i], tmp, suffix );
			delete tmp;
			// delete this content from our old-content list
			for(vector<pair<string,string> >::iterator j = old_contents.begin(); j != old_contents.end(); j++) {
				if (*j == p){
					old_contents.erase(j);
					break;
				}
			}
		}
		delete fileh;
		// push removed files in the backup and delete the actual scdb_file
		for (unsigned int i=0; i<old_contents.size(); i++){
			SCDBFile* tmp;
			if (db->FileExists( profile, type, old_contents[i].second)){
				// has to be removed
				tmp = db->FilePushBackup(profile, type, old_contents[i].second, true);
				delete tmp;
			}
		}
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database error ("+e.reason+")" );
		return false;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "could not save resource "+name+" (file) (resource not existent)" );
		return false;
	}
	catch( SCDBFile::write_error ) {
		LOGWARN( "could not save "+name+" (file)" );
		return false;
	}
	return true;
}

bool File::Restore( ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;

	this->ResolveFiles();
	if (files.size() == 0 ) {
			// should not be possible
		LOGWARN( "Could not resolve files for "+name+" (file)" );
		return false;
	}
	SCDB *db = SCDB::GetHandle();
	
		// clean resource first
		// remove all non directory files
	struct stat st;
	for ( unsigned int i=0; i<files.size(); i++ ) {
		if ( stat( (resource_root+files[i]).c_str(), &st ) != 0 )
			LOGDEBUG( "stat() on "+(resource_root+files[i])+" failed ("+strerror(errno)+")" );
		else {
			if ( !S_ISDIR(st.st_mode) ) {
				LOGDEBUG( "removing file "+files[i] );
				unlink( (resource_root+files[i]).c_str() );
			}
		}
	}
		// now remove direcotries in reverse order
	for ( unsigned int i=files.size()-1; i; i-- ) {
		if ( stat( (resource_root+files[i-1]).c_str(), &st ) != 0 )
			LOGDEBUG( "stat() on "+files[i-1]+" failed ("+strerror(errno)+")" );
		else {
			if ( S_ISDIR(st.st_mode) ) {
				LOGDEBUG( "removing directory "+files[i-1] );
				rmdir( (resource_root+files[i-1]).c_str() );
			}
		}
	}

	SCDBFile *fileh=NULL;
	SCDBFile *tmph=NULL;
	try {
		fileh = db->FileGetHandle( profile, type, files[0] );

		if ( !this->RestoreFile( fileh ) ) {
			LOGWARN( "could not restore file "+files[0] );
			delete fileh;
			return false;
		}

		for (unsigned int i=0; i<fileh->GetNumContents(); i++) {
			pair<string,string> &p = fileh->GetContent( i );
			tmph = db->FileGetHandle( profile, type, p.second );
			this->RestoreFile( tmph );
			delete tmph;
			tmph=NULL;
		}
	}
	catch( SCDB::db_action_failed &e ) {
		if (fileh) delete fileh;
		if (tmph) delete tmph;
		LOGWARN( "could not save "+name+" (file)" );
		LOGWARN( "database I/O error, reason "+e.reason );
		return false;
	}
	Modlib ml;
	if (ml.HasHandler( "file", name, "postrestore" )) {
		if (resource_root.empty()) {
			LOGDEBUG( "Calling postrestore script for resource "+name+" (file)" );
			vector<string> out;
			vector<string> err;
			if (ml.CallHandler( "file", name, "postrestore", name, out, err ) != 0) {
				LOGWARN( "postrestore script of resource "+name+" (file) failed" );
				for (unsigned int i=0;i<err.size();i++) LOGDEBUG( "postrestore: "+err[i]);
			}
		}
		else
			LOGDEBUG( "Resource root is set, skipping postrestore handler of resource "+name+" (file)" );
	}
	if (fileh) delete fileh;
	if (tmph) delete tmph;
	return true;
}

bool File::Patch( const string &ref_profile ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;
	if ( !this->Modified() ) return false;

	bool ret = false;
	char reffile[]="/tmp/stored_ref.XXXXXX";
	char patchfile[]="/tmp/stored.XXXXXX";
	char difffile[]="/tmp/diff.XXXXXX";
	if ( mktemp(reffile) == NULL ) {
		LOGWARN( "could not create temp file" );
		return false;
	}
	if ( mktemp(patchfile) == NULL ) {
		unlink( reffile );
		LOGWARN( "could not create temp file" );
		return false;
	}
	if ( mktemp(difffile) == NULL ) {
		unlink( reffile );
		unlink( patchfile );
		LOGWARN( "could not create temp file" );
		return false;
	}
		
	this->ResolveFiles();
	for ( unsigned int i=0; i<files.size(); i++ ) {
		string file_type = GetType( files[i] );
		if ( file_type != "regular" ) {
			LOGWARN( "file "+files[i]+" (type "+file_type+") is non regluar, skipping" );
		}
		else {
			SCDB *db = SCDB::GetHandle();

			if ( !db->FileExists( ref_profile, "file", files[i]) ) {
				LOGWARN( "can not patch file "+files[i]+" (does not exist in profile "+ref_profile+")" );
				continue;
			}
			
			ofstream tmp( reffile );
			SCDBFile *fh = NULL;
			try {
				db->FileRestore( ref_profile, "file", files[i], tmp );
				tmp.close();

				int dfd = open(difffile, O_WRONLY|O_CREAT, 0600);
				if (dfd == -1) {
					LOGWARN( string("patching failed (could not open ")+difffile+" for writing)" );
					goto out;
				}

				int status=0;
				pid_t pid = fork();
				if (pid == -1) {
					LOGWARN("patching failed (can not fork)");
					goto out;
				}

				if (pid==0) {
					close(2);
					dup2(dfd,1);
					execlp( "diff", "diff", "-u", "-a", "-r", "--label", patchfile,
									files[i].c_str(), reffile, NULL );
					exit(-1);
				}
				waitpid(pid, &status,0);
				close(dfd);
				if (!WIFEXITED(status)) {
					LOGWARN( "patching failed (diff returned error)");
					goto out;
				}

				ifstream diff( difffile );
				if ( diff.get() == EOF ) {
					diff.close();
					LOGDEBUG( "file "+files[i]+" (type "+file_type+") is not modified, skipping" );
				}
				else {
					diff.close();
					ofstream tmp2( patchfile );
					db->FileRestore( profile, "file", files[i], tmp2 );
					tmp2.close();
					dfd = open( difffile, O_RDONLY );
					if (dfd == -1 ) {
						LOGWARN( "patch failed (could not open diff file)" );
						goto out;
					}
					pid = fork();
					if (pid == -1) {
						LOGWARN( "patch failed (could not fork)" );
						goto out;
					}
					if (pid == 0) {
						setuid(0);
						dup2(dfd,0);
						execlp( "patch", "patch", "-p0", "-s", "-t", "-R", "--no-backup-if-mismatch", NULL );
						exit(-1);
					}
					waitpid(pid, &status, 0);
					if (!WIFEXITED(status)) {
						LOGWARN( "patching failed (patch returned error)");
						goto out;
					}

					fh = db->FilePushBackup( profile, "file", files[i] );
					this->SaveFile( patchfile, fh );
					delete fh;
				}
			}
			catch( SCDB::no_such_file ) {
				if (fh) delete fh;
				LOGWARN( "resource "+name+" (file) not in storage" );
				LOGWARN( "could not patch file &"+files[i]+" of profile &"+profile );
				goto out;
			}
			catch( SCDB::db_action_failed &e ) {
				if (fh) delete fh;
				LOGWARN( "scdb error ("+e.reason+")" );
				LOGWARN( "could not patch file &"+files[i]+" of profile &"+profile );
				goto out;
			}
	
		}
	}
	ret = true;
	out:
	unlink( reffile );
	unlink( difffile );
	unlink( patchfile );
	return ret;
}

void File::UpdateData( ) throw( Resource::not_available )
{
	resource_data.clear();
}

string File::ReadLink( const string &link )
{
	int n;
	char buf[250];

	n=readlink( link.c_str(), buf, 249 );
	if ( n == -1 ) return( "" );

	buf[n]='\0';

	return buf;
}

void File::ShowChanges( ostream &output, bool backup, string version ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return;
	if ( !backup && !this->Modified() ) {
		output << "files are identical\n";
		return;
	}

	SCDB *db = SCDB::GetHandle();
	char tmpfile[]="/tmp/stored.XXXXXX";
	char difffile[]="/tmp/diff.XXXXXX";
	if ( mktemp(tmpfile) == NULL ) {
		LOGWARN( "could not create temp file" );
		output << LOGTRANS( "  could not determine changes for resource &"+name+" (file)" ) << endl;
		return;
	}
	if ( mktemp(difffile) == NULL ) {
		LOGWARN( "could not create temp file" );
		output << LOGTRANS( "  could not determine changes for resource &"+name+" (file)" ) << endl;
		return;
	}
		
	this->ResolveFiles();
	for ( unsigned int i=0; i<files.size(); i++ ) {
		string file_type = GetType( files[i] );
		SCDBFile *fileh=NULL;
			// maybe type differs
		try {
			if (backup)
				fileh = db->FileGetBackupHandle( profile, type, files[i], false, (version=="" ? NULL : &version) );
			else
				fileh = db->FileGetHandle( profile, type, files[i], false );
		}
		catch( SCDB::no_such_file ) {
			LOGDEBUG( "resource "+name+" (file) (or backup of) not in storage" );
			output << "file " << files[i] << endl;
			if (backup) 
				output << LOGTRANS( "  file has no backup in profile &"+profile ) << endl;
			else
				output << LOGTRANS( "  file does not exist in profile &"+profile ) << endl;
			continue;
		}
		catch( SCDB::db_action_failed &e ) {
			LOGWARN( "scdb error ("+e.reason+")" );
			output << "file " << files[i] << endl;
			output << LOGTRANS( "  could not access file in profile &"+profile ) << endl;
			continue;
		}
		if (file_type != fileh->GetSubtype()) {
			output << "file " << files[i] << endl
						 << "  current type: " << file_type << endl
						 << "  saved type: " << fileh->GetSubtype() << endl;
		}
		else if ( file_type == "regular" ) {
			ofstream tmp( tmpfile );
			try {
				fileh->Restore( tmp );
				tmp.close();
		
//				string cmdline = "/usr/bin/diff -u ";
//				cmdline += tmpfile;
//				cmdline += " ";
//				cmdline += files[i];
//				cmdline += " >";
//				cmdline += difffile;

					// FIXME: output redirection does not work with Executor
//	vector<string> out, err;
//	Executor e( cmdline );
//	try {
//		e.Run( out, err );
//	}
//	catch( Executor::exec_failed ) {
//		LOGWARN( "executing diff command failed");
//		if ( !err.empty() ) {
//			LOGDEBUG( "error messages" );
//			for (unsigned int i=0; i<err.size(); i++)
//				LOGDEBUG( err[i] );
//		}
//		return;
//	}
				int difffd = open( difffile, O_WRONLY|O_TRUNC|O_CREAT, 0600 );
				if (difffd == -1 ) {
					LOGWARN( "executing diff command failed (could not open temp file)");
					return;
				}

				pid_t pid = fork();
				if (pid == -1) {
					LOGWARN( "executing diff command failed (could not fork)");
					close(difffd);
					return;
				}
				if (pid == 0) {
					setuid( 0 );
					if ( dup2(difffd,1) == -1 )
						cerr << "dup2() failed" << endl;
					execlp( "diff", "diff", "-u", tmpfile, files[i].c_str(), NULL );
					LOGWARN( "executing diff command failed");
				}
				else {
					wait(NULL);
					close(difffd);
					ifstream diff( difffile );
					if ( diff.get() != EOF ) {
						diff.seekg( 0, ios::beg );
						output << "file " << files[i] << " (type " << file_type << ")\n"
									 << diff.rdbuf();
					}
					diff.close();
				}

				struct stat &saved_attribs = fileh->GetAttribs();
				struct stat local_attribs;
				if (stat((fileh->GetName().c_str()), &local_attribs) == -1) {
					LOGDEBUG( "could not stat file "+fileh->GetName() );
				}
				else {
					if ( (saved_attribs.st_mode != local_attribs.st_mode) ||
							 (saved_attribs.st_uid != local_attribs.st_uid) ||
							 (saved_attribs.st_gid != local_attribs.st_gid) ) {
						output << "  " << LOGTRANS( "current attributes:" )
									 << " mode " << local_attribs.st_mode
									 << " owner " << local_attribs.st_uid << ":" << saved_attribs.st_gid
									 << "\n	 " << LOGTRANS( "saved attributes:" )
									 << " mode " << saved_attribs.st_mode
									 << " owner " << saved_attribs.st_uid << ":" << saved_attribs.st_gid << endl;
					}
				} 
			}
			catch( SCDB::no_such_file ) {
				tmp.close();
				LOGWARN( "resource "+name+" (file) (or backup of) not in storage" );
				if (backup)
					output << LOGTRANS( "  file has no backup in profile &"+profile ) << endl;
				else
					output << LOGTRANS( "  file does not exist in profile &"+profile ) << endl;
			}
			catch( SCDB::db_action_failed &e ) {
				tmp.close();
				unlink( tmpfile );
				unlink( difffile );
				LOGWARN( "scdb error ("+e.reason+")" );
				output << LOGTRANS( "  could not access file in profile &"+profile ) << endl;
				return;
			}
		}
		else if ( file_type == "link" ) {
			stringstream ss;
			try {
				fileh->Restore( ss );
				string l;
				ss >> l;
				if (this->ReadLink(files[i]) != l) {
					output << "file " << files[i] << " (type " << file_type << ")" << endl;
					output << LOGTRANS( "  current link destination: " ) << this->ReadLink(files[i]) << endl
								 << LOGTRANS( "  saved link destination: " ) << l << endl;
				}
			}
			catch( SCDB::no_such_file ) {
				LOGWARN( "resource "+name+" (file) (or backup of) not in storage" );
				if (backup)
					output << LOGTRANS( "  file has no backup in profile &"+profile ) << endl;
				else
					output << LOGTRANS( "  file does not exist in profile &"+profile ) << endl;
			}
			catch( SCDB::db_action_failed &e ) {
				LOGWARN( "scdb error ("+e.reason+")" );
				output << LOGTRANS( "  could not access file in profile &"+profile ) << endl;
				return;
			}
		}

		if (fileh) delete fileh;
	}
		// it is possible that there are files saved in the profile but are not in
		// the system (anymore)
	try {
		SCDBFile *fileh;
		if (backup)
			fileh = db->FileGetBackupHandle( profile, "file", name, false );
		else
			fileh = db->FileGetHandle( profile, "file", name, false );
		for( unsigned int i=0; i<fileh->GetNumContents(); i++ ) {
			bool found=false;
			for( unsigned int j=1; j<files.size(); j++ ) {
				if ( files[j] == (*fileh)[i].second ) {
					found=true;
					break;
				}
			}
			if ( !found ) {
				output << "file " << (*fileh)[i].second	 << " (type " << (*fileh)[i].first << ")" << endl;
				output << LOGTRANS( "  file only exists in profile &"+profile ) << endl;
			}
		}
		delete fileh;
	}
	catch( SCDB::no_such_file ) {
			// should not happen
		LOGDEBUG( "unexpected no_such_file, maybe SCDB corrupt" );
		LOGWARN( "resource "+name+" (file) not in storage" );
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
	}
	unlink( tmpfile );
	unlink( difffile );
}

void File::ResolveFiles( )
{
	vector<string> err;
	files.clear();
	Modlib ml;
	
	try {
		ml.CallHandler( "file", name, "get_files", resource_root, files, err );
	}
	catch( Modlib::handler_not_available ) {
		LOGWARN( "could not call get_files handler for resource "+name+ " (file)" );
		LOGERR( "could not determine files for resource &"+name+" (file)" );
	}
}

void File::SaveFile( const string &file, SCDBFile *fh, string suffix) throw( SCDBFile::write_error )
{
	string ftype = this->GetType( file );
	LOGDEBUG( "saving file "+file+" type "+ftype );
	fh->SetSubtype( ftype );
	if ( ftype == "regular" ) {
		ifstream in( (resource_root+file).c_str() );
		if ( in.good() ) {
			fh->Save( in, true, suffix );
		}
		else {
			LOGWARN( "could not open file "+file );
			LOGWARN( "could not save resource &"+name+" (file)" );
			return;
		}
		in.close();
		Modlib ml;
		fh->SetMD5sum( ml.GetMD5sum( file ) );
		struct stat file_stat;
		if ( stat( (resource_root+file).c_str(), &file_stat ) != 0 ) {
			LOGWARN( "can not save attributes for "+file+" ("+strerror( errno)+")" );
		}
		else {
			fh->SaveAttribs( file_stat );
		}
	}
	else if ( ftype == "ghost" ) {
		fh->Remove();
	}
	else if ( ftype == "link" ) {
		stringstream s;
		s << this->ReadLink( resource_root+file ) << endl;
		fh->Save( s, true, suffix );
	}
	else if ( ftype == "dir" ) {
		fh->Create(suffix);
		struct stat file_stat;
		if ( stat( (resource_root+file).c_str(), &file_stat ) != 0 ) {
			LOGWARN( "can not save attributes for "+file+" ("+strerror( errno)+")" );
		}
		else {
			fh->SaveAttribs( file_stat );
		}
	}
	else {
		LOGWARN( "can not save "+name+" (file), unknow subtype "+ftype );
		return;
	}
}

bool File::RestoreFile( SCDBFile *f )
{
	string &fname = f->GetName();
	string &ftype = f->GetSubtype();
	LOGDEBUG( "restoring file "+fname );
	if ( ftype == "link" ) {
		unlink( (resource_root+fname).c_str() );
		stringstream tmp;
		try {
			f->Restore( tmp );
			char buf[255];
			tmp.getline( buf, 254 );
			if ( symlink( buf, (resource_root+fname).c_str() ) == -1 ) {
				LOGDEBUG( "could not create link "+fname+" ("+strerror(errno)+")" );
				LOGWARN( "could not restore resource &"+fname+" (file)" );
				return false;
			}
		}
		catch( SCDBFile::read_error ) {
			LOGDEBUG( "could not read saved link destination of "+fname );
			LOGWARN( "could not restore resource &"+fname+" (file)" );
			return false;
		}
		catch( SCDBFile::write_error ) {
			LOGDEBUG( "could not restore link "+fname );
			LOGWARN( "could not restore resource &"+fname+" (file)" );
			return false;
		}
	}
	else if ( ftype == "ghost" ) {
		unlink( (resource_root+fname).c_str() );
	}
	else if ( ftype == "dir" ) {
		struct stat file_stat;
		if ( lstat( (resource_root+fname).c_str(), &file_stat ) == 0 ) {
				// the file exists
			if ( !S_ISDIR( file_stat.st_mode ) ) {
					// there is something out there, but it is not a dir
				unlink( (resource_root+fname).c_str() );
			}
		}
			// now there are two possibilities: the file exists and is
			// a dir or it exists not at all (anymore)
		if ( lstat( (resource_root+fname).c_str(), &file_stat ) == -1 ) {
			mkdir( (resource_root+fname).c_str(), 0755 );
			try {
				struct stat &file_stat = f->GetAttribs();
				struct utimbuf ub;
				ub.actime = file_stat.st_atime;
				ub.modtime = file_stat.st_mtime;
				utime( (resource_root+fname).c_str(), &ub );
				chmod( (resource_root+fname).c_str(), file_stat.st_mode );
				chown( (resource_root+fname).c_str(), file_stat.st_uid, file_stat.st_gid );
			}
			catch( SCDBFile::read_error ) {
				LOGWARN( "could not restore attribs for file "+fname );
			}
		}
	}
	else if ( ftype == "regular" ) {
		ofstream o( (resource_root+fname).c_str() );
		try {
			f->Restore( o );
		}
		catch( SCDBFile::read_error ) {
			LOGWARN( "could not restore resource &"+fname+" (file)" );
			return false;
		}
		catch( SCDBFile::write_error ) {
			LOGWARN( "could not restore resource &"+fname+" (file)" );
			return false;
		}
		try {
			struct stat &file_stat = f->GetAttribs();
			struct utimbuf ub;
			ub.actime = file_stat.st_atime;
			ub.modtime = file_stat.st_mtime;
			utime( (resource_root+fname).c_str(), &ub );
			chmod( (resource_root+fname).c_str(), file_stat.st_mode );
			chown( (resource_root+fname).c_str(), file_stat.st_uid, file_stat.st_gid );
		}
		catch( SCDBFile::read_error ) {
			LOGWARN( "could not restore attributes for file "+fname );
		}
		o.close();
	}
	else {
		LOGWARN( "unknown file type '" +ftype+"' of file "+fname+" resource "+name+" (file)" );
		LOGWARN( "could not restore resource &"+name+" (file)" );
		return false;
	}
	return true;
}

string File::GetType( const string &file )
{
	struct stat status_buf;
	if ( lstat( (resource_root+file).c_str(), &status_buf ) == -1 )
		return "ghost";
	if ( S_ISLNK( status_buf.st_mode ) ) return "link";
	if ( S_ISREG( status_buf.st_mode ) ) return "regular";
	if ( S_ISDIR( status_buf.st_mode ) ) return "dir";
	return "unkown";
}

bool File::NeedsUpdate( ) throw( not_available )
{
	this->ResolveFiles();
	if (files.size() == 0 ) {
			// should not be possible
		LOGWARN( "Could not resolve files for "+name+" (file)" );
		return false;
	}
	for (unsigned int i=0; i<files.size(); i++) {
		if ( this->GetType( files[i] ) != "regular" ) {
			LOGDEBUG( "file "+files[i]+" is non regular, skipping" );
			continue;
		}
		struct stat s;
		if ( ::stat( (files[i]+".rpmsave").c_str(), &s ) == 0 )
			return true;
			/* we do not need to save .rpmorig files, would be the same in every profile anyway
				 if ( ::stat( (files[i]+".rpmorig").c_str(), &s ) == 0 )
				 return true;
			*/
	}
	return false;
}

bool File::ModifiedInProfile( const string &dest_profile ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;

	SCDB *db = SCDB::GetHandle();
	SCDBFile *srch, *desth;
	try {
		srch = db->FileGetHandle( profile, type, name, false );
		desth = db->FileGetHandle( dest_profile, type, name, false );

			// compare type
		if (srch->GetSubtype() != desth->GetSubtype()) {
			LOGDEBUG( "subtype differs for file "+name );
			delete srch;
			delete desth;
			return true;
		}

		Modlib ml;
			// maybe we have an external handler
		if ( ml.HasHandler( "file", name, "status" ) ) {
			LOGDEBUG( "file "+name+" has external handler, not supported by ModifiedInProfile()" );
			delete srch;
			delete desth;
			return false;
		}		 
			// compare md5 (only for regular files)
		if ( srch->GetSubtype() == "regular") {
			if ( srch->MD5sum() != desth->MD5sum() ) {
				LOGDEBUG( "md5 differs for file "+name+" "+srch->MD5sum()+" != "+desth->MD5sum() );
				delete srch;
				delete desth;
				return true;
			}
		}

		if ( srch->GetNumContents() != desth->GetNumContents() ) {
			LOGDEBUG( "number of contents differs for file "+name );
			delete srch;
			delete desth;
			return true;
		}

			// type and md5 are same, but maybe resource contents are different
			// (in case they have content)
		if ( srch->GetNumContents() == 0 ) {
				// no, resource must be unchanged
			delete srch;
			delete desth;
			return false;
		}
		for (unsigned int i=0; i<srch->GetNumContents(); i++) {
			pair<string, string> &p1 = srch->GetContent(i);
			pair<string, string> &p2 = desth->GetContent(i);
				// check type and name for equality
			if ( p1.first != p2.first || p1.second != p2.second ) {
				LOGDEBUG( "subfile "+p1.second+" of file "+name+" differs in name/type" );
				delete srch;
				delete desth;
				return false;
			}
				// check md5 for regular files (more is not supported, but should be sufficiant)
			if ( p1.first == "regular" ) {
				if (db->FileGetMD5(profile, p1.second, p1.first) != 
						db->FileGetMD5(profile, p2.second, p2.first) ) {
					LOGDEBUG( "subfile "+p1.second+" of file "+name+" differs in md5" );
					delete srch;
					delete desth;
					return true;
				}
			}
		}
		delete srch;
		delete desth;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "could not determine modification state for "+name+" (file) ("+e.reason+")" );
		LOGWARN( "assuming resource is not modified" );
		return false;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "could not find resoure "+name+" (file) in data storage" );
		LOGWARN( "assuming resource is not modified" );
		return false;
	}
		// all tests failed, resource is unchanged
	return false;
}

bool File::HasSystemAlternative( ) throw( Resource::not_available )
{
	this->ResolveFiles();
	if (files.size() == 0 ) {
			// should not be possible
		LOGWARN( "Could not resolve files for "+name+" (file)" );
		return false;
	}
	for (unsigned int i=0; i<files.size(); i++) {
		if ( this->GetType( files[i] ) != "regular" ) {
			LOGDEBUG( "file "+files[i]+" is non regular, skipping" );
			continue;
		}
		struct stat s;
		if ( ::stat( (files[i]+".rpmorig").c_str(), &s ) == 0 )
			return true;
	}
	return false;
}

bool File::HasSystemBackup( ) throw( Resource::not_available )
{
	this->ResolveFiles();
	if (files.size() == 0 ) {
			// should not be possible
		LOGWARN( "Could not resolve files for "+name+" (file)" );
		return false;
	}
	for (unsigned int i=0; i<files.size(); i++) {
		if ( this->GetType( files[i] ) != "regular" ) {
			LOGDEBUG( "file "+files[i]+" is non regular, skipping" );
			continue;
		}
		struct stat s;
		if ( ::stat( (files[i]+".rpmsave").c_str(), &s ) == 0 )
			return true;
	}
	return false;
}

bool File::AddBackup( ) throw( Resource::not_available )
{
		// this function does basically the same like Save(), but
		// uses FileGetBackupHandle instead of FilePushBackup for getting file handles
		// FIXME: code could be shared
	if ( profile == "__unset" ) return false;

	this->ResolveFiles();
	if (files.size() == 0 ) {
			// should not be possible
		LOGWARN( "Could not resolve files for "+name+" (file)" );
		return false;
	}
	SCDB *db = SCDB::GetHandle();

	try {
		string suffix="";
		SCDBFile *fileh = db->FileGetBackupHandle( profile, type, files[0], true );

		fileh->ClearContent();
		this->SaveFile( files[0], fileh );
		utime(fileh->GetLocation().c_str(), NULL);
		suffix = fileh->GetBackupVersion();
		for (unsigned int i=1; i<files.size(); i++) {
			pair<string,string> p= make_pair( this->GetType( files[i] ), files[i] );
			fileh->AddContent( p );
			SCDBFile *tmp = db->FileGetBackupHandle( profile, type, files[i] );
			this->SaveFile( files[i], tmp, suffix );
			utime(tmp->GetLocation().c_str(), NULL);
			delete tmp;
		}
		delete fileh;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database error ("+e.reason+")" );
		return false;
	}
	catch( SCDB::no_such_file ) {
			// oops, can not happen
		LOGWARN( "could not create SCDB file" );
		return false;
	}
	catch( SCDBFile::write_error ) {
		LOGWARN( "could not save "+name+" (file)" );
		return false;
	}
	return true;
}

bool File::HasBackup( ) throw( Resource::not_available )
{
	if (backup_state == 1) return true;
	if (backup_state == 2) return false;
	if ( profile == "__unset" ) {
		LOGWARN( "profile is unset, can not check for backup files" );
		return false;
	}
	SCDB *db = SCDB::GetHandle();
	if (db->FileHasBackup( profile, "file", name ) ) return true;
	try {
		SCDBFile *fileh = db->FileGetHandle( profile, "file", name, false );
		for (unsigned int i=0; i<fileh->GetNumContents(); i++)
			if (db->FileHasBackup( profile, "file", (*fileh)[i].second ) ) return true;
		delete fileh;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error ("+e.reason+")" );
		LOGWARN( "could not query backup status for resource &"+name+" (file)" );
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "scdb error (file not found)" );
		LOGWARN( "could not query backup status for resource &"+name+" (file)" );
	}
	return false;
}

vector< pair<time_t, string> > File::GetBackupVersions( ) throw( Resource::not_available )
{
	vector< pair<time_t, string > > v;
	if (backup_state ==2 ) return v;
	if (profile == "__unset" ) {
		LOGWARN("profile is unset, can not check for backup files");
		return v;
	}
	SCDB *db = SCDB::GetHandle();
	try{
		// get all backups for this resource
		db->FileGetBackupVersions( profile, "file", name, v);
	}
	catch(SCDB::db_action_failed &e ) {
		LOGWARN("database I/O error ("+e.reason+")");
		LOGWARN("could not query backups for resource &"+name+" (file)");
	}
	
	return v;
}

bool File::RestoreBackup(const string * version ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) {
		LOGWARN( "profile is unset, can not restore backup files" );
		return false;
	}
	SCDB *db = SCDB::GetHandle();
	try {
		if (!db->FileHasBackup( profile, "file", name )) {
			LOGWARN( "resource "+name+" (file) does not have any backups" );
			return false;
		}
		// FilePopBackup must be called before handle is created
		// otherwise we may overwrite changes
		SCDBFile *fileh = db->FilePopBackup( profile, "file", name, version);
		bool res = this->RestoreFile( fileh );

		// if we have contents we have to restore these too
		for (unsigned int i=0; i<fileh->GetNumContents(); i++) {
			// Restore Content File
			SCDBFile *fh = db->FilePopBackup( profile, "file", (*fileh)[i].second, version );
			if (!this->RestoreFile( fh )) res=false;
			delete fh;
		}
		delete fileh;
		return res;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error ("+e.reason+")" );
		LOGWARN( "could not prepare backup restore of resource &"+name+" (file)" ); 
		return false;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "scdb error (file not found)" );
		LOGWARN( "this should not happen and may be a bug" );
		LOGWARN( "could not prepare backup restore of resource &"+name+" (file)" ); 
		return false;
	}
}

bool File::RemoveBackup( const string * version) throw( Resource::not_available )
{
 	if ( profile == "__unset" ) {
		LOGWARN( "profile is unset, can not restore backup files" );
		return false;
	}
	SCDB *db = SCDB::GetHandle();
	try {
		if (!db->FileHasBackup( profile, "file", name )) {
			LOGWARN( "resource "+name+" (file) does not have any backups" );
			return false;
		}
		return db->FileRemoveBackup(profile, "file", name, version);
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error ("+e.reason+")" );
		LOGWARN( "could not prepare backup restore of resource &"+name+" (file)" ); 
		return false;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "scdb error (file not found)" );
		LOGWARN( "this should not happen and may be a bug" );
		LOGWARN( "could not prepare backup restore of resource &"+name+" (file)" ); 
		return false;
	}
}

bool File::Refresh( ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;

	this->ResolveFiles();
	if (files.size() == 0 ) {
			// should not be possible
		LOGWARN( "Could not resolve files for "+name+" (file)" );
		return false;
	}
	SCDB *db = SCDB::GetHandle();

	try {
		SCDBFile *fileh;
		if (!db->FileExists(profile, type, files[0])) {
			LOGDEBUG( "resource "+name+" (file) does not exist, no backup will be made" );
			fileh = db->FileGetHandle( profile, type, files[0], true );
		}
		else 
			fileh = db->FilePushBackup( profile, type, files[0] );
		fileh->ClearContent();
		fileh->SetMD5sum( fileh->RealMD5sum( ) );
		this->SaveFile( files[0], fileh );
		for (unsigned int i=1; i<files.size(); i++) {
			pair<string,string> p= make_pair( this->GetType( files[i] ), files[i] );
			fileh->AddContent( p );
			SCDBFile *tmp;
			if (!db->FileExists( profile, type, files[i] )) {
				LOGDEBUG( "file "+files[i]+" seems to be new in resource "+name+" (file), creating it" );
				tmp = db->FileGetHandle( profile, type, files[i], true );
			}
			else 
				tmp = db->FilePushBackup( profile, type, files[i] );
			tmp->SetMD5sum( tmp->RealMD5sum( ) );
			delete tmp;
		}
		delete fileh;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database error ("+e.reason+")" );
		return false;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "could not save resource "+name+" (file) (resource not existent)" );
		return false;
	}
	catch( SCDBFile::write_error ) {
		LOGWARN( "could not save "+name+" (file)" );
		return false;
	}
	return true;
}

void File::CreatePath( const string &path )
{
	struct stat buf;
	if ( stat( path.c_str(), &buf) == 0 ) return;
	if ( mkdir( path.c_str(), 0700 ) != 0 ) {
		CreatePath( path.substr(0, path.find_last_of('/') ) );
		mkdir( path.c_str(), 0700 );
	}
}

bool File::Export( const string &root ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;

	SCDB *db = SCDB::GetHandle();
	
	SCDBFile *fileh=NULL;
	SCDBFile *tmph=NULL;
	string tmp = resource_root;
	resource_root = root+"/";

		// this is bit hackish
		// we need to create the directory tree for the resource, as the
		// RestoreFile function does not take care of this
	this->CreatePath( (resource_root+name).substr(0, (resource_root+name).find_last_of('/')) );

	try {
		fileh = db->FileGetHandle( profile, type, name );

		if ( !this->RestoreFile( fileh ) ) {
			LOGWARN( "could not export file "+name );
			delete fileh;
			return false;
		}
		for (unsigned int i=0; i<fileh->GetNumContents(); i++) {
			pair<string,string> &p = fileh->GetContent( i );
			tmph = db->FileGetHandle( profile, type, p.second );
			this->RestoreFile( tmph );
			delete tmph;
			tmph=NULL;
		}
	}
	catch( SCDB::db_action_failed &e ) {
		resource_root = tmp;
		if (fileh) delete fileh;
		if (tmph) delete tmph;
		LOGWARN( "could not export "+name+" (file)" );
		LOGWARN( "database I/O error, reason "+e.reason );
		return false;
	}
	if (fileh) delete fileh;
	if (tmph) delete tmph;
	resource_root = tmp;
	return true;
}

bool File::Import( const string &root ) throw( Resource::not_available )
{
	string tmp = resource_root;
	resource_root = root+"/";
	bool ret = this->Save();
	resource_root = tmp;
	return ret;
}
