/*
 * $Id: service.cc,v 1.4 2007/03/09 09:43:26 gleissner Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	service resource
 * File				 :	service.cc
 * Description :	handling module for service resource type
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2001 SuSE GmbH
 *					 2002-2004 SuSE Linux AG
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "service.hh"
#include <log.hh>
#include <executor.hh>
#include <modlib.hh>
#include <scpm_conf.hh>
#include <sstream>
#include <fstream>
#include <utmp.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <scdb.hh>

using namespace std;

#define module_name "service"

Service::Service( string name_string, string type_string, bool allow_missing ) 
		: Resource( name_string, type_string, allow_missing )
{
	log = Log::GetHandle();
	provides.clear();

	for (unsigned int i=0; i<resource_data.size(); i++) {
		if ( resource_data[i].data_name == "provides" )
			provides.push_back( resource_data[i].data_value );
	}

		// check out provides
	if ( provides.size() == 0 ) {
		try {
			Modlib ml;
			vector<string> out;
			vector<string> err;
			ml.CallHandler( "service", name, "get_provides", "", out, err );

			if ( out.size() == 0) {
				LOGDEBUG( "could not determine 'provides' for service "+name+
									" (using service name instead)" );
				provides.push_back( name );
			}
			else {
				provides = out;
			}
		}
		catch( Modlib::handler_not_available ) {
			LOGDEBUG( "could not determine 'provides' for service "+name+
								" (using service name instead)" );
			provides.push_back( name );
			return;
		}
	}
}

Service::~Service( )
{

}

bool Service::UpToDate( ) throw( Resource::not_available )
{
		// for the service resource, UpToDate is given in case the resource
		// is unmodified an its state (running/offline) is the same as the saved one
	if ( this->Modified() ) return false;
	bool run = this->Running();
	bool act = this->Active();
	if ( run == act ) return true;
	return false;
}

bool Service::Active( ) throw( Resource::not_available )
{
	Modlib ml;

	char runlevel = this->GetRunlevel();
	if ( runlevel == '0' ) {
		LOGWARN( "could not determine current runlevel" );
		return false;
	}

		// if profile is unset, check the actual status
	if ( profile == "__unset" ) {
		try {
			vector<string> out;
			vector<string> err;
			if ( ml.CallHandler( "service", name, "active", "", out, err ) != 0 ) {
				LOGDEBUG( "handler for active state (file "+name+") returned non-zero" );
				if (out.size() ) {
					LOGDEBUG( "output was:" );
					for (unsigned int i=0; i<out.size(); i++) LOGDEBUG( out[i] );
				}
				LOGWARN( "could not query active state of service "+name );
				LOGWARN( "assuming service "+name+" is down" );
				return false;
			}
			for (unsigned int i=0; i<out.size(); i++) {
				if ( out[i].at( 0 ) == runlevel ) {
					if ( out[i].find( "on" ) != string::npos ) 
						return true;
					if ( out[i].find( "off" ) != string::npos ) 
						return false;
				}
			}
			LOGDEBUG( "could not find active status for current runlevel "+runlevel );
			LOGWARN( "could not query active state of service "+name );
			LOGWARN( "assuming service "+name+" is down" );
			return false;
		}
		catch( Modlib::handler_not_available ) {
			LOGWARN( "could not query active status of service "+name );
			LOGWARN( "assuming 'not active' for service "+name );
			return false;
		}
	} 
		// profile is set, therefor the saved active status for this one is wanted
	SCDB *db = SCDB::GetHandle();
	stringstream s;
	try {
		db->FileRestore( profile, "service", name, s );
		string slink;
		while ( s >> slink ) {
			if ( slink.at(0) == runlevel ) {
				if ( slink.substr( 2 ) == "on" )
					return true;
				else 
					return false;
			}
		}
		LOGWARN( "could not determine saved active state for service "+name );
		LOGWARN( "assuming 'not active'" );
		return false;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "could not access saved data for service "+name+" (reason: "+e.reason+")" );
		return false;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "could not access saved data for service "+name+" (no such file)" );
		return false;
	}
	return false;
}

bool Service::Modified( ) throw( Resource::not_available )
{
	SCDB *db = SCDB::GetHandle();
	Modlib ml;
	stringstream s;
	vector<string> stats;
	vector<string> err;

	try {
		if ( ml.CallHandler( "service", name, "active", "", stats, err ) == 2 ) {
			LOGDEBUG( "service "+name+" is not installed" );
			LOGDEBUG( "treating resource "+name+" (service)"+" as never changing" );
			return false;
		}
		for (unsigned int i=0; i<stats.size(); i++) s << stats[i] << endl;

		string stored_md5 = db->FileGetMD5( profile, string("service"), name );
		string actual_md5 = ml.GetMD5sum( s );
		if ( actual_md5 != stored_md5 ) {
			LOGDEBUG( "service "+name+" differs" );
			for (unsigned int i=0; i<stats.size(); i++) LOGDEBUG( stats[i] );
			return true;
		}
		if ( this->GetStartPriority() != this->CalcStartPriority() ) {
			LOGDEBUG( "service "+name+" differs in start priority" );
			return true;
		}
		if ( this->GetStopPriority() != this->CalcStopPriority() ) {
			LOGDEBUG( "service "+name+" differs in stop priority" );
			return true;
		}
		return false;
	}
	catch ( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGERR( "could not query saved status for service &"+name );
		return true;
	}
	catch ( SCDB::no_such_file ) {
		LOGWARN( "no such resource "+name+" (service) in profile "+profile	);
		return true;
	}
	catch ( Modlib::handler_not_available ) {
		LOGWARN( "no 'active' handler for resource "+name+" (service)" );
		LOGERR( "could not query saved status for service &"+name );
		return true;
	}
}

bool Service::Running( ) throw( Resource::not_available )
{
	int retval;
	vector<string> out;
	vector<string> err;
	Modlib ml;

	try {
		retval = ml.CallHandler( "service", name, "status", "", out, err );
	}
	catch ( Modlib::handler_not_available ) {
		LOGWARN( "could not determine status of resource "+name+
						 " (service), assuming resource is down" );
		return false;
	}

	if ( retval == 0 ) return true;

	LOGDEBUG( "service "+name+" not running" );
	return false;
}

bool Service::Start( )	throw( Resource::not_available )
{
	int retval;
	vector<string> out;
	vector<string> err;
	Modlib ml;

	try {
		retval = ml.CallHandler( "service", name, "start", "", out, err );
	}
	catch ( Modlib::handler_not_available ) {
		LOGWARN( "could not execute start handler for resource "+name+" (service)" );
		return false;
	}

	if ( retval == 0 ) return true;

	if ( out.size() ) LOGDEBUG( "output of start handler was" );
	for( unsigned int i=0; i<out.size(); i++ )
		LOGDEBUG( out[i] );
	LOGWARN( "could not start service "+name );
	return false;
}

bool Service::Stop( ) throw( Resource::not_available )
{
	int retval;
	vector<string> out;
	vector<string> err;
	Modlib ml;

	LOGDEBUG( "stopping service "+name );

	try {
		retval = ml.CallHandler( "service", name, "stop", "", out, err );
	}
	catch ( Modlib::handler_not_available ) {
		LOGWARN( "could not execute stop handler for resource "+name+" (service)" );
		return false;
	}

	if ( retval == 0 ) return true;

	if ( out.size() ) LOGDEBUG( "output of stop handler was" );
	for( unsigned int i=0; i<out.size(); i++ )
		LOGDEBUG( out[i] );
	LOGWARN( "could not stop service "+name );
	return false;
}

bool Service::Restart( )	throw( Resource::not_available )
{
	int retval;
	vector<string> out;
	vector<string> err;
	Modlib ml;

	try {
		retval = ml.CallHandler( "service", name, "restart", "", out, err );
	}
	catch ( Modlib::handler_not_available ) {
		LOGWARN( "could not execute restart handler for resource "+name+" (service)" );
		return false;
	}

	if ( retval == 0 ) return true;

	LOGWARN( "could not reload service "+name );
	return false;
}

vector<string> Service::GetDependencies( string resource_type )	 throw( Resource::not_available )
{
	vector<string> deps;
	vector<string> err;
	try {
		Modlib ml;
		ml.CallHandler( "service", name, "get_deps", resource_type, deps, err );
	}
	catch( Modlib::handler_not_available ) {
		LOGWARN( "could not determine dependencies for service "+name );
		deps.clear();
	}

	return deps;
}

int Service::GetStartPriority( ) throw( Resource::not_available )
{
	if (profile=="unset") return 99;
	int prio;
	try {
		SCDBFile *fileh = SCDB::GetHandle()->FileGetHandle( profile, "service", name );
		prio = fileh->GetStartPriority();
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGWARN( "could not query saved priority for service &"+name );
		return 99;
	}
	return prio;
}

int Service::GetStopPriority( ) throw( Resource::not_available )
{
	if (profile=="unset") return 0;
	int prio;
	try {
		SCDBFile *fileh = SCDB::GetHandle()->FileGetHandle( profile, "service", name );
		prio = fileh->GetStopPriority();
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGWARN( "could not query saved priority for service &"+name );
		return 0;
	}
	return prio;
}

int Service::CalcStartPriority( )
{
	SCPM_conf *cfg = SCPM_conf::GetHandle();
	string link = this->GetStartLink( cfg->GetInitDir()+"/rc"+this->GetRunlevel()+".d" );

	if ( link == "__unset" ) {
		LOGWARN( "could not determine priority for service "+name );
		return 99;
	}
	char linkno[3];
	linkno[0] = link.at( link.find_last_of("/") + 2 );
	linkno[1] = link.at( link.find_last_of("/") + 3 );
	linkno[2] = '\0';
	return atoi( linkno );
}

int Service::CalcStopPriority( )
{
	SCPM_conf *cfg = SCPM_conf::GetHandle();
	string link = this->GetStopLink( cfg->GetInitDir()+"/rc"+this->GetRunlevel()+".d" );

	if ( link == "__unset" ) {
		LOGWARN( "could not determine priority for service "+name );
		return 0;
	}
	char linkno[3];
	linkno[0] = link.at( link.find_last_of("/") + 2 );
	linkno[1] = link.at( link.find_last_of("/") + 3 );
	linkno[2] = '\0';
	return atoi( linkno );
}

void Service::UpdateData( ) throw( Resource::not_available )
{
	resource_data.clear();

	for ( unsigned int i=0; i<provides.size(); i++ ) {
		data_element d;
		d.data_name = "provides";
		d.data_value = provides[i];
		resource_data.push_back( d );
	}
}

bool Service::Save( ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;

	Modlib ml;
	SCDB *db = SCDB::GetHandle();
	SCDBFile *fileh;

	try {
		if (!db->FileExists(profile, "service", name)) {
			LOGDEBUG( "resource "+name+" (service) does not exits, no backup will be made" );
			fileh = db->FileGetHandle( profile, "service", name, true );
		}
		else
			fileh = db->FilePushBackup( profile, "service", name );
		vector<string> stats;
		vector<string> err;
		stringstream s;
		ml.CallHandler( "service", name, "active", "", stats, err );
		for( unsigned int i=0; i<stats.size(); i++ ) s << stats[i] << endl;
		fileh->SetSubtype( "regular" ); // this is the only one
		fileh->ClearContent();
		fileh->Save( s, true );
		fileh->SetStopPriority( this->CalcStopPriority() );
		fileh->SetStartPriority( this->CalcStartPriority() );
		delete fileh;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGERR( "could not save resource &"+name+" (service)" );
		return false;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "scdb error (internal error)" );
		LOGERR( "could not save resource &"+name+" (service)" );
		return false;
	}
	catch( Modlib::handler_not_available ) {
		delete fileh;
		LOGWARN( "could not query active status of service "+name );
		LOGERR( "could not save resource &"+name+" (service)" );
		return false;
	}
	catch( SCDBFile::write_error ) {
		delete fileh;
		LOGWARN( "scdb error (write error)" );
		LOGERR( "could not save resource &"+name+" (service)" );
		return false;
	}
	return true;
}

bool Service::Restore( ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;

	Modlib ml;
	SCDB *db = SCDB::GetHandle();

	SCDBFile *fileh=NULL;
	try {
		stringstream s;
		fileh = db->FileGetHandle( profile, type, name );
		fileh->Restore( s );
		vector<string> out;
		vector<string> err;
		string args;
		string tmp;
		while ( s >> tmp ) args+=tmp+" ";
		if (ml.CallHandler( "service", name, "restore", args, out, err ) != 0) {
			if (out.size()) LOGDEBUG( "output of restore handler was:" );
			for (unsigned int i=0; i<out.size(); i++)
				LOGDEBUG( out[i] );
			LOGWARN( "could not restore resource &"+name+" (service)" );
			delete fileh;
			return false;
		}
		if (fileh->GetStartPriority()==-1) {
			fileh->SetStartPriority( this->CalcStartPriority() );
			fileh->SetStopPriority( this->CalcStopPriority() );
		}
		delete fileh;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "could not access saved service "+name+" (no such file)" );
		LOGWARN( "could not restore resource &"+name+" (service)" );
		return false;
	}
	catch( SCDB::db_action_failed &e ) {
		if (fileh) delete fileh;
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGWARN( "could not restore resource &"+name+" (service)" );
		return false;
	}
	catch( Modlib::handler_not_available ) {
		delete fileh;
		LOGWARN( "could not call restore handler for service "+name );
		LOGWARN( "could not restore resource &"+name+" (service)" );
		return false;
	}
	catch( SCDBFile::read_error ) {
		delete fileh;
		LOGWARN( "SCDB read error on restoring service "+name );
		LOGWARN( "could not restore resource &"+name+" (service)" );
		return false;
	}
	catch( SCDBFile::write_error ) {
		delete fileh;
		LOGWARN( "SCDB write error on restoring service "+name );
		LOGWARN( "could not restore resource &"+name+" (service)" );
		return false;
	}
	return true;
}

string Service::GetStartLink( string directory )
{
	DIR *dp;
	struct dirent *ep;
	dp = opendir( directory.c_str() );

	while ( (ep = readdir( dp )) ) {
		string fname = ep->d_name;
		if ( (fname.find( name ) != string::npos) && (fname.find("S")==0) ) {
			closedir( dp );
			return directory+"/"+fname;
		}
	} ;
	closedir( dp );
	return "__unset";
}

string Service::GetStopLink( string directory )
{
	DIR *dp;
	struct dirent *ep;
	dp = opendir( directory.c_str() );

	while ( (ep = readdir( dp )) ) {
		string fname = ep->d_name;
		if ( (fname.find( name ) != string::npos) && (fname.find("K")==0) ) {
			closedir( dp );
			return directory+"/"+fname;
		}
	} ;
	closedir( dp );
	return "__unset";
}

char Service::GetRunlevel( )
{
	struct utmp *ut;
	char runlevel='0';

	setutent();
	while ( ((ut = getutent()) != NULL) && (runlevel=='0') ) {
		if (ut->ut_type == RUN_LVL) {
			runlevel = ut->ut_pid % 256;
			endutent();
		}
	}

	return runlevel;
}

bool Service::AddBackup( ) throw( Resource::not_available )
{
	SCDBFile *fileh=NULL;
	SCDB *db = SCDB::GetHandle();
	Modlib ml;

	try {
		fileh = db->FileGetBackupHandle( profile, type, name, true );
		vector<string> stats;
		vector<string> err;
		stringstream s;
		ml.CallHandler( "service", name, "active", "", stats, err );
		for( unsigned int i=0; i<stats.size(); i++ ) s << stats[i] << endl;
		fileh->SetSubtype( "regular" ); // this is the only one
		fileh->ClearContent();
		fileh->Save( s, true );
		delete fileh; fileh=NULL;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error, reason "+e.reason );
		LOGWARN( "could not create backup for resource "+name+" (file)" );
	}
	catch( SCDBFile::read_error ) {
		LOGWARN( "storage read error" );
		LOGWARN( "could not create backup for resource "+name+" (file)" );
	}
	catch( SCDBFile::write_error ) {
		LOGWARN( "storage write error" );
		LOGWARN( "could not create backup for resource "+name+" (file)" );
	}
	if (fileh) delete fileh;
	return true;
}

bool Service::HasBackup( ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) {
		LOGWARN( "profile is unset, can not check for backup files" );
		return false;
	}
	SCDB *db = SCDB::GetHandle();
	return db->FileHasBackup( profile, "service", name );
}

vector< pair<time_t, string> > Service::GetBackupVersions( ) throw( Resource::not_available )
{
	vector< pair<time_t, string > > v;
	if (profile == "__unset" ) {
		LOGWARN("profile is unset, can not check for backup files");
		return v;
	}
	SCDB *db = SCDB::GetHandle();
	try{
		// get all backups for this resource
		db->FileGetBackupVersions( profile, "service", name, v);
	}
	catch(SCDB::db_action_failed &e ) {
		LOGWARN("database I/O error ("+e.reason+")");
		LOGWARN("could not query backups for resource &"+name+" (file)");
	}
	
	return v;
}

bool Service::RestoreBackup(const string* version ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) {
		LOGWARN( "profile is unset, can not restore backup files" );
		return false;
	}
	SCDB *db = SCDB::GetHandle();
	try {
		if (!db->FileHasBackup( profile, "service", name )) {
			LOGWARN( "resource "+name+" (service) does not have any backups" );
			return false;
		}
		db->FilePopBackup( profile, "service", name, version );
		this->Restore();
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error ("+e.reason+")" );
		LOGWARN( "could not prepare backup restore of resource &"+name+" (file)" ); 
		return false;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "scdb error (file not found)" );
		LOGWARN( "this should not happen and may be a bug" );
		LOGWARN( "could not prepare backup restore of resource &"+name+" (file)" ); 
		return false;
	}
	return true;
}

bool Service::RemoveBackup( const string * version) throw( Resource::not_available )
{
 	if ( profile == "__unset" ) {
		LOGWARN( "profile is unset, can not restore backup files" );
		return false;
	}
	SCDB *db = SCDB::GetHandle();
	try {
		if (!db->FileHasBackup( profile, "service", name )) {
			LOGWARN( "resource "+name+" (service) does not have any backups" );
			return false;
		}
		return db->FileRemoveBackup(profile, "service", name, version);
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "database I/O error ("+e.reason+")" );
		LOGWARN( "could not prepare backup restore of resource &"+name+" (service)" ); 
		return false;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "scdb error (file not found)" );
		LOGWARN( "this should not happen and may be a bug" );
		LOGWARN( "could not prepare backup restore of resource &"+name+" (service)" ); 
		return false;
	}
}

bool Service::Refresh( ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;

	SCDB *db = SCDB::GetHandle();
	SCDBFile *fileh;

	try {
		if (!db->FileExists(profile, "service", name)) {
			LOGDEBUG( "resource "+name+" (service) does not exits, no backup will be made" );
			fileh = db->FileGetHandle( profile, "service", name, true );
		}
		else
			fileh = db->FilePushBackup( profile, "service", name );
		fileh->SetMD5sum( fileh->RealMD5sum() );
		delete fileh;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		return false;
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "scdb error (internal error)" );
		return false;
	}
	catch( SCDBFile::write_error ) {
		delete fileh;
		LOGWARN( "scdb error (write error)" );
		return false;
	}
	return true;
}

void Service::ShowChanges( ostream &output, bool backup, string version ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return;
	if ( !backup && !this->Modified() ) return;

	SCDB *db = SCDB::GetHandle();
	stringstream buf;

	SCDBFile *fileh = NULL;
	try {
		if (backup)
			fileh = db->FileGetBackupHandle( profile, type, name, false, (version=="" ? NULL : &version ) );
		else
			fileh = db->FileGetHandle( profile, type, name, false );
		fileh->Restore( buf );
		vector<string> out;
		vector<string> err;
		Modlib ml;
		if ( ml.CallHandler( "service", name, "active", "", out, err ) != 0 ) {
			LOGDEBUG( "handler for active state (file "+name+") returned non-zero" );
			if (out.size() ) {
				LOGDEBUG( "output was:" );
				for (unsigned int i=0; i<out.size(); i++) LOGDEBUG( out[i] );
			}
			LOGWARN( "could not query active state of service "+name );
			output << "service " << name << ": status query failed" << endl;
			delete fileh;
			return;
		}
		string foo;
		output << "service " << name << endl
					 << " current     saved" << endl;
		for (unsigned int i=0; i<=5; i++) {
			output << " " << out[i];
			for (unsigned int j=0; j<12-out[i].size(); j++) output << " ";
			buf >> foo;
			output << foo << endl;
		}
		output << " start priority: " << this->CalcStartPriority() << " (current) "
					 << this->GetStartPriority() << " (saved)" << endl;
		output << " stop priority: " << this->CalcStopPriority() << " (current) "
					 << this->GetStopPriority() << " (saved)" << endl;
	}
	catch( SCDB::no_such_file ) {
		LOGDEBUG( "resource "+name+" (service) (or backup of) not in storage" );
		output << "service " << name << endl;
		if (backup) 
			output << LOGTRANS( "  service has no backup in profile &"+profile ) << endl;
		else
			output << LOGTRANS( "  service does not exist in profile &"+profile ) << endl;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		output << "service " << name << endl;
		output << LOGTRANS( "  could not access file in profile &"+profile ) << endl;
	}
	if (fileh) delete fileh;
}

bool Service::Export( const string &root ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;

	SCDB *db = SCDB::GetHandle();
	try {
		ofstream os( (root+"/"+name).c_str(), ios::trunc );
		db->FileRestore( profile, type, name, os );
		os.close();
	}
	catch( SCDB::no_such_file ) {
		LOGWARN( "could not access saved service "+name+" (no such file)" );
		LOGWARN( "could not export resource "+name+" (service)" );
		return false;
	}
	catch( SCDB::db_action_failed &e ) {
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGWARN( "could not export resource "+name+" (service)" );
		return false;
	}
	return true;
}

bool Service::Import( const string &root ) throw( Resource::not_available )
{
	if ( profile == "__unset" ) return false;

	SCDB *db = SCDB::GetHandle();
	SCDBFile *fileh = NULL;

	try {
		ifstream is( (root+"/"+name).c_str() );
		if (is.bad() ) return false;
		fileh=db->FileGetHandle( profile, type, name, true );
		fileh->SetSubtype( "regular" ); // this is the only one
		fileh->ClearContent();
		fileh->Save( is, true );
		fileh->SetStopPriority( this->CalcStopPriority() );
		fileh->SetStartPriority( this->CalcStartPriority() );
		delete fileh;
		is.close();
	}
	catch( SCDB::db_action_failed &e ) {
		if (fileh) delete fileh;
		LOGWARN( "scdb error ("+e.reason+")" );
		LOGWARN( "could not import resource "+name+" (service)" );
		return false;
	}
	return true;
}
