/*
 * $Id: local_file.cc,v 1.2 2006/09/21 07:47:59 hschaa Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	SCPM database
 * File				 :	scdb_file.hh
 * Description :	SCDB storage interface
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2002-2004 SuSE Linux AG
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "local_file.hh"
#include <scpm_conf.hh>
#include <log.hh>
#include <iostream>
#include <fstream>
#include <modlib.hh>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <utime.h>
#include <cstring>

using namespace std;

#define module_name "local_file"

LocalFile::LocalFile( SCDBNode<backend_node> &node, const string &resname,
											const string &restype, const string &profile_name, bool &modified, bool is_backup )
		: SCDBFile( node, resname, restype, profile_name, modified, is_backup	 )
{
	if ( read_location.empty() ) {
			// maybe we are a new one, so no location has been generated yet
		read_location = SCPM_conf::GetHandle()->GetProfileDir()+"/"+profile+"/"+type+"/"+
			this->EncodeFilename( name );
	}
}

string &LocalFile::RealMD5sum()
{
	static string md5;
	Modlib ml;
	if ( read_location.empty() ) md5.clear();
	else md5 = ml.GetMD5sum( read_location );
	return md5;
}

void LocalFile::Save( istream &ins, bool auto_md5, string suffix ) throw( SCDBFile::write_error )
{
		// maybe we need to create dirs
	if ( write_location.empty() ) this->GenerateWriteLocation(suffix);

	int fd;
	if ( (fd=::open( write_location.c_str(), O_WRONLY|O_TRUNC|O_CREAT, 0600) ) == -1 ) {
		LOGDEBUG( "could not open file "+write_location );
		throw write_error();
	}
	char buf[4097];
	int count;
	int written;
	while ( !ins.eof() ) {
		ins.read( buf, 4096 );
		count = ins.gcount();
		if ( count ) {
			written = ::write( fd, buf, count );
			if ( written != count ) {
					// error (disk full?)
				LOGDEBUG( "writing in "+write_location+" failed (disk full?)" );
				::close( fd );
				throw write_error();
			}
		}
	}
	::close( fd );
		// update md5 sum (if wanted)
	if ( auto_md5 ) {
		Modlib ml;
		md5sum=ml.GetMD5sum( write_location );
	}
		// read location may be outdated now
	read_location=write_location;
}

void LocalFile::Restore( ostream &outs ) throw( SCDBFile::read_error, SCDBFile::write_error )
{
	int fd;
	if ( (fd=::open( read_location.c_str(), O_RDONLY) ) == -1 ) {
		LOGDEBUG( "could not open file "+read_location );
		throw read_error();
	}
	char buf[4097];
	int count;
	do {
		count = ::read( fd, buf, 4096 );
		if ( count ) {
			outs.write( buf, count );
			if ( outs.bad() ) {
					// error (disk full?)
				LOGDEBUG( "write failed (disk full?)" );
				::close( fd );
				throw write_error();
			}
		}
	} while( count );
	::close( fd );
}

void LocalFile::Remove( ) throw( SCDBFile::write_error )
{
	if ( unlink( read_location.c_str() ) == -1 ) {
		if ( errno != ENOENT ) {
				// unlink failed and not because of missing file
			LOGDEBUG( "could not unlink "+write_location+" ("+std::strerror(errno)+")" );
			throw write_error();
		}
	}
}

void LocalFile::Create( string suffix ) throw( SCDBFile::write_error )
{
		// maybe we need to create dirs
	if ( write_location.empty() ) this->GenerateWriteLocation(suffix);
	int fd = ::open( write_location.c_str(), O_WRONLY|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR );
	if ( fd == -1 ) {
		LOGDEBUG( "could not create "+write_location+" ("+std::strerror(errno)+")" );
		throw write_error();
	}
	::close( fd );
	read_location=write_location;
}

void LocalFile::SaveAttribs( struct stat &attribs ) throw( SCDBFile::write_error )
{
	if ( write_location.empty() ) this->GenerateWriteLocation();
	struct utimbuf ub;
	ub.actime = attribs.st_atime;
	ub.modtime = attribs.st_mtime;
	if ( utime( write_location.c_str(), &ub ) != 0 ) {
		LOGDEBUG( "could not set utime on "+write_location );
		throw write_error();
	}
	if ( chmod( write_location.c_str(), attribs.st_mode ) != 0 ) {
		LOGDEBUG( "could not chmod "+write_location );
		throw write_error();
	}
	if ( chown( write_location.c_str(), attribs.st_uid, attribs.st_gid ) != 0 ) {
		LOGDEBUG( "could not chown "+write_location );
		throw write_error();
	}
	read_location=write_location;
}

struct stat &LocalFile::GetAttribs( ) throw( SCDBFile::read_error )
{
	static struct stat attribs;
	if ( ::stat( read_location.c_str(), &attribs ) != 0 ) {
		LOGDEBUG( "could not stat "+read_location+" ("+std::strerror(errno)+")" );
		throw read_error();
	}
	return attribs;
}

string &LocalFile::EncodeFilename( string &s )
{
	static string n;
	unsigned int i=0;
	n.clear();
	n.reserve( s.size() + 10 );
	while( i<s.size() ) {
		if ( s.at(i) == '_' ) {
			n += "__";
		}
		else if ( s.at(i) == '/' ) n += '_';
		else n += s.at(i);
		i++;
	}
	return n;
}

void LocalFile::GenerateWriteLocation(string suffix) throw( write_error )
{
	struct stat status_buf;
	string base_dir = SCPM_conf::GetHandle()->GetProfileDir()+"/"+profile;
	if ( stat( (base_dir+"/"+type).c_str(), &status_buf ) == -1 ) {
		if ( stat( (base_dir).c_str(), &status_buf ) == -1 ) {
			if ( mkdir( (base_dir).c_str(), 0750 ) !=0 ) {
				LOGWARN( "could not make dir "+base_dir );
				throw write_error();
			}
		}
		if ( mkdir( (base_dir+"/"+type).c_str(), 0750 ) !=0 ) {
			LOGWARN( "could not make dir "+base_dir+"/"+type );
			throw write_error();
		}
	}
	else if (! S_ISDIR( status_buf.st_mode ) ) {
		LOGWARN( base_dir+"/"+type+" is not a directory" );
		throw write_error();
	} 
	write_location = base_dir + "/" + type + "/" + this->EncodeFilename( name );
		// maybe we need a backup extension
		write_location += this->GetSuffix( write_location, suffix );
}

istream *LocalFile::GetStream( ) throw( SCDBFile::read_error )
{
	static ifstream *in;
	struct stat buf;
	if ( stat( read_location.c_str(), &buf ) == -1 ) throw read_error();
	in = new ifstream( read_location.c_str() );
	if ( in->bad() ) throw read_error();
	return in;
}

string LocalFile::GetSuffix( string &fname, string suffix )
{
	struct stat buf;
	if ( stat( (fname+suffix).c_str(), &buf ) == 0)
	{
		int i=1;
		while( stat( (fname+"."+itoa(i)).c_str(), &buf ) == 0 ) {
			i++;
		}
		suffix = "."+itoa(i);
	}
	return suffix;
}

string LocalFile::GetBackupVersion()
{
	// get the original filename the backup would have without suffix
	string orgname = this->EncodeFilename( name );
	string loc = read_location;
	// delete encoded filname out of real filename to get the suffix (may be empty)
	loc.erase(0, loc.find(orgname) + orgname.size());
	
	return loc;

}
