/*
 * $Id: scdb.cc,v 1.6 2006/09/27 14:45:01 hschaa Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	SCPM database
 * File				 :	scdb.cc
 * Description :	implements SCDB access and management functions
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2002-2004 SuSE Linux AG
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "scdb.hh"
#include "xml.hh"
#include <log.hh>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdio>
#include <scpm_conf.hh>
#include <utime.h>
#define module_name "scdb"

using namespace std;

SCDB::SCDB( )
{
	db = NULL;
	modified = false;
	backend = new XML( );
	root = backend->GetRoot();
}

SCDB::~SCDB( )
{
	delete backend;
	if (db) delete db;
}

int SCDB::LoadDB( ) throw( db_action_failed )
{
	SCPM_conf *cfg = SCPM_conf::GetHandle();
	return this->LoadDB( cfg->GetDBFile() );
}

int SCDB::LoadDB( string dbfile ) throw( db_action_failed )
{
	db = new ifstream( dbfile.c_str() );
	if ( db->good() ) {
		return this->LoadDB( *db, true );
			// do not clean up ifstream object as we may need it later for
			// branch on demand loading
	}
	else {
		LOGWARN( "could not open "+dbfile );
		delete db;
		db=NULL;
		return db_state_open_failed;
	}
}

int SCDB::LoadDB( istream &input, bool quick ) throw( db_action_failed )
{
	if (!backend->Load( input, true )) return db_state_corrupt;
	string version_string = backend->GetSCDBVersion();
	if ( (version_string != "1.3") )
		return db_state_version_unsupported;
	return db_state_ok;
}

void SCDB::SaveDB( ) throw( db_action_failed )
{
	SCPM_conf *cfg = SCPM_conf::GetHandle();
	char s[255];
	std::sprintf( s, "%s.XXXXXX", cfg->GetDBFile().c_str() );

	if ( mktemp( s ) == NULL ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "mktemp failed" );
		throw db_action_failed( "mktemp failed" );
	}		 

	ofstream tmpdb( s );

	if ( tmpdb.bad() ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, (string)"could not open "+s );
		throw db_action_failed( "database I/O error" );
	}

	backend->Save( tmpdb );
	if (tmpdb.bad() ) {
		tmpdb.close();
		throw db_action_failed( "database I/O error" );
	}
	tmpdb.close();

	ifstream in( s );
	ofstream out( cfg->GetDBFile().c_str() );

	if ( out.bad() ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not open "+cfg->GetDBFile() );
		throw db_action_failed( "database I/O error" );
	}

	out << in.rdbuf();
	in.close();
	out.close();
	modified = false;
}

void SCDB::SaveDB( ostream &output ) throw( db_action_failed )
{
	backend->Save( output );
	if (output.bad() )
		throw db_action_failed( "database I/O error" );
	modified = false;
}

void SCDB::DropDB( )
{
	delete backend;
	backend = new XML( );
	root = backend->GetRoot();
	modified = false;
}

string SCDB::GetSCDBVersion( )
{ 
	return backend->GetSCDBVersion();
}

void SCDB::SetSCDBVersion( string version )
{
	backend->SetSCDBVersion( version );
}

string SCDB::GetFormatVersion( )
{ 
	return backend->GetFormatVersion();
}

bool SCDB::UpToDate( )
{
	if ( backend->GetSCDBVersion() != "1.3" ) return false;
	if ( ! backend->UpToDate() ) return false;
	return true;
}
void SCDB::SetStatusFlag( string status_flag, bool value ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*status*"+status_flag );
		if ( value )
			node.SetValue( "yes" );
		else
			node.SetValue( "no" );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not set status flag "+status_flag );
		throw db_action_failed( "key not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not set status flag "+status_flag );
		throw db_action_failed( "key not unique" );
	}
}

bool SCDB::GetStatusFlag( string status_flag ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*status*"+status_flag );
		if ( node.GetValue() == "yes" )
			return true;
		else if ( node.GetValue() == "no" )
			return false;
		else {
			Log *log = Log::GetHandle();
			log->WriteMessage( "scdb", Log::CLASS_DEBUG, "status flag "+status_flag+
												 "contains senseless value "+node.GetValue() );
			throw db_action_failed( "db corrupt" );
		}
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not get status flag "+status_flag );
		throw db_action_failed( "key not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not get status flag "+status_flag );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::SetStatusKey( string status_key, string value ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*status*"+status_key );
		node.SetValue( value );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not set status key "+status_key );
		throw db_action_failed( "key not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not set status key "+status_key );
		throw db_action_failed( "key not unique" );
	}
}

string SCDB::GetStatusKey( string status_key ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*status*"+status_key );
		return node.GetValue();
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not get status flag "+status_key );
		throw db_action_failed( "key not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not get status key "+status_key );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::SetActiveProfile( string profile_name ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*status*active-profile" );
		node.SetValue( profile_name );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not set actve profile" );
		throw db_action_failed( "key not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not set active profile" );
		throw db_action_failed( "key not unique" );
	}
}

string SCDB::GetActiveProfile( ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*status*active-profile" );
		return node.GetValue( );
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not get active profile" );
		throw db_action_failed( "key not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not get active profile" );
		throw db_action_failed( "key not unique" );
	}
}

vector<string> SCDB::GetProfiles( ) throw( db_action_failed )
{
	vector<string> profiles;
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
			for ( unsigned int j=0; j<node[i].GetNumSub(); j++ ) {
				if ( (node[i][j].GetName() == "name") &&
						 !node[i].Deleted() && !node[i][j].Deleted() )
					profiles.push_back( node[i][j].GetValue() );
			}
		}
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find profile root node" );
		throw db_action_failed( "key not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not determine profile root node status flag" );
		throw db_action_failed( "key not unique" );
	}
	return profiles;
}

bool SCDB::ResourceExists( string resource_name, string resource_type )
	throw( db_action_failed )
{
	try {
		this->GetNode( "root*resources*"+resource_type+"*resource|name="+resource_name );
		return true;
	}
	catch( not_found ) {
		return false;
	}
	catch( not_unique ) {
		return true;
	}
}

void SCDB::ResourceAdd( string resource_name, string resource_type ) throw( db_action_failed )
{
	try {
		if ( this->KeyExists( "root*resources*"+resource_type+"*resource|name="+resource_name ) )
			throw db_action_failed( "resource exists" );
		SCDBNode<backend_node> &node = this->AddNode( "root*resources*"+resource_type, "resource" );
		backend_node *n1 = new backend_node( "name" );
		backend_node *n2 = new backend_node( "deps" );
		backend_node *n3 = new backend_node( "data" );
		node.AddSub( *n1 );
		n1->SetValue( resource_name );
		node.AddSub( *n2 );
		node.AddSub( *n3 );
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find resource type root node" );
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not add resource "+resource_name+
											 " ("+resource_type+")" );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "resource type root node not unique (uh?)" );
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not add resource "+resource_name+
											 " ("+resource_type+")" );
		throw db_action_failed( "db corrupt" );
	}
}

void SCDB::ResourceAddType( string resource_type ) throw( db_action_failed )
{
	try {
		if ( this->KeyExists( "root*resources*"+resource_type ) )
			throw db_action_failed( "resource type exists" );
		this->AddNode( "root*resources", resource_type );
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find resource root node" );
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not add resource type "
											 +resource_type );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "resource root node not unique (uh?)" );
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not add resource type "
											 +resource_type );
		throw db_action_failed( "db corrupt" );
	}
}

void SCDB::ResourceDelete( string resource_name, string resource_type ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*resources*"+resource_type+"*resource|name="+resource_name );
		node.SetDeleted( true );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find resource node" );
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not delete resource "+resource_name+
											 " ("+resource_type+")" );
		throw db_action_failed( "resource not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not determine resource node" );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::ResourceAddDep( string resource_name, string resource_type, string dep_name, string dep_type )
	throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*resources*"+resource_type
																									+"*resource|name="+resource_name+"*deps" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
			if ( (node[i].GetName() == resource_type) && (node[i].GetValue() == resource_name ) )
				throw db_action_failed( "resource dependency exists" );
		}
		backend_node *n1 = new backend_node( dep_type );
		n1->SetValue( dep_name );
		node.AddSub( *n1 );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find resource node" );
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not add resource dependency "
											 +dep_name+" ("+dep_type+") to "
											 +resource_name+" ("+resource_type+")" );
		throw db_action_failed( "resource not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not determine resource node" );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::ResourceCleanDeps( string resource_name, string resource_type ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*resources*"+resource_type+
																									"*resource|name="+resource_name+"*deps" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
			node[i].SetDeleted( true );
			modified = true;
		}
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not delete resource deps for "+resource_name+
											 " ("+resource_type+")" );
		throw db_action_failed( "resource not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not delete resource deps for "+resource_name+
											 " ("+resource_type+")" );
		throw db_action_failed( "key not unique" );
	}
}

vector<SCDB::resource_dep_t> SCDB::ResourceGetDeps( string resource_name, string resource_type )
	throw( db_action_failed )
{
	try {
		vector<resource_dep_t> deps;
		SCDBNode<backend_node> &node = this->GetNode( "root*resources*"+resource_type+"*resource|name="+resource_name+
																									"*deps" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
			if ( node[i].Deleted() ) continue;
			resource_dep_t d;
			d.type = node[i].GetName();
			d.name = node[i].GetValue();
			deps.push_back( d );
		}
		return deps;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find deps for resource "+resource_name+
											 " ("+resource_type+")" );
		throw db_action_failed( "resource not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find deps for resource "+resource_name+
											 " ("+resource_type+")" );
		throw db_action_failed( "key not unique" );
	}
}

vector<string> SCDB::ResourceGetTypes( ) throw( db_action_failed )
{
	vector<string> restypes;
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*resources" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++) {
			if ( node[i].Deleted() ) continue;
			restypes.push_back( node[i].GetName() );
		}
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not find resource root node" );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find resource root node" );
		throw db_action_failed( "key not unique" );
	}
	return restypes;
}

vector<string> SCDB::ResourceGetNames( string resource_type ) throw( db_action_failed )
{
	vector<string> resnames;
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*resources*"+resource_type );
		for ( unsigned int i=0; i<node.GetNumSub(); i++) {
			unsigned int j;
			for ( j=0; j<node[i].GetNumSub(); j++ ) {
				if ( (node[i][j].GetName() == "name") && !node[i].Deleted() && !node[i][j].Deleted() ) {
					resnames.push_back( node[i][j].GetValue() );
					break;
				}
			}
			if ( j == node[i].GetNumSub() ) {
				Log *log = Log::GetHandle();
				log->WriteMessage( "scdb", Log::CLASS_WARNING, 
													 "found a resource node without name subnode, db possibly corrupt" );
			}
		}
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not find resource type "
											 +resource_type );
		throw db_action_failed( "resource type not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not find resource type "+resource_type );
		throw db_action_failed( "key not unique" );
	}
	return resnames;
}

void SCDB::ResourceAddData( string resource_name, string resource_type,
														string data_entry, string data_value )
	throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*resources*"+resource_type+"*resource|name="
																									+resource_name+"*data" );
		backend_node *data_node = new backend_node( data_entry );
		data_node->SetValue( data_value );
		node.AddSub( *data_node );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not find data node of resource "+
											 resource_name+" ("+resource_type+")" );
		throw db_action_failed( "resource node/subnode not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not find data node of "+resource_name 
											 +" ("+resource_type+")" );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::ResourceClearData( string resource_name, string resource_type )
	throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*resources*"+resource_type+"*resource|name="
																									+resource_name+"*data" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
			node[i].SetDeleted( true );
			modified = true;
		}
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find date node of resource "+
											 resource_name+" ("+resource_type+")" );
		throw db_action_failed( "resource node/subnode not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find data node of resource "+resource_name 
											 +" ("+resource_type+")" );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::ResourceGetData( string resource_name, string resource_type,
														vector<string> &data_entries, vector<string> &data_values )
	throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*resources*"+resource_type+"*resource|name="
																									+resource_name+"*data" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
			if ( node[i].Deleted() ) continue;
			data_entries.push_back( node[i].GetName() );
			data_values.push_back( node[i].GetValue() );
		}
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find date node of resource "+
											 resource_name+" ("+resource_type+")" );
		throw db_action_failed( "resource node/subnode not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find data node of resource "+resource_name 
											 +" ("+resource_type+")" );
		throw db_action_failed( "key not unique" );
	}
}

bool SCDB::ProfileExists( string profile ) throw( db_action_failed )
{
	try {
		if ( this->KeyExists( "root*profiles*profile|name="+profile ) )
			return true;
		else
			return false;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not query profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not query profile "+profile );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::ProfileAdd( string profile_name ) throw( db_action_failed )
{
	try {
		if ( this->KeyExists( "root*profiles*profile|name="+profile_name ) )
			throw db_action_failed( "profile exists" );

		SCDBNode<backend_node> &parent = this->GetNode( "root*profiles" );
		backend_node *node = new backend_node( "profile" );
		parent.AddSub( *node );

		backend_node *node_name = new backend_node( "name" );
		backend_node *node_descr = new backend_node( "description" );
		backend_node *node_clean = new backend_node( "clean" );
		backend_node *node_active = new backend_node( "active" );
		backend_node *node_scripts = new backend_node( "scripts" );
//		backend_node *node_promagic = new backend_node( "promagic" );
//		backend_node *node_packmagic = new backend_node( "packmagic" );
		backend_node *node_resources = new backend_node( "resources" );

		node->AddSub( *node_name );
		node->AddSub( *node_descr );
		node->AddSub( *node_clean );
		node->AddSub( *node_active );
		node->AddSub( *node_scripts );
//		node->AddSub( *node_promagic );
//		node->AddSub( *node_packmagic );
		node->AddSub( *node_resources );

		backend_node *node_prestart = new backend_node( "prestart" );
		backend_node *node_poststart = new backend_node( "poststart" );
		backend_node *node_prestop = new backend_node( "prestop" );
		backend_node *node_poststop = new backend_node( "poststop" );

		node_scripts->AddSub( *node_prestart );
		node_scripts->AddSub( *node_prestop );
		node_scripts->AddSub( *node_poststart );
		node_scripts->AddSub( *node_poststop );

			// create data nodes
		SCDBNode<backend_node> &data_node = this->GetNode( "root*data" );
		backend_node *data_profile_node = new backend_node( "profile" );
		backend_node *data_name_node = new backend_node( "name" );

		data_node.AddSub( *data_profile_node );
		data_profile_node->AddSub( *data_name_node );
		data_name_node->SetValue( profile_name );

			// setting builtin defaults
		node_active->SetValue( "no" );
		node_clean->SetValue( "yes" );

		node_name->SetValue( profile_name );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not add profile "+profile_name );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not add profile "+profile_name );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::ProfileDelete( string profile_name ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile_name );
		node.SetDeleted( true );
		SCDBNode<backend_node> &node1 = this->GetNode( "root*data*profile|name="+profile_name );
		node1.SetDeleted( true );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not delete profile "+profile_name );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not delete profile "+profile_name );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::ProfileAddResource( string profile_name, string resource_name, string resource_type )
	throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile_name+"*resources" );
		for( unsigned int i=0; i<node.GetNumSub(); i++)
			if ( (node[i].GetName() == resource_type) && 
					 (node[i].GetValue() == resource_name) &&
					 (!node[i].Deleted() ) )
				throw db_action_failed( "profile resource exists" );
		backend_node *n1 = new backend_node( resource_type );
		n1->SetValue( resource_name );
		node.AddSub( *n1 );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find resources subnode of profile "
											 +profile_name );
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not add resource "+resource_name+
											 " ("+resource_type+") to profile "+profile_name );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find resource subnode of profile "+profile_name );
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not add resource "+resource_name+
											 " ("+resource_type+") to profile "+profile_name );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::ProfileDeleteResource( string profile_name, string resource_name, string resource_type )
	throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile_name+"*resources" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
			if ( (node[i].GetName() == resource_type) &&
					 (node[i].GetValue() == resource_name) ) {
				node[i].SetDeleted( true );
				modified = true;
			}
		}
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not remove resource "+resource_name+
											 " ("+resource_type+") from profile "+profile_name );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not remove resource "+resource_name+
											 " ("+resource_type+") from profile "+profile_name );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::ProfileClearResources( string profile_name, string resource_type )
	throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile_name+"*resources" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
			if ( (resource_type == "all") || (resource_type == node[i].GetName()) ) 
				node[i].SetDeleted( true );
			modified = true;
		}
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find resources subnode of profile "
											 +profile_name );
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could remove resources from profile "
											 +profile_name );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "could not find resource subnode of profile "
											 +profile_name );
		throw db_action_failed( "key not unique" );
	}
}

bool SCDB::ProfileHasResource( string profile_name, string resource_name, string resource_type )
	throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile_name+"*resources" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++ )
			if ( !node[i].Deleted() &&
					 node[i].GetName() == resource_type &&
					 node[i].GetValue() == resource_name ) return true;
		return false;
	}
	catch( not_found ) {
		throw db_action_failed( "no such profile" );
	}
	catch( not_unique ) {
		throw db_action_failed( "key not unique" );
	}
}

bool SCDB::ProfileGetFlag( string profile, string flag ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile+"*"+flag );
		if ( node.GetValue() == "yes" )
			return true;
		else if ( node.GetValue() == "no" )
			return false;
		else {
			Log *log = Log::GetHandle();
			log->WriteMessage( "scdb", Log::CLASS_WARNING, "flag "+flag+" of profile "+profile+
												 " contains senseless value "+node.GetValue() );
			throw db_action_failed( "db corrupt" );
		}
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not get flag "+flag+
											 " of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not get flag "+flag+
											 " of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
}

void SCDB::ProfileSetFlag( string profile, string flag, bool value ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile+"*"+flag );
		if ( value )
			node.SetValue( "yes" );
		else
			node.SetValue( "no" );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not set flag "+flag+
											 " of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not set flag "+flag+
											 " of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
}

string SCDB::ProfileGetKey( string profile, string key ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile+"*"+key );
		return node.GetValue();
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not get key "+key+
											 " of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not get key "+key+
											 " of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
}

void SCDB::ProfileSetKey( string profile, string key, string value ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile+"*"+key );
		node.SetValue( value );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not get key "+key+
											 " of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not get key "+key+
											 " of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
}

string SCDB::ProfileGetScript( string profile, string script_name )
	throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile+
																									"*scripts*"+script_name );
		return node.GetValue();
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not find "+script_name+
											 " script subnode of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not find "+script_name+
											 " script subnode of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
}

void SCDB::ProfileSetScript( string profile, string script_name, string script_location )
	throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile+
																									"*scripts*"+script_name );
		node.SetValue( script_location );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not find "+script_name+
											 " script subnode of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not find "+script_name+
											 " script subnode of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
}

vector<string> SCDB::ProfileGetResourceTypes( string profile )
	throw( db_action_failed )
{
	vector<string> restypes;
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile+"*resources" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
			bool found = false;
			for ( unsigned int j=0; j<restypes.size(); j++ ) 
				if ( (restypes[j] == node[i].GetName()) && !node[i].Deleted() ) found = true;
			if ( !found ) restypes.push_back( node[i].GetName() );
		}
		return restypes;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING,
											 "could not find resources subnode of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING,
											 "could not find resources subnode of profile "+profile );
		throw db_action_failed( "db corrupt" );
	}
}

vector<string> SCDB::ProfileGetResources( string profile, string resource_type )
	throw( db_action_failed )
{
	vector<string> resnames;
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*profiles*profile|name="+profile+"*resources" );
		for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
			if ( (node[i].GetName() == resource_type) &&
					 !node[i].Deleted() )
				resnames.push_back( node[i].GetValue() );
		}
		return resnames;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not find resources subnode of profile"+
											 profile );
		throw db_action_failed( "db corrupt" );
	}
}

void SCDB::Dump( ostream &output, string key ) throw( db_action_failed )
{
	vector<SCDBNode<backend_node>*> nc;
	string k;

		// check weather key has "root" prepended, if not, do so
	if ( key.find("root") == string::npos )
		k = "root*" + key;
	else
		k = key;

	if ( k == "root" ) {
			// full tree should be dumped
			// so we do not the SearchNodes() and also skip printing the root node
		for ( unsigned int i=0; i<root->GetNumSub(); i++) {
			DumpTree( (*root)[i], output );
		}
		return;
	}

	SearchNodes( *root, k, nc );

	for ( unsigned int i=0; i<nc.size(); i++ ) {
		DumpTree( *nc[i], output );
	}
}

void SCDB::SetValue( string key, string value ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( key );
		if ( node.GetNumSub() > 0 )
			throw db_action_failed( "key is not a leaf node" );
		node.SetValue( value );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "key "+key+" not found" );
		throw db_action_failed( "key not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "key "+key+" not found" );
		throw db_action_failed( "key not unique" );
	}
}

string SCDB::GetValue( string key ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( key );
		return node.GetValue();
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "key "+key+" not found" );
		throw db_action_failed( "key not found" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "key "+key+" not unique" );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::AddNodes( string key, string value ) throw( db_action_failed )
{
	try {
		string parent = key.substr( 0, key.find_last_of( "*" ) );
		string node_name = key.substr( key.find_last_of( "*" ) +1 );
		vector<SCDBNode<backend_node>*> nc;
		string k;
		if ( parent.find("root") == 0 )
			k = parent;
		else
			k = "root*" + parent;
		SearchNodes( *root, k, nc );
		if ( nc.size() == 0 ) throw db_action_failed( "parent node not found" );
		for ( unsigned int i=0; i<nc.size(); i++ ) {
			SCDBNode<backend_node> &n = AddNode( *nc[i], node_name );
			n.SetValue( value );
		}
	}
	catch( not_found ) {
		throw db_action_failed( "parent node not found" );
	}
}

void SCDB::DeleteNodes( string key ) throw( db_action_failed )
{
	try {
		vector<SCDBNode<backend_node>*> nc;
		string k;
		
		if ( key.find("root") == 0 )
			k = key;
		else
			k = "root*" + key;
		
		SearchNodes( *root, k, nc );
		for ( unsigned int i=0; i<nc.size(); i++ ) nc[i]->SetDeleted( true );
	}
	catch( not_found ) {
		Log::GetHandle()->WriteMessage( "scdb", Log::CLASS_DEBUG, 
																		"key "+key+" did not match any node" );
	}
}

void SCDB::RenameNodes( string key, string node_name ) throw( db_action_failed )
{
	try {
		vector<SCDBNode<backend_node>*> nc;
		string k;
		
		if ( key.find("root") == 0 )
			k = key;
		else
			k = "root*" + key;
		
		SearchNodes( *root, k, nc );
		for ( unsigned int i=0; i<nc.size(); i++ ) nc[i]->SetName( node_name );
	}
	catch( not_found ) {
		Log::GetHandle()->WriteMessage( "scdb", Log::CLASS_DEBUG, 
																		"key "+key+" did not match any node" );
	}
}

SCDBNode<SCDB::backend_node> &SCDB::GetNode( string key ) throw( not_found, not_unique )
{
	vector<SCDBNode<backend_node>*> nc;
	string k;

	if ( key.find("root") == 0 )
		k = key;
	else
		k = "root*" + key;

	SearchNodes( *root, k, nc );
	
	if ( nc.size() > 1 ) {
			// tag is not unique
		LOGDEBUG( "key "+key+" is not unique" );
		throw not_unique();
	}
	if ( nc.size() == 0 ) {
			// nc is empty
		throw not_found();
	}
		
	return *nc[0];
}


void SCDB::SearchNodes( SCDBNode<backend_node> &node, string tagspec,
												vector<SCDBNode<backend_node>*> &node_container)
	throw ( not_found )
{
	string key;
	string attr_name;
	string attr_value;
	bool	 attr_set;

		//LOGDEBUG( "entered SearchNodes(), node "+node.GetName()+" tagspec "+tagspec );

	if ( tagspec == "" ) {
			// we are at the end of the tagspec
		if ( node_container.size() > 0 )
			return;
		throw not_found();
	}

	if ( node.Deleted() ) {
			// we are on a deleted node, skipping
		return;
	}

		// grab the tag
	if ( tagspec.find("*") != string::npos )
		key = tagspec.substr( 0, tagspec.find("*") );
	else
		key = tagspec; // last one

		// check whether an attribute is wanted
	if ( key.find("|") != string::npos ) {
		if ( key.find("=") == string::npos ) {
			Log *log = Log::GetHandle();
			log->WriteMessage( "scdb", Log::CLASS_DEBUG, "invalid search key "+key );
			throw not_found();
		}
		attr_set = true;
		attr_name = key.substr( key.find("|")+1, key.find("=")-key.find("|")-1 );
		attr_value = key.substr( key.find("=")+1 );
			// setting key to the key itself
		key = key.substr( 0, key.find("|") );
	}
	else
		attr_set = false;

	if ( key != node.GetName() ) {
			// we left the right path
		return;
	}

	if ( !node.Available() ) {
			// we entered a branch which has not yet been loaded
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "branch at node "+key+" not available, trying to load" );
		try {
			backend->LoadBranch( static_cast<backend_node&>(node) );
		}
		catch( XML::op_failed ) {
			log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not load db branch, node "+key );
			throw not_found();
		}
	}

	if ( attr_set ) {
		bool found_attr=false;
			// check attributes
		for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
			if ( !node[i].Deleted() ) {
				if ( node[i].GetName() == attr_name ) {
					// found the key
					if ( node[i].GetValue() == attr_value ) {
						// even found the value
						if ( (key == node.GetName()) && (tagspec.find("*")==string::npos) ) {
							// this node exactly matches
							node_container.push_back( &node );
						}
						else found_attr=true;
					}
				}
			}
		}
		if ( found_attr )
			for ( unsigned int i=0; i<node.GetNumSub(); i++ )
				SearchNodes( node[i], tagspec.substr( tagspec.find("*")+1 ), node_container );
	}
	else {
			// no attribute set, this makes life easier
		if ( tagspec == node.GetName() ) {
			node_container.push_back( &node );
		}
		else {
				// no match, digging deeper
			for ( unsigned int i=0; i<node.GetNumSub(); i++ ) {
				SearchNodes( node[i], tagspec.substr( tagspec.find("*")+1 ), node_container );
			}
		}
	}
		// here means all subsearches are done, returning quietly
}

bool SCDB::KeyExists( string key )
{
	try {
		vector<SCDBNode<backend_node>*> nc;
		string k;

			// check weather key has "root" prepended, if not, do so
		if ( key.find("root") == string::npos )
			k = "root*" + key;
		else
			k = key;
		SearchNodes( *root, k, nc );
		if (nc.size()) return true;
		return false;
	}
	catch( not_found ) {
		return false;
	}
}

SCDBNode<SCDB::backend_node> &SCDB::AddNode( string key, string node_name )
	throw ( not_found, not_unique )
{
	vector<SCDBNode<backend_node>*> nc;
	string k;

	if ( key.find("root") == 0 )
		k = key;
	else
		k = "root*" + key;

	SearchNodes( *root, k, nc );
	
	if ( nc.size() > 1 ) {
			// tag is not unique
		LOGDEBUG( "key "+key+" is not unique" );
		throw not_unique();
	}

	if ( nc.size() == 0 ) {
		throw not_found();
	}

	backend_node *n = new backend_node( node_name );
	nc[0]->AddSub( *n );

	modified=true;

	return *n;
}

SCDBNode<SCDB::backend_node> &SCDB::AddNode( SCDBNode<backend_node> &parent, string node_name )
{
	backend_node *n = new backend_node( node_name );
	parent.AddSub( *n );
	modified=true;
	return *n;
}

void SCDB::DumpTree( SCDBNode<backend_node> &node, ostream &out )
{
	static int depth=0;

		// do not dump in case tree has deleted flag set
	if ( node.Deleted() ) return;

	if ( !node.Available() ) {
			// we entered a branch which has not yet been loaded
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, "branch at node "+node.GetName()+
											 " not avaible, trying to load" );
		try {
			backend->LoadBranch( static_cast<backend_node&>(node) );
		}
		catch( XML::op_failed ) {
			log->WriteMessage( "scdb", Log::CLASS_WARNING, "could not load db branch, node "
												 +node.GetName() );
			throw not_found();
		}
	}

	if ( node.GetValue() == "" ) {
			// key only
		for ( int i=0; i<depth*3; i++ ) out << " ";
		out << "<" << node.GetName() << ">" << endl;
		
			// search subs
		depth++;
		for (unsigned int i=0; i<node.GetNumSub(); i++)
			DumpTree( node[i], out );
		depth--;

		for ( int i=0; i<depth*3; i++ ) out << " ";
		out << "</" << node.GetName() << ">" << endl;
	}
	else {
		for ( int i=0; i<depth*3; i++ ) out << " ";
		out << "<" << node.GetName() << ">"
				<< node.GetValue()
				<< "</" << node.GetName() << ">" << endl;
	}
}

bool SCDB::FileExists( const string &profile, const string &resource_type, 
											 const string &resource_name )
	throw( db_action_failed )
{
	try {
		if ( this->KeyExists( "root*data*profile|name="+profile+"*"+resource_type+"|name="+resource_name ) )
			return true;
		else
			return false;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, 
											 "could not query file data for profile "+profile );
		return false;
			//throw db_action_failed( "db corrupt" );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, 
											 "could not query file data for profile "+profile );
		throw db_action_failed( "key not unique" );
	}
}

SCDBFile *SCDB::FileGetHandle( const string &profile, const string &resource_type, 
															 const string &resource_name, bool create )
	throw( db_action_failed, no_such_file )
{
	try {
		SCDBNode<backend_node> *node;
		if ( create && !this->KeyExists( 
					 "root*data*profile|name="+profile+"*"+resource_type+"|name="+resource_name )) {
			node = &this->AddNode( "root*data*profile|name="+profile, resource_type );
			SCDBNode<backend_node> &node_name = this->AddNode( *node, "name" );
			node_name.SetValue( resource_name );
			this->AddNode( *node, "subtype" );
			this->AddNode( *node, "md5" );
			this->AddNode( *node, "location" );
			this->AddNode( *node, "contents" );
			this->AddNode( *node, "start-prio" );
			this->AddNode( *node, "stop-prio" );
		}
		else {
			node = &this->GetNode( "root*data*profile|name="+profile+"*"+resource_type+
														 "|name="+resource_name );
				// maybe it's a bound resource
			string bind_res;
			bool found_prio=false;
			bind_res.clear();
			for (unsigned int i=0; i<node->GetNumSub(); i++) {
				if ((*node)[i].GetName() == "bind" ) {
					bind_res = (*node)[i].GetValue();
				}
				if ((*node)[i].GetName()=="start-prio") found_prio=true; 
			}
			if ( !bind_res.empty() ) {
					// it is one, so we use the node the resource is linked to
				node = &this->GetNode( "root*data*profile|name="+profile+"*"+resource_type+
															 "name="+bind_res );
			}
				// HACK: older resources may not have start/stop-prio nodes yet
				// we create them on the fly
			if (!found_prio) {
				this->AddNode( *node, "start-prio" );
				this->AddNode( *node, "stop-prio" );
			}
		}
		backend_file *file = 
			new backend_file( *node, resource_name, resource_type, profile, modified );
		return file;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, 
											 "could not query file data for profile "+profile+" (file not found)" );
		throw no_such_file( );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, 
											 "could not query file data for profile "+profile );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::FileCopy( const string &profile, const string &dest_profile, 
										 const string &resource_type, const string &resource_name ) 
	throw( db_action_failed, no_such_file, file_exists )
{
	if ( this->FileExists( dest_profile, resource_type, resource_name ) ) {
		LOGWARN( "file "+resource_name+" ("+resource_type+") already exist in profile "+dest_profile );
		throw file_exists();
	}
	try {
		SCDBNode<backend_node> &src = this->GetNode( "data*profile|name="+profile+"*"+resource_type+
																								 "|name="+resource_name );
		SCDBNode<backend_node> &dest = this->AddNode( "data*profile|name="+dest_profile, resource_type );
		LOGDEBUG( "copying "+resource_name+" ("+resource_type+") from "+profile+" to "+dest_profile );
		this->CopyNode( src, dest );
			// FIXME: this is an evil hack to fix broken links
			// copy file itself
		SCDBFile *sfile = this->FileGetHandle( profile, resource_type, resource_name, false );
		if ( sfile->GetSubtype() == "ghost" ) {
				// ghost file, nothing to copy
			delete sfile;
		}
		else {
			SCDBFile *dfile = this->FileGetHandle( dest_profile, resource_type, resource_name, true );
			istream *ins = sfile->GetStream();
			dfile->Save( *ins );
			dfile->SetSubtype( sfile->GetSubtype() );
			dfile->SetMD5sum( sfile->MD5sum() );
			dfile->SaveAttribs( sfile->GetAttribs() );
			if (dfile->GetSubtype() == "service" ) {
				dfile->SetStartPriority( sfile->GetStartPriority() );
				dfile->SetStopPriority( sfile->GetStopPriority() );
			}
			delete ins;
			delete sfile;
			delete dfile;
		}
			// maybe it has some subfiles
		backend_file *file = new backend_file( src, resource_name, resource_type, profile, modified );
		for ( unsigned int i=0; i<file->GetNumContents(); i++ ) {
			pair<string,string> &p = file->GetContent( i );
			this->FileCopy( profile, dest_profile, resource_type, p.second );
		}
		delete file;
	}
	catch( not_found ) {
		throw no_such_file();
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, 
											 "could not query file data for profile "+profile );
		throw db_action_failed( "key not unique" );
	}
}

void SCDB::FileDelete( const string &profile, const string &resource_type, 
											 const string &resource_name )
	throw( db_action_failed, no_such_file )
{
	string n = "root*data*profile|name="+profile+"*"+resource_type+"|name="+resource_name;
	try {
		SCDBNode<backend_node> &node = this->GetNode( n );
		backend_file *file = 
			new backend_file( node, resource_name, resource_type, profile, modified );
		unsigned int sub = file->GetNumContents();
		for( unsigned int i=0; i<sub; i++ ) {
			pair<string, string> p;
			p = file->GetContent( i );
			this->FileDelete( profile, resource_type, p.second );
		}
		file->Remove();
		delete file;
		node.SetDeleted( true );
		modified = true;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, 
											 "could not query file data for profile "+profile+" (file not found)" );
		throw no_such_file( );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, 
											 "could not query file data for profile "+profile );
		throw db_action_failed( "key not unique" );
	}
	catch( SCDBFile::write_error ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, 
											 "could not remove resource "+resource_name+" ("+resource_type+")" );
		throw db_action_failed( "I/O error" );
	}		 
}

string &SCDB::FileGetMD5( const string &profile, const string &resource_type, 
													const string &resource_name )
	throw( db_action_failed, no_such_file )
{
	static string md5;
	md5.clear();
	try {
		SCDBNode<backend_node> &node = this->GetNode( "root*data*profile|name="+profile+"*"+resource_type+"|name="+resource_name );
		backend_file file( node, resource_name, resource_type, profile, modified );
		md5 = file.MD5sum();
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, 
											 "could not query file data for profile "+profile+" (file not found)" );
		throw no_such_file( );
	}
	catch( not_unique ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, 
											 "could not query file data for profile "+profile );
		throw db_action_failed( "key not unique" );
	}
	return md5;
}

void SCDB::CopyNode( SCDBNode<backend_node> &src, SCDBNode<backend_node> &dest )
{
	for (unsigned int i=0; i<dest.GetNumSub(); i++) 
		dest[i].SetDeleted(true);
	for (unsigned int i=0; i<src.GetNumSub(); i++) {
		if (src[i].Deleted()) continue;
		SCDBNode<backend_node> &tmp = this->AddNode( dest, src[i].GetName() );
		if ( src[i].GetNumSub() )
			this->CopyNode( src[i], tmp );
		else
			tmp.SetValue( src[i].GetValue() );
	}
}

void SCDB::FileSave( const string &profile, const string &resource_type, 
										 const string &resource_name, istream &in, struct stat *st)
	throw( db_action_failed )
{
	try {
		SCDBFile *fileh = this->FileGetHandle( profile, resource_type, resource_name );
		fileh->Save( in );
		if (st) fileh->SaveAttribs( *st );
		delete fileh;
	}
	catch( SCDBFile::write_error ) {
		throw db_action_failed( "write error" );
	}
	catch( no_such_file ) {
		throw db_action_failed( "internal database error, could not make file handle" );
	}
}

void SCDB::FileRestore( const string &profile, const string &resource_type, 
												const string &resource_name, ostream &out, struct stat *st )
	throw( db_action_failed, no_such_file )
{
	try {
		SCDBFile *fileh = this->FileGetHandle( profile, resource_type, resource_name, false );
		fileh->Restore( out );
		if (st) st=&fileh->GetAttribs( );
		delete fileh;
	}
	catch( SCDBFile::read_error ) {
		throw db_action_failed( "read error" );
	}
	catch( SCDBFile::write_error ) {
		throw db_action_failed( "write error" );
	}
}

SCDBFile *SCDB::FilePushBackup( const string &profile, const string &resource_type,
																const string &resource_name, bool remove_file)
	throw( db_action_failed, no_such_file )
{
	try {
		SCDBNode<backend_node> &file_node = this->GetNode( "root*data*profile|name="+profile+
																											 "*"+resource_type+"|name="+resource_name );
		if (!this->KeyExists( "root*data*profile|name="+profile+"*backup" ))
			this->AddNode( "root*data*profile|name="+profile, "backup" );
			// if there are too many backups, delete the oldest one
		try {
			vector<SCDBNode<backend_node>*> nodes;
			this->SearchNodes( *root, "root*data*profile|name="+profile+"*backup*"+resource_type +"|name="+resource_name, nodes );
			if (nodes.size() > SCPM_conf::GetHandle()->GetMaxBackups()) {
				backend_file *fileh = new backend_file( *nodes[0], resource_name, resource_type, profile, modified, false );
				fileh->Remove();
				delete fileh;
				nodes[0]->SetDeleted(true);
			}
		}
		catch( not_found ) {
				// resource has no backups, that is fine
		}
		SCDBNode<backend_node> &backup_node = this->AddNode( "root*data*profile|name="+profile+"*backup",
																												 resource_type );
		// Copy Nodedata from Backup to File
		this->CopyNode( file_node, backup_node );
		
		// touch the new backupfile for setting mtime and ctime
		backend_file *backup_bf = new backend_file( backup_node, resource_name, resource_type, profile, modified, true);
		utime(backup_bf->GetLocation().c_str(), NULL);
		delete backup_bf;
		
		// return a new backend_file when it should be created
		backend_file *bf = NULL;
		if (remove_file)
			file_node.SetDeleted(true);
		else
			bf = new backend_file( file_node, resource_name, resource_type, profile, modified, true );
		return bf;
	}
	catch( not_found ) {
		throw no_such_file();
	}
	catch( not_unique ) {
		throw db_action_failed("key not unique");
	}
}

bool SCDB::FileRemoveBackup( const string &profile, const string &resource_type,
														 const string &resource_name, const string * version2remove)
	throw( db_action_failed, no_such_file )
{
	string errstr;
	try {
		// basically the same like FileDelete bot works ob backups 
		vector<SCDBNode<backend_node>*> backup_nodes;
		unsigned int backup_idx=0;
		backend_file *fileh = NULL;

		// find all backup nodes off this resource
		this->SearchNodes( *root, "root*data*profile|name="+profile+"*backup*"+resource_type+"|name="+resource_name, backup_nodes );
		
		// no backups found
		if (backup_nodes.size() == 0)
			throw db_action_failed("resource (" + resource_type + " " + resource_name  + ") has no backups");

		// remove specific version or latest?
		if (version2remove) {
			// find the right backupnode
			for(backup_idx=0; backup_idx<backup_nodes.size(); backup_idx++) {
				fileh = new backend_file(*(backup_nodes[backup_idx]), resource_name, resource_type, profile, modified, true);
				if ("{"+fileh->GetBackupVersion()+"}" == *version2remove) 
					break;
				delete fileh;
			}	
			// if the version was not found
			if (backup_idx >= backup_nodes.size())
				throw db_action_failed("resource (" + resource_type + " " + resource_name + ") has no backups");
		}
		else {
			backup_idx = backup_nodes.size() - 1;
			fileh = new backend_file( *(backup_nodes[backup_idx]), resource_name, resource_type, profile, modified, true);
		}

		// remove contents 
		unsigned int sub = fileh->GetNumContents();
		for( unsigned int i=0; i<sub; i++ ) {
			pair<string, string> p;
			p = fileh->GetContent( i );
			this->FileRemoveBackup( profile, resource_type, p.second, version2remove );
		}

		// remove backupfile
		fileh->Remove();
		delete fileh;
		
		// remove node from db
		backup_nodes[backup_idx]->SetDeleted(true);
		modified = true;

		return true;
	}
	catch( not_found ) {
		errstr="resource does not exist or has no backups";
	}
	catch( SCDBFile::write_error ) {
		errstr="write error";
	}
	catch ( not_unique ) {
		errstr="key not unique";
	}
	throw db_action_failed(errstr);
	
	return false;
}



SCDBFile *SCDB::FilePopBackup( const string &profile, const string &resource_type,
															 const string &resource_name, const string * version2restore)
	throw( db_action_failed, no_such_file )
{
	string errstr;
	try {
		vector<SCDBNode<backend_node>*> backup_nodes;
		unsigned int backup_idx = 0;

		// get all Backupnodes
		this->SearchNodes( *root, "root*data*profile|name="+profile+"*backup*"+resource_type+"|name="+resource_name, backup_nodes );
		
		// no backups found
		if (backup_nodes.size() == 0)
			throw db_action_failed("resource (" + resource_type + " " + resource_name  + ") has no backups");

		// restore specific version or the latest?
		if (version2restore) {
			// find the right backupnode
			for(backup_idx=0; backup_idx<backup_nodes.size(); backup_idx++) {
				backend_file bf(*(backup_nodes[backup_idx]), resource_name, resource_type, profile, modified, true);
				if ("{"+bf.GetBackupVersion()+"}" == *version2restore) 
					break;
			}
		}
		else
			backup_idx = backup_nodes.size() - 1;

		if (backup_idx >= backup_nodes.size())
			throw db_action_failed("resource (" + resource_type + " " + resource_name + ") has no backups");
	
		SCDBNode<backend_node> *file_node;	
		string key = "root*data*profile|name="+profile+"*"+resource_type+"|name="+resource_name;
																											 	
		if (this->KeyExists( key )) {
			// get the file_node
			file_node = &this->GetNode( key );
			
			// current version is obsolete
			SCDBFile *fileh = this->FileGetHandle( profile, resource_type, resource_name, false );
			fileh->Remove();
			delete fileh;
		}
		else { 
			// the file does not exist any more, but we want the backup restored, so lets create it
			file_node = &this->AddNode( "root*data*profile|name="+profile, resource_type);
		}
		
			// clear file node
		unsigned int numsub = file_node->GetNumSub();
		while(numsub) (*file_node)[--numsub].SetDeleted(true);

			// copy backup
		this->CopyNode( *(backup_nodes[backup_idx]), (*file_node) );

			// remove backup node
		backup_nodes[backup_idx]->SetDeleted(true);
		modified = true;

		// return the file
		backend_file *bf = new backend_file( *file_node, resource_name, resource_type, profile, modified, false );
		return bf;
	}
	catch( not_found ) {
		errstr="resource does not exist or has no backups";
	}
	catch( SCDBFile::write_error ) {
		errstr="write error";
	}
	catch ( not_unique ) {
		errstr="key not unique";
	}
	throw db_action_failed(errstr);
}

bool SCDB::FileHasBackup( const string &profile, const string &resource_type, const string &resource_name )
{
	return this->KeyExists("root*data*profile|name="+profile+"*backup*"+resource_type+"|name="
												 +resource_name );
}

void SCDB::FileGetBackupVersions( const string &profile, const string &resource_type,
																	const string &resource_name, vector<pair<time_t, string> > &versions)
	throw( db_action_failed, no_such_file)
{
	try {
		// insert backup-key if not present
		if (!this->KeyExists( "root*data*profile|name="+profile+"*backup"))
			this->AddNode( "root*data*profile|name="+profile, "backup");
		
		// search all nodes for this resource
		vector<SCDBNode<backend_node>*> nodes;
		this->SearchNodes( *root, "root*data*profile|name="+profile+"*backup*"+resource_type+"|name="+resource_name, nodes);
		
		// no backups found
		if (nodes.size() == 0)
			throw db_action_failed("resource (" + resource_type + " " + resource_name  + ") has no backups");

		// get files for all nodes
		for (unsigned int i=0; i < nodes.size(); i++){
			backend_file bf(*nodes[i], resource_name, resource_type, profile, modified, true);
			// we add surrounding brackets to the Version-String because it can be empty
			// and we dont want to run in any problems
			if (bf.GetSubtype() != "ghost")
			{
				pair<time_t, string> version(bf.GetAttribs().st_mtime, "{"+bf.GetBackupVersion()+"}" );
				versions.push_back(version);
			}
			else
			{
				pair<time_t, string> version(0, "{"+bf.GetBackupVersion()+"}");
				versions.push_back(version);
			}

			// older resources may not have Start-Stop-Prio,
			// create them on the fly	
			bool found_prio = false;
			for (unsigned int j=0; j < nodes[i]->GetNumSub(); j++) 
				if ((*nodes[i])[j].GetName()=="start-prio") { found_prio=true; break;}
			if (!found_prio) {
				this->AddNode(*nodes[i], "start-prio");
				this->AddNode(*nodes[i], "stop-prio");
			}
		}
		
	}
	catch( not_found ) {
		Log* log = Log::GetHandle();
		log->WriteMessage("scdb", Log::CLASS_DEBUG, "could not query file data for profile "+profile+" (file not found)" );
		throw no_such_file();
	}
}

SCDBFile *SCDB::FileGetBackupHandle( const string &profile, const string &resource_type, 
																		 const string &resource_name, bool create, string * version )
	throw( db_action_failed, no_such_file )
{
	try {
		if (!this->KeyExists( "root*data*profile|name="+profile+"*backup" ))
			this->AddNode( "root*data*profile|name="+profile, "backup" );
		SCDBNode<backend_node> *node;
		if (create) {
			node = &this->AddNode( "root*data*profile|name="+profile+"*backup", resource_type );
			SCDBNode<backend_node> &node_name = this->AddNode( *node, "name" );
			node_name.SetValue( resource_name );
			this->AddNode( *node, "subtype" );
			this->AddNode( *node, "md5" );
			this->AddNode( *node, "location" );
			this->AddNode( *node, "contents" );
			this->AddNode( *node, "start-prio" );
			this->AddNode( *node, "stop-prio" );
		}
		else {
			// should use FileGetBackupHandles or at least share code with it
			vector<SCDBNode<backend_node>*> nodes;
			this->SearchNodes( *root,
												 "root*data*profile|name="+profile+"*backup*"+resource_type+"|name="+resource_name, nodes );
			if (nodes.size() == 0)
				throw no_such_file();

			if (version) {
				// find the right backupnode
				unsigned int backup_idx;
				backend_file *fileh;
				for(backup_idx=0; backup_idx<nodes.size(); backup_idx++) {
					fileh = new backend_file(*(nodes[backup_idx]), resource_name, resource_type, profile, modified, true);
					if ("{"+fileh->GetBackupVersion()+"}" == *version) 
						break;
					delete fileh;
				}	
				// if the version was not found
				if (backup_idx >= nodes.size())
					throw db_action_failed("resource (" + resource_type + " " + resource_name + ") has no backups");
				node = nodes[backup_idx];
			}
			else
				node = nodes[nodes.size()-1];

			bool found_prio;
			for (unsigned int i=0; i<node->GetNumSub(); i++) {
				if ((*node)[i].GetName()=="start-prio") found_prio=true; 
			}
				// HACK: older resources may not have start/stop-prio nodes yet
				// we create them on the fly
			if (!found_prio) {
				this->AddNode( *node, "start-prio" );
				this->AddNode( *node, "stop-prio" );
			}
		}
		backend_file *file = 
			new backend_file( *node, resource_name, resource_type, profile, modified, true );
		return file;
	}
	catch( not_found ) {
		Log *log = Log::GetHandle();
		log->WriteMessage( "scdb", Log::CLASS_DEBUG, 
											 "could not query file data for profile "+profile+" (file not found)" );
		throw no_such_file( );
	}
}

void SCDB::CreateSubTree( string rootspec, istream &inputstream ) throw( db_action_failed )
{
	try {
		SCDBNode<backend_node> &node = this->GetNode( rootspec );
		backend->CreateSubTree( static_cast<backend_node&>(node), inputstream, false );
		modified=true;
	}
	catch( XML::op_failed ) {
		throw db_action_failed( "database error" );
	}
	catch( not_unique ) {
		throw db_action_failed( "key not unique" );
	}
	catch( not_found ) {
		throw db_action_failed( "key not found" );
	}
}
