/*
 * $Id: scdb_file.hh,v 1.3 2008/10/22 10:12:42 hschaa Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	SCPM database
 * File				 :	scdb_file.hh
 * Description :	SCDB storage interface
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2002-2004 SuSE Linux AG
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#ifndef SCDB_FILE_H
#define SCDB_FILE_H
#include <iostream>
#include <string>
#include <cstdlib>
#include <vector>
#include <list>
#include <utility>
#include "scdb_node.hh"
#include "xml.hh"

static inline std::string itoa( int i )
{
	std::ostringstream o;
	o << i;
	return o.str();
}

class SCDBFile
{
public:
	friend class SCDB;
	typedef XMLNode backend_node;

	struct read_error{ };
	struct write_error{ };
	struct index_invalid{ };

	std::string &MD5sum( ) { return md5sum; };
	virtual std::string &RealMD5sum( ) =0;
	virtual void SetMD5sum( const std::string &md5 ) { md5sum=md5; scdb_modified=true; };
	virtual void Save( std::istream &ins, bool auto_md5=false, std::string suffix="" ) throw( write_error ) =0;
	virtual void Restore( std::ostream &outs ) throw( read_error, write_error ) =0;
	virtual void Remove( ) throw( write_error ) =0;
	virtual void Create( std::string suffix ) throw( write_error ) =0;
	virtual void SaveAttribs( struct stat &attribs ) throw( write_error ) =0;
	virtual struct stat &GetAttribs( ) throw( read_error ) =0;
	virtual std::istream *GetStream( ) throw( read_error ) =0;
	virtual std::string GetBackupVersion() = 0;
	std::string &GetName( ) { return name; };
	std::string &GetType( ) { return type; };
	void SetSubtype( const std::string &stype ) { subtype=stype; scdb_modified=true; };
	std::string &GetSubtype( ) { return subtype; };
	unsigned int GetNumContents( ) { return contents.size(); };
	std::pair<std::string, std::string> &GetContent( unsigned int idx ) throw( index_invalid ) {
		if (idx>=contents.size()) throw index_invalid();
		return contents[idx]; };
	std::pair<std::string, std::string> &operator[]( unsigned int idx ) throw( index_invalid )
		{ return this->GetContent(idx); };
	void ClearContent( ) { contents.clear(); scdb_modified=true; };
	void AddContent( std::pair<std::string,std::string> &data ) {
		contents.push_back( data );
		scdb_modified=true;
	};
	int GetStartPriority( ) { return start_prio; };
	int GetStopPriority( ) { return stop_prio; };
	void SetStartPriority( int prio ) { start_prio=prio; scdb_modified=true; };
	void SetStopPriority( int prio ) { stop_prio=prio; scdb_modified=true; };
	std::string GetLocation( ) { return read_location; };
	virtual ~SCDBFile( ) {
		for ( unsigned int i=0; i<scdb_node.GetNumSub(); i++ ) {
			if (scdb_node[i].GetName() == "subtype") scdb_node[i].SetValue( subtype );
			if (scdb_node[i].GetName() == "location") scdb_node[i].SetValue( read_location );
			if (scdb_node[i].GetName() == "md5") scdb_node[i].SetValue( md5sum );
			if (scdb_node[i].GetName() == "contents") {
				unsigned int numsub=scdb_node[i].GetNumSub();
				while( numsub-- ) scdb_node[i].DelSub( numsub );
				for ( unsigned int j=0; j<contents.size(); j++ ) {
					backend_node *n = new backend_node( contents[j].first );
					n->SetValue( contents[j].second );
					scdb_node[i].AddSub( *n );
				}
			}
			if (scdb_node[i].GetName() == "start-prio")
				if (start_prio != -1) scdb_node[i].SetValue( itoa(start_prio) );
			if (scdb_node[i].GetName() == "stop-prio")
				if (stop_prio != -1) scdb_node[i].SetValue( itoa(stop_prio) );
		}
	};
	
private:
	SCDBNode<backend_node> &scdb_node;
	std::list<SCDBNode<backend_node>*> subnodes;
	std::vector<std::pair<std::string, std::string> > contents;

protected:
	std::string name;
	std::string type;
	std::string subtype;
	std::string profile;
	bool &scdb_modified;
	std::string read_location;
	std::string write_location;
	std::string md5sum;
	bool backup;
	int start_prio;
	int stop_prio;
	std::vector<SCDBNode<backend_node> *> backup_nodes;
	SCDBFile( SCDBNode<backend_node> &node, const std::string &resname, const std::string &restype,
						const std::string &profile_name, bool &modified, bool is_backup=false )
			: scdb_node( node ), name( resname ), type( restype ), profile( profile_name ),
				scdb_modified( modified ), backup( is_backup ), start_prio(-1), stop_prio(-1) {
		for ( unsigned int i=0; i<scdb_node.GetNumSub(); i++ ) {
			if (!(scdb_node[i].Deleted())) {
				if (scdb_node[i].GetName() == "subtype") subtype = scdb_node[i].GetValue();
				if (scdb_node[i].GetName() == "location") read_location = scdb_node[i].GetValue();
				if (scdb_node[i].GetName() == "md5") md5sum = scdb_node[i].GetValue();
				if (scdb_node[i].GetName() == "contents")
					for ( unsigned int j=0; j<scdb_node[i].GetNumSub(); j++ )
						if (!(scdb_node[i][j].Deleted()))
							contents.push_back( make_pair(scdb_node[i][j].GetName(), scdb_node[i][j].GetValue()) );
				if (scdb_node[i].GetName() == "start-prio" && scdb_node[i].GetValue().size()) 
					start_prio = (int)strtol((scdb_node[i].GetValue()).c_str(),NULL,10);
				if (scdb_node[i].GetName() == "stop-prio" && scdb_node[i].GetValue().size()) 
					stop_prio = (int)strtol((scdb_node[i].GetValue()).c_str(),NULL,10);
			}
		}
	};
};

#endif
