/*
 * $Id: scdb_util.cc,v 1.1.1.1 2005/05/11 09:03:12 gleissner Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	SCDB manipulation utility
 * File				 :	scdb_util.cc
 * Description :	Class containing utility functions for SCDB
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2002-2004 SuSE Linux AG
 *								2004 SUSE LINUX Products GmbH
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "scdb_util.hh"
#include <resource_helpers.hh>
#include <resource.hh>
#include <scpm_conf.hh>
#include <fstream>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>
#include <utime.h>

using namespace std;

const string module_name="scdb_util";

void SCDBUtil::CreateResource( string resource_name, string resource_type,
															 string profile, bool skip_update )
	throw( operation_failed )
{
	try {
			// do check for profile existance first
		if ( profile != "all" && profile != "none" &&
				 !db->ProfileExists( profile ) )
			throw( operation_failed( "no such profile" ) );

			// create resource in database
		db->ResourceAdd( resource_name, resource_type );

		Resource_helpers rh;
		Resource *r = rh.CreateObject( resource_type, resource_name );

			// add deps
		vector<string> types = rh.GetResourceTypes();
		for ( unsigned int i=0; i<types.size(); i++ ) {
			try {
				vector<string> deps = r->GetDependencies( types[i] );
				for ( unsigned int j=0; j<deps.size(); j++ ) {
					db->ResourceAddDep( resource_name, resource_type, deps[j], types[i] );
				}
			}
			catch( Resource::not_available ) {
					// not all resource types have dependencies
			}
		}

			// let the resource update its data
		try {
			if ( !skip_update ) {
				r->UpdateData();
				r->WriteData();
			}
		}
		catch( Resource::not_available ) {
				// resource has no UpdateData()
		}

		LOGDEBUG( "resource "+resource_name+" ("+resource_type+") created" );

			// add resource to profiles
		vector<string> profiles = db->GetProfiles();
		if ( profile == "all" )
			for ( unsigned int i=0; i<profiles.size(); i++ ) {
				db->ProfileAddResource( profiles[i], resource_name, resource_type );
				r->SetProfile( profiles[i] );
				r->Save();
				LOGDEBUG( "resource "+resource_name+" ("+resource_type+") added to profile "
									+profiles[i] );
			}
		else if ( profile != "none" ) {
			db->ProfileAddResource( profile, resource_name, resource_type );
			r->SetProfile( profile );
			r->Save();
			LOGDEBUG( "resource "+resource_name+" ("+resource_type+") added to profile "+profile );
		}
		delete r;
	}
	catch( SCDB::db_action_failed &e ) {
		throw operation_failed( e.reason );
	}
	catch( Resource_helpers::type_invalid ) {
		throw operation_failed( "invalid resource type" );
	}
	catch( Resource::resource_unknown ) {
		throw operation_failed( "unknown resource addressed" );
	}
	catch( ... ) {
		throw operation_failed( scpm_error );
	}
}

void SCDBUtil::DropResource( string resource_name, string resource_type, bool full )
	throw( operation_failed )
{
	try {
		vector<string> profiles = db->GetProfiles();
		for ( unsigned int i=0; i<profiles.size(); i++ ) {
			db->ProfileDeleteResource( profiles[i], resource_name, resource_type );
			db->FileDelete( profiles[i], resource_type, resource_name );
			LOGDEBUG( "resource "+resource_name+" ("+resource_type+") removed from profile "
								+profiles[i] );
		}
		if ( full ) db->ResourceDelete( resource_name, resource_type );
	}
	catch( SCDB::db_action_failed &e ) {
		throw operation_failed( e.reason );
	}
	catch( ... ) {
		throw operation_failed( scpm_error );
	}
}

void SCDBUtil::BindResource( string resource_name, string resource_type,
														 string src_profile, string dest_profile )
	throw( operation_failed )
{
	throw operation_failed( "resource binding not yet implemented" );
}


void SCDBUtil::ReleaseResource( string resource_name, string resource_type,
																string profile )
	throw( operation_failed )
{
	throw operation_failed( "resource binding not yet implemented" );
}

void SCDBUtil::FindUnusedResources( vector<string> &resource_names, vector<string> &resource_types )
	throw( operation_failed )
{
	Resource_helpers rh;
	
	try {
		vector<string> types = db->ResourceGetTypes();
		for ( unsigned int i=0; i<types.size(); i++ ) {
			vector<string> names = db->ResourceGetNames( types[i] );
			for ( unsigned int j=0; j<names.size(); j++ ) {
				bool in_use = false;
				vector<string> profiles = db->GetProfiles();
				for ( unsigned int k=0; k<profiles.size(); k++ ) {
					Resource *r = rh.CreateObject( types[i], names[j] );
					r->SetProfile( profiles[k] );
					if ( r->Modified() ) {
						in_use = true;
							//delete r;
							//break;
					}
					else {
							// resource is unmodified, but maybe a dependend resource
							// has been modified, so this resource may be in use though
						vector<SCDB::resource_dep_t> deps = db->ResourceGetDeps( names[j], types[i] );
						for ( unsigned int l=0; l<deps.size(); l++ ) {
							try {
								Resource *rd = rh.CreateObject( deps[l].type, deps[l].name );
								rd->SetProfile( profiles[k] );
								if ( rd->Modified() ) {
									in_use = true;
									delete rd;
									break;
								}
								delete rd;
							}
							catch( Resource::resource_unknown ) {
									// a dependend resource does not exist
									// this is harmless, we ignore it
							}
						}
					}
					delete r;
				}
				if ( !in_use ) {
					resource_names.push_back( names[j] );
					resource_types.push_back( types[i] );
				}
			}
		}
	}
	catch( SCDB::db_action_failed &e ) {
		throw operation_failed( e.reason);
	}
	catch( Resource_helpers::type_invalid ) {
		throw operation_failed( "internal error, a db resource type is not known" );
	}
	catch( Resource::resource_unknown ) {
		throw operation_failed( "internal error, a resource vanished (maybe db corruption)" );
	}
	catch( ... ) {
		throw operation_failed( scpm_error );
	}
}

void SCDBUtil::FindUnhandledResources( vector<string> &resource_names,
																			 vector<string> &resource_types ) throw( operation_failed )
{
	Resource_helpers rh;
	vector<string> types;
	try {
		types = rh.GetResourceTypes();
		for ( unsigned int i=0; i<types.size(); i++ ) {
			vector<string> names = rh.GetResourceNames( types[i] );
			for ( unsigned int j=0; j<names.size(); j++ ) {
				if ( db->ResourceExists( names[j], types[i] ) )
					continue;
				else {
					resource_types.push_back( types[i] );
					resource_names.push_back( names[j] );
				}
			}
		}
	}
	catch( Resource_helpers::type_invalid ) {
		throw operation_failed( "invalid type addressed, this is a bug" );
	}
	catch( Resource::resource_unknown ) {
		throw operation_failed( "unknown resource type addressed, this is a bug" );
	}
	catch( ... ) {
		throw operation_failed( scpm_error );
	}
}

void SCDBUtil::FindDeletedResources( vector<string> &resource_names,
																		 vector<string> &resource_types )
	throw( operation_failed )
{
	Resource_helpers rh;
	try {
		vector<string> types = db->ResourceGetTypes();
		for ( unsigned int i=0; i<types.size(); i++ ) {
			vector<string> names = db->ResourceGetNames( types[i] );
			for ( unsigned int j=0; j<names.size(); j++ ) {
				bool found = false;
				vector<string> avail = rh.GetResourceNames( types[i] );
				for ( unsigned int k=0; k<avail.size(); k++ )
					if ( avail[k] == names[j] ) { found = true; break; }
				if ( !found ) {
					resource_names.push_back( names[j] );
					resource_types.push_back( types[i] );
				}
			}
		}
	}
	catch( SCDB::db_action_failed &e ) {
		throw operation_failed( e.reason);
	}
	catch( ... ) {
		throw operation_failed( scpm_error );
	}
}

void SCDBUtil::MinimizeDB( bool full ) throw( operation_failed )
{
	vector<string> names;
	vector<string> types;
	this->FindUnusedResources( names, types );
	for ( unsigned int i=0; i<names.size(); i++ )
		this->DropResource( names[i], types[i], full );
}

void SCDBUtil::MaximizeDB( bool add_to_profiles ) throw( operation_failed )
{
	Resource_helpers rh;
	vector<string> types;
	try {
		types = rh.GetResourceTypes();
		for ( unsigned int i=0; i<types.size(); i++ ) {
			vector<string> names = rh.GetResourceNames( types[i] );
			for ( unsigned int j=0; j<names.size(); j++ ) {
				if ( db->ResourceExists( names[j], types[i] ) )
					continue;
				else {
					if ( add_to_profiles )
						this->CreateResource( names[j], types[i], "all" );
					else
						this->CreateResource( names[j], types[i], "none" );
				}
			}
		}
	}
	catch( Resource_helpers::type_invalid ) {
		throw operation_failed( "invalid type addressed, this is a bug" );
	}
	catch( Resource::resource_unknown ) {
		throw operation_failed( "unknown resource type addressed, this is a bug" );
	}
	catch( SCDB::db_action_failed &e ) {
		throw operation_failed( e.reason );
	}
	catch( operation_failed &e ) {
		throw e;
	}
	catch( ... ) {
		throw operation_failed( scpm_error );
	}
}

void SCDBUtil::RebuildDeps( ) throw( operation_failed )
{
	Resource_helpers rh;
	try {
		vector<string> types = db->ResourceGetTypes();
		for ( unsigned int i=0; i<types.size(); i++ ) {
			vector<string> names = db->ResourceGetNames( types[i] );
			for ( unsigned int j=0; j<names.size(); j++ ) {
				Resource *r = rh.CreateObject( types[i], names[j] );
				db->ResourceCleanDeps( names[j], types[i] );
				try {
					for ( unsigned int k=0; k<types.size(); k++ ) {
						vector<string> deps = r->GetDependencies( types[k] );
						for ( unsigned int l=0; l<deps.size(); l++ ) {
							db->ResourceAddDep( names[j], types[i], deps[l], types[k] );
						}
					}
				}
				catch( Resource::not_available ) {
						// resource type has no dependencies, skipping this type
				}
				delete r;
			}
		}
	}
	catch( Resource_helpers::type_invalid ) {
		throw operation_failed( "invalid resource type addressed" );
	}
	catch( SCDB::db_action_failed &e ) {
		throw operation_failed( e.reason );
	}
	catch( ... ) {
		throw operation_failed( scpm_error );
	}
}

void SCDBUtil::RebuildResources( ) throw( operation_failed )
{
	try {
		Resource_helpers rh;
		vector<string> profiles = db->GetProfiles();
		vector<string> types = rh.GetResourceTypes();
		for (unsigned int i=0; i<profiles.size(); i++) {
			db->ProfileClearResources( profiles[i] );
			for (unsigned int j=0; j<types.size(); j++) {
				vector<string> names = rh.GetResourceNames( types[j] );
				for (unsigned int k=0; j<names.size(); k++) {
					db->ProfileAddResource( profiles[i], names[k], types[j] );
				}
			}
		}
	}
	catch( SCDB::db_action_failed &e ) {
		throw operation_failed( e.reason );
	}
}

void SCDBUtil::FindUnreferencedFiles( vector<string> &filenames ) throw( operation_failed )
{
	SCPM_conf *conf = SCPM_conf::GetHandle();
	DIR *dp;
	struct dirent *ep;
	vector<string> profiles = db->GetProfiles();
	for (unsigned int i=0; i<profiles.size(); i++) {
		string path = conf->GetProfileDir()+"/"+profiles[i]+"/file";
		dp = opendir( path.c_str() );
		if (!dp) {
			LOGDEBUG( path+": "+strerror(errno) );
			throw operation_failed( "could not open data directory for profile "+profiles[i] );
		}
		while( (ep = readdir( dp ) ) ) {
			if (ep->d_name[0]=='.') continue;
			if (db->KeyExists( "root*data*profile|name="+profiles[i]+"*file|location="+path+"/"+
												 string(ep->d_name) )) continue;
			if (db->KeyExists( "root*data*profile|name="+profiles[i]+"*backup*file|location="+path+"/"+
												 string(ep->d_name) )) continue;
			LOGDEBUG( "file "+path+"/"+string(ep->d_name)+" is not referenced in database" );
			filenames.push_back( path+"/"+string(ep->d_name) );
		}
		closedir( dp );
	}
}

void SCDBUtil::ExportProfile( const string &root, const string &profile ) throw( operation_failed )
{
	struct stat st;
	if ( stat( root.c_str(), &st ) == -1 )
		throw operation_failed( "could not stat "+root );
	if ( !S_ISDIR(st.st_mode) )
		throw operation_failed( root+" is not a directory" );

	Resource *r;
	Resource_helpers rh;
	try {
		if (!db->ProfileExists(profile)) throw operation_failed( "no such profile "+profile );
		vector<string> resource_types = db->ProfileGetResourceTypes( profile );
		for (unsigned int i=0; i<resource_types.size(); i++) {
			if ( stat( (root+"/"+resource_types[i]).c_str(), &st ) == -1 )
				if (mkdir((root+"/"+resource_types[i]).c_str(),0755) == -1) {
					LOGDEBUG( root+"/"+resource_types[i]+": "+strerror(errno) );
					throw operation_failed( "could not create directory for resource type "+resource_types[i] );
				}
			vector<string> resource_names = db->ProfileGetResources( profile, resource_types[i] );
			for (unsigned int j=0; j<resource_names.size(); j++) {
				r=rh.CreateObject( resource_types[i], resource_names[j], false );
				r->SetProfile( profile );
				if (!r->Export( root+"/"+resource_types[i] ))
					throw operation_failed( "export of resource "+resource_names[j]+" ("+
																	resource_types[i]+") failed" );
				delete r;
			}
		}
	}
	catch( Resource::not_available ) {
		LOGWARN( "at least one resource type does not support export" );
		throw operation_failed( "export profile support incomplete" );
	}
}

void SCDBUtil::ImportProfile( const string &root, const string &profile ) throw( operation_failed )
{
	struct stat st;
	if ( stat( root.c_str(), &st ) == -1 )
		throw operation_failed( "could not stat "+root );
	if ( !S_ISDIR(st.st_mode) )
		throw operation_failed( root+" is not a directory" );

	Resource *r;
	Resource_helpers rh;

	try {
		if (db->ProfileExists(profile)) throw operation_failed( "profile "+profile+" exists" );
		db->ProfileAdd( profile );
		vector<string> resource_types = db->ResourceGetTypes();
		for (unsigned int i=0; i<resource_types.size(); i++) {
			if ( stat( (root+"/"+resource_types[i]).c_str(), &st ) == -1 ) {
				LOGDEBUG( root+"/"+resource_types[i]+": "+strerror(errno) );
				throw operation_failed( "import directory incomplete" );
			}
			vector<string> resource_names = db->ResourceGetNames( resource_types[i] );
			for (unsigned int j=0; j<resource_names.size(); j++) {
				db->ProfileAddResource( profile, resource_names[j], resource_types[i] );
				r=rh.CreateObject( resource_types[i], resource_names[j], false );
				r->SetProfile( profile );
				if (!r->Import( root+"/"+resource_types[i] )) {
					LOGWARN( "import of resource "+resource_names[j]+" ("+resource_types[i]+") failed" );
				}
				delete r;
			}
		}
	}
	catch( Resource::not_available ) {
		LOGWARN( "at least one resource type does not support export" );
		throw operation_failed( "could not export profile" );
	}
}


/*
	void SCDBUtil::ExportProfile( const string &root, const string &profile ) throw( operation_failed )
	{
	struct stat st;
	if ( stat( root.c_str(), &st ) == -1 )
	throw operation_failed( "could not stat "+root );
	if ( !S_ISDIR(st.st_mode) )
	throw operation_failed( root+" is not a directory" );
	SCDBFile *fileh = NULL;
	SCDBFile *sfileh = NULL;
	string errstr;
	try {
	if (!db->ProfileExists(profile)) throw operation_failed( "no such profile "+profile );
	vector<string> resource_types = db->ProfileGetResourceTypes( profile );
	for (unsigned int i=0; i<resource_types.size(); i++) {
	if ( stat( (root+"/"+resource_types[i]).c_str(), &st ) == -1 )
	if (mkdir((root+"/"+resource_types[i]).c_str(),0755) == -1) {
	LOGDEBUG( root+"/"+resource_types[i]+": "+strerror(errno) );
	throw operation_failed( "could not create directory for resource type "+resource_types[i] );
	}
	vector<string> resource_names = db->ProfileGetResources( profile, resource_types[i] );
	for (unsigned int j=0; j<resource_names.size(); j++) {
	fileh = db->FileGetHandle( profile, resource_types[i], resource_names[j] );
	string curloc = fileh->GetLocation();
	string newloc = root+"/"+resource_types[i]+"/"+string(basename((const char*)curloc.c_str()));
	ofstream outf( newloc.c_str(), ios::trunc );
		// HACK
		if (fileh->GetSubtype() != "ghost" ) {
		fileh->Restore( outf );
		try {
		struct stat &file_stat = fileh->GetAttribs();
		struct utimbuf ub;
		ub.actime = file_stat.st_atime;
		ub.modtime = file_stat.st_mtime;
		utime( newloc.c_str(), &ub );
		chmod( newloc.c_str(), file_stat.st_mode );
		chown( newloc.c_str(), file_stat.st_uid, file_stat.st_gid );
		}
		catch( SCDBFile::read_error ) {
		LOGWARN( "could not set attribs for file "+newloc );
		}
		}
		outf.close();
		for (unsigned int k=0; k<fileh->GetNumContents(); k++) {
		pair<string, string> p = fileh->GetContent(k);
		sfileh = db->FileGetHandle( profile, resource_types[i], p.second, false );
		string curloc = sfileh->GetLocation();
		string newloc = root+"/"+resource_types[i]+"/"+string(basename((const char*)curloc.c_str()));
		ofstream outf( newloc.c_str(), ios::trunc );
			// HACK again
			if (sfileh->GetSubtype() != "ghost" ) {
			sfileh->Restore( outf );
			try {
			struct stat &file_stat = sfileh->GetAttribs();
			struct utimbuf ub;
			ub.actime = file_stat.st_atime;
			ub.modtime = file_stat.st_mtime;
			utime( newloc.c_str(), &ub );
			chmod( newloc.c_str(), file_stat.st_mode );
			chown( newloc.c_str(), file_stat.st_uid, file_stat.st_gid );
			}
			catch( SCDBFile::read_error ) {
			LOGWARN( "could not set attribs for file "+newloc );
			}
			}
			outf.close();
			}
			delete fileh;
			}
			}
			ofstream outf( (root+"/profile.scdb").c_str() );
			db->Dump( outf, "root*profiles*profile|name="+profile );
			outf.close();
			ofstream outf2( (root+"/data.scdb").c_str() );
			db->Dump( outf2, "root*data*profile|name="+profile );
			outf2.close();
			}
			catch( SCDB::db_action_failed &e ) {
			if (fileh) delete fileh;
			throw operation_failed( "SCDB error: "+e.reason );
			}
			catch( SCDB::no_such_file ) {
			throw operation_failed( "File not found in SCDB" );
			}
			catch( SCDBFile::read_error ) {
			throw operation_failed( "SCDB read error" );
			}
			catch( SCDBFile::write_error ) {
			throw operation_failed( "SCDB write error" );
			}
			}

			void SCDBUtil::ImportProfile( const string &root, const string &profile ) throw( operation_failed )
			{
			SCDBFile *dest_fh = NULL;
			SCDBFile *src_fh = NULL;
			SCDBFile *dest2_fh = NULL;
			SCDBFile *src2_fh = NULL;
			try {
			if (db->ProfileExists( profile ))
			throw operation_failed( "profile exists" );
			ifstream sprofile( (root+"/profile.scdb").c_str() );
			ifstream sdata( (root+"/data.scdb").c_str() );
			if (sprofile.bad()) throw operation_failed( "error accessing profile.scdb" );
			if (sdata.bad()) throw operation_failed( "error accessing data.scdb" );
			db->AddNode( "root", "temp_profiles" );
			db->AddNode( "root", "temp_data" );
			db->CreateSubTree( "root*temp_profiles", sprofile );
			sprofile.close();
			db->CreateSubTree( "root*temp_data", sdata );
			sdata.close();
		
			db->ProfileAdd( profile );

				// copy some profile information
				{
				SCDBNode<SCDB::backend_node> &src = db->GetNode( "root*temp_profiles*profile*description" );
				SCDBNode<SCDB::backend_node> &dest = db->GetNode( "root*profiles*profile|name="+profile+"*description" );
				db->CopyNode( src, dest );
				}
				{
				SCDBNode<SCDB::backend_node> &src = db->GetNode( "root*temp_profiles*profile*scripts" );
				SCDBNode<SCDB::backend_node> &dest = db->GetNode( "root*profiles*profile|name="+profile+"*scripts" );
				db->CopyNode( src, dest );
				}

					// add resources
					vector<string> resource_types = db->ResourceGetTypes();
					for (unsigned int i=0; i<resource_types.size();i++) {
					vector<string> resource_names = db->ResourceGetNames( resource_types[i] );
					for (unsigned int j=0; j<resource_names.size(); j++) {
					db->ProfileAddResource( profile, resource_names[j], resource_types[i] );
					dest_fh = db->FileGetHandle( profile, resource_types[i], resource_names[j], true );
					SCDBNode<SCDB::backend_node> &n = db->GetNode( "root*temp_data*profile*"+resource_types[i]+
					"|name="+resource_names[j] );
						// this is a bit hackish here
						bool fake_modified;
						src_fh = new SCDB::backend_file( n, resource_names[j], resource_types[i], "temp", fake_modified );
						if (src_fh->GetSubtype()!="ghost") {
						string loc = root+"/"+resource_types[i]+"/"+string(basename(src_fh->GetLocation().c_str()));
						ifstream ins( loc.c_str() );
						if (ins.fail()) {
						delete src_fh;
						delete dest_fh;
						throw operation_failed( "could not open "+loc );
						}
						dest_fh->Save( ins );
						dest_fh->SetMD5sum( src_fh->MD5sum() );
						dest_fh->SaveAttribs( src_fh->GetAttribs() );
						}
						dest_fh->SetSubtype( src_fh->GetSubtype() );
						dest_fh->SetStartPriority( src_fh->GetStartPriority() );
						dest_fh->SetStopPriority( src_fh->GetStopPriority() );
						for (unsigned int k=0; k<src_fh->GetNumContents(); k++) {
						pair<string,string> &p = src_fh->GetContent(k);
						SCDBNode<SCDB::backend_node> &n = db->GetNode( "root*temp_data*profile*"+resource_types[i]+
						"|name="+p.second );
						src2_fh = new SCDB::backend_file( n, p.second, resource_types[i], "temp", fake_modified );
						dest2_fh = db->FileGetHandle( profile, resource_types[i], p.second, true );
						if (src2_fh->GetSubtype()!="ghost") {
						string loc = root+"/"+resource_types[i]+"/"+string(basename(src2_fh->GetLocation().c_str()));
						ifstream ins( loc.c_str() );
						if (ins.fail()) {
						delete src_fh;
						delete dest_fh;
						delete src2_fh;
						delete dest2_fh;
						throw operation_failed( "could not open "+loc );
						}
						dest2_fh->Save( ins );
						dest2_fh->SetMD5sum( src2_fh->MD5sum() );
						dest2_fh->SaveAttribs( src2_fh->GetAttribs() );
						}
						dest2_fh->SetSubtype( src2_fh->GetSubtype() );
						dest2_fh->SetStartPriority( src2_fh->GetStartPriority() );
						dest2_fh->SetStopPriority( src2_fh->GetStopPriority() );
						dest_fh->AddContent( p );
						}
						delete src_fh;
						delete dest_fh;
						delete src2_fh;
						delete dest2_fh;
						src_fh=NULL;
						dest_fh=NULL;
						src2_fh=NULL;
						dest2_fh=NULL;
						}
						}
						db->GetNode( "root*temp_data" ).SetDeleted(true);
						db->GetNode( "root*temp_profiles" ).SetDeleted(true);
						LOGINFO( "imported profile "+profile );
						}
						catch( SCDB::db_action_failed &e ) {
						if (src_fh) delete src_fh;
						if (dest_fh) delete dest_fh;
						if (src2_fh) delete src_fh;
						if (dest2_fh) delete dest_fh;
						throw operation_failed( "SCDB error: "+e.reason );
						}
						catch( SCDBFile::read_error ) {
						if (src_fh) delete src_fh;
						if (dest_fh) delete dest_fh;
						if (src2_fh) delete src_fh;
						if (dest2_fh) delete dest_fh;
						throw operation_failed( "SCDB read error" );
						}
						catch( SCDBFile::write_error ) {
						if (src_fh) delete src_fh;
						if (dest_fh) delete dest_fh;
						if (src2_fh) delete src_fh;
						if (dest2_fh) delete dest_fh;
						throw operation_failed( "SCDB write error" );
						}

/*		vector<string> resource_types = db->ProfileGetResourceTypes( profile );
for (unsigned int i=0; i<resource_types.size();i++) {
vector<string> resource_names = db->ProfileGetResources( profile, resource_types[i] );
for (unsigned int j=0; j<resource_names.size(); j++) {
cerr << "Importing " << resource_names[j] << " (" << resource_types[i] << ")\n" ;
fileh = db->FileGetHandle( profile, resource_types[i], resource_names[j], false );
if (fileh->GetSubtype()!="ghost") {
string loc = root+"/"+resource_types[i]+"/"+string(basename(fileh->GetLocation().c_str()));
ifstream ins( loc.c_str() );
if (ins.fail()) {
delete fileh;
throw operation_failed( "could not open "+loc );
}
fileh->Save( ins );
struct stat file_stat;
if ( stat( loc.c_str(), &file_stat ) != 0 ) {
LOGWARN( "can not save attributes for "+loc+" ("+strerror( errno)+")" );
}
else {
fileh->SaveAttribs( file_stat );
}
}
for (unsigned int k=0; k<fileh->GetNumContents(); k++) {
pair<string, string> p = fileh->GetContent(k);
sfileh = db->FileGetHandle( profile, resource_types[i], p.second, false );
string loc = root+"/"+resource_types[i]+"/"+string(basename(sfileh->GetLocation().c_str()));
if (sfileh->GetSubtype() != "ghost" ) {
ifstream ins( loc.c_str() );
if (ins.fail()) {
delete fileh;
delete sfileh;
throw operation_failed( "could not open "+loc );
}
sfileh->Save( ins );
ins.close();
struct stat file_stat;
if ( stat( loc.c_str(), &file_stat ) != 0 ) {
LOGWARN( "can not save attributes for "+loc+" ("+strerror( errno)+")" );
}
else {
fileh->SaveAttribs( file_stat );
}
}
delete sfileh;
sfileh=NULL;
}
delete fileh;
fileh=NULL;
}
}
	// whipe backups
	vector<SCDBNode<SCDB::backend_node>*> nc;
	try {
	db->SearchNodes( *db->root, "root*data*profile|name="+profile+"*backup", nc );
	for (unsigned int i=0; i<nc.size(); i++)
	nc[i]->SetDeleted(true);
	}
	catch( SCDB::not_found ) {
		// ok, no backups
		}
		LOGINFO( "imported profile "+profile );
		}
		catch( SCDB::db_action_failed &e ) {
		if (fileh) delete fileh;
		if (sfileh) delete sfileh;
		throw operation_failed( "SCDB error: "+e.reason );
		}
		catch( SCDBFile::read_error ) {
		if (fileh) delete fileh;
		if (sfileh) delete sfileh;
		throw operation_failed( "SCDB read error" );
		}
		catch( SCDBFile::write_error ) {
		if (fileh) delete fileh;
		if (sfileh) delete sfileh;
		throw operation_failed( "SCDB write error" );
		}
		}
*/

/*
	void SCDBUtil::RebuildData( ) throw( operation_failed )
	{
	try {
	vector<string> &profiles = db->GetProfiles();
		// clear data tree first, but preserve structure
		for (unsigned int i=0; i<profiles.size(); i++) {
		SCDBNode *data_node = db->GetNode( "data*profile|name="+profiles[i] );
		}
		vector<string> &types = db->ResourceGetTypes();
		for (unsigned int i=0; i<profiles.size(); i++) {
		db->ProfileClearResources( profiles[i] );
		for (unsigned int j=0; j<types.size(); j++) {
		vector<string> &names = db->ResourceGetNames( types[j] );
		for (unsigned int k=0; j<names.size(); k++) {
		db->ProfileAddResource( profiles[i], names[k], types[j] );
		}
		}
		}
		}
		catch( SCDB::db_action_failed &e ) {
		throw operation_failed( e.reason );
		}
		catch( SCDB::not_found ) {
		throw operation_failed( "scdb error (key not found)" );
		}
		catch( SCDB::not_unique ) {
		throw operation_failed( "scdb error (key not unique)" );
		}
		}
*/
