/*
 * $Id: xml.cc,v 1.2 2008/10/22 10:12:42 hschaa Exp $
 *
 *
 * Project		 :	SCPM (System Configuration Profile Management)
 * Module			 :	SCPM database
 * File				 :	xml.cc
 * Description :	class for loading and saving SCDB from/to a local file in XML format
 * Author			 :	Joachim Gleissner <jg@suse.de>
 *
 * Copyright 2002 SuSE Linux AG
 *
 * Released under the terms of the GNU General Public License
 * (see file COPYRIGHT in project root directory).
 *
 */
#include "xml.hh"
#include <cstdio>
#include <sstream>
#include <cstring>

using namespace std;

XML::XML( )
{
	root = new XMLNode( "root" );
	db_stats.node_count = 0;
	db_stats.deleted_node_count = 0;
	db_stats.mem_usage = 0;
}

XML::~XML( )
{
		// clean up tree
	delete root;
}


bool XML::Load( istream &input, bool quick ) throw( )
{
	Log *log = Log::GetHandle();
		// save istream object
	dbistream = &input;
		// seek scdb version
	string line;
	scdb_version_str="n/a";
	format_version_str="n/a";
		// do not search for version string after first non comment
	while ( getline( input, line ) && line.find("<?") != string::npos) {
			// compatibility
		if ( line.find( "<?scdb version " ) != string::npos ) {
			line.erase( line.find("<?") , strlen("<?scdb version ") );
			scdb_version_str = line.substr( 0, line.find_first_of(" >\"") );
		}
		if ( line.find( "<?scdb version=" ) != string::npos ) {
			line.erase( line.find("<?"), strlen("<?scdb version=") );
			if ( line.at(0) == '"' ) line.erase( 0, 1 );
			scdb_version_str = line.substr( 0, line.find_first_of(" >\"") );
		}
		if ( line.find( "<?scdb formatversion=" ) != string::npos ) {
			line.erase( line.find("<?"), strlen("<?scdb formatversion=") );
			if ( line.at(0) == '"' ) line.erase( 0, 1 );
			format_version_str = line.substr( 0, line.find_first_of(" >\"") );
		}			 
	}
	input.seekg(0, ios::beg);
	log->WriteMessage( "xml", Log::CLASS_DEBUG, "scdb version "+scdb_version_str );
	if ( format_version_str == "n/a" ) {
			// older SCPM version (pre 0.7) did not have formatversions
		log->WriteMessage( "xml", Log::CLASS_WARNING,
											 "db has no format version, assuming old (pre 0.7) version" );
		format_version_str = "ancient";
	}
	else if ( format_version_str != "1.0" ) {
		log->WriteMessage( "xml", Log::CLASS_WARNING, "unsupported format version "
											 +format_version_str+" (supported: 1.0)" );
	}
	try {
		CreateSubTree( *root, input, quick );
		char tmp[50];
		sprintf( tmp, "%i items read", db_stats.node_count );
		log->WriteMessage( "xml", Log::CLASS_DEBUG, tmp );
	}
	catch( op_failed ) {
		return false;
	}
	return true;
}

void XML::Save( ostream &output )
{
	output << "<?xml version=\"1.0\"?>" << endl;
	output << "<?scdb version=\""+scdb_version_str+"\"?>" << endl;
	output << "<?scdb formatversion=\"1.0\"?>" << endl;
	for( unsigned int i=0; i<root->GetNumSub(); i++ )
		this->DumpTree( (*root)[i], output );
}

bool XML::UpToDate( )
{
	if ( format_version_str != "1.0" ) return false;
	return true;
}

void XML::DumpTree( XMLNode &node, ostream &output )
{
	static int depth=0;

		// do not dump in case tree has deleted flag set
	if ( node.Deleted() ) return;

		// maybe we are on a branch which is not loaded yet
	if (!node.Available()) this->LoadBranch( node );

	if ( node.GetValue() == "" ) {
			// key only
		for ( int i=0; i<depth*3; i++ ) output << " ";
		output << "<" << node.GetName() << ">" << endl;
		
			// search subs
		depth++;
		for (unsigned int i=0; i<node.GetNumSub(); i++)
			DumpTree( node[i], output );
		depth--;

		for ( int i=0; i<depth*3; i++ ) output << " ";
		output << "</" << node.GetName() << ">" << endl;
	}
	else {
		for ( int i=0; i<depth*3; i++ ) output << " ";
		output << "<" << node.GetName() << ">"
					 << node.GetValue()
					 << "</" << node.GetName() << ">" << endl;
	}
}

void XML::LoadBranch( XMLNode &node ) throw( op_failed )
{
	if ( node.Available() ) return;
	Log *log = Log::GetHandle();
	if ( dbistream->bad() ) {
			// ohoh, we lost the database stream and we have unread nodes
		log->WriteMessage( "xml", Log::CLASS_ERROR, "database I/O error, lost stream" );
	}
	dbistream->seekg( node.filepos );
	char c;
	dbistream->get(c); // filepos is one byte before the wanted node
	CreateSubTree( node, *dbistream, false );
	node.available = true;
	char tmp[50];
	sprintf( tmp, "db_stats update: %i items read", db_stats.node_count );
	log->WriteMessage( "xml", Log::CLASS_DEBUG, "loaded branch at node "+node.GetName() );
	log->WriteMessage( "xml", Log::CLASS_DEBUG, tmp );
}

void XML::CreateSubTree( XMLNode &node, istream &input, bool quick ) throw( op_failed )
{
	Log *log = Log::GetHandle();
	XMLNode *new_node;
	key_t key;

	for (;;) {
		try {
			key = GetKey( input );
		}
		catch ( end_of_tokens ) {
			return;
		}
		catch ( syntax_error &s ) {
			ostringstream msg;
			msg << "syntax error (" << s.error << ") line "
					<< s.line << " near '" << s.key << "'";
			log->WriteMessage( "xml", Log::CLASS_WARNING, msg.str() );
			throw op_failed();
		}

		if ( key.is_begin_token ) {
			new_node = new XMLNode( key.str );
			new_node->filepos = key.fpos;
				//node.filepos = key.fpos;
			node.AddSub( *new_node );
			if ( quick && (&node==root) && (key.str=="data") ) {
					// in quick mode, we skip reading the data part
				log->WriteMessage( "xml", Log::CLASS_DEBUG, "skip reading data branch" );
				new_node->available=false;
				return;
			}
			else 
				CreateSubTree( *new_node, input, quick );
			db_stats.mem_usage += sizeof( node );
			db_stats.node_count++;
		}
		else if (key.is_end_token) {
			if ( node.GetName() == key.str ) return;
		}
		else {
			node.available = true;
			node.SetValue( key.str );
		}
	}
}

XML::key_t XML::GetKey( istream &input ) throw( syntax_error, end_of_tokens )
{
	static const int STATE_START=0;
	static const int STATE_TOKEN_OPEN=1;
	static const int STATE_TOKEN_READ=2;
	static const int STATEVALUE_READ=3;
	static const int STATE_SKIPPING=4;
	int state;
	static int line_number=1;
	static key_t k;
	char c;

	k.str.clear();
	k.is_begin_token = false;
	k.is_end_token = false;

//	line_number = 0;
	state = STATE_START;

	for (;;) {
		streampos p = input.tellg();
		c = input.get();
		k.fpos = p;

		if ( c=='\n' ) line_number++; // always count line numbers
		if ( (state==STATE_SKIPPING) && (c!='<') ) continue;
		if ( (state==STATE_SKIPPING) && (c=='<') ) state=STATE_START;

		switch (c) {
			case '\xFF':
				if (state!=STATE_START) 
					throw syntax_error( "", "unexpected end of file", line_number);
				else throw end_of_tokens();
			case '!':
			case '?':
				if (state == STATE_TOKEN_OPEN) {
						// comment or ?xml ... reached, skipping
					state=STATE_SKIPPING;
					continue;
				}
				if (state != STATEVALUE_READ) {
					throw syntax_error( k.str, "unexpected '!' or '?'", line_number );
				}
				k.str += c;
				break;
			case ' ':
				if (state == STATEVALUE_READ) {
						// white spaces are allowed in values
						// so add it (ignore in all other cases)
					k.str += c;
					continue;
				}
			case '\t':
				if (state == STATEVALUE_READ) {
						// tabs are allowed in values
						// so add it (ignore in all other cases)
					k.str += c;
				}
				continue;
			case '\n':
				if (state == STATEVALUE_READ) {
						// newlines are allowed in values
						// so add it (ignore in all other cases)
					k.str += c;
				}
				continue;
			case '<':
				if (state == STATEVALUE_READ) {
						// value fully read, next token reached
					state = STATE_TOKEN_OPEN;
						// need to roll back
					input.seekg( p );
					return k;
				}
				if (state != STATE_START) { 
					throw syntax_error( k.str, "unexpected '<'", line_number );
				}
				state=STATE_TOKEN_OPEN;
				k.is_begin_token=true;
				break;
			case '/':
				if (state == STATE_TOKEN_OPEN) {
					k.is_end_token=true;
					k.is_begin_token=false;
					continue;
				}
				if (state == STATE_START) state=STATEVALUE_READ;
				if (state == STATE_TOKEN_READ) {
					throw syntax_error( k.str, "invalid char / in tag specifier", line_number );
				}
				k.str += c;
				break;
			case '>':
				if (state != STATE_TOKEN_READ) {
					throw syntax_error( k.str, "unexpected >", line_number );
				}
				return k;
			default:
				if (state == STATE_START) state=STATEVALUE_READ;
				if (state == STATE_TOKEN_OPEN) state=STATE_TOKEN_READ;
				k.str += c;
		}
	}
}
